#include "Solution.h"
#include "Manager.h"

using namespace std;

BranchAndBound::BranchAndBound(Random *rnd, Graph* pin, Control *ctrl) :
	ColoringManager(rnd, pin, ctrl) {
	N = pin->Vertices();
	NodesData.resize(N);

	U.resize(N, 0);
	C.resize(N);

	for (Vertex i = 0; i < N; i++) {
		//NodesData[i].Node = i;
		NodesData[i].used_col.clear();
		NodesData[i].used_col.resize(N + 1, 0);
		NodesData[i].diff_col = 0;
		NodesData[i].adj_uncol = pin->degree[i];
		U[i] = 1;
		C[i] = N + 1;
	}
	tot_visited = 0;
}


BranchAndBound::BranchAndBound(Random *rnd, Graph* pin, Control *ctrl,
		Coloring &initial) :
	ColoringManager(rnd, pin, ctrl) {

	N = pin->Vertices();
	NodesData.resize(N);

	U.resize(N, 0);
	C.resize(N);

	for (Vertex i = 0; i < N; i++) {
		NodesData[i].used_col.clear();
		NodesData[i].used_col.resize(N + 1, 0);
		NodesData[i].diff_col = 0;
		NodesData[i].adj_uncol = 0;

		VertexSet::iterator pt;
		for (pt = p_in->adjacency_set[i].begin(); pt
				!= p_in->adjacency_set[i].end(); pt++) {

			if (initial.col[*pt] < (Color) N + 1) {
				if (NodesData[i].used_col[initial.col[*pt]] == 0) {
					NodesData[i].used_col[initial.col[*pt]] = N + 2;
					NodesData[i].diff_col++;
				}
			} else
				NodesData[i].adj_uncol++;

		}

		//cout<<NodesData[i].adj_uncol<<endl;
		C[i] = initial.col[i];
		if (initial.col[i] == (Color) N + 1) {
			U[i] = 1;
		} else {
			U[i] = 0;
		}

	}
	tot_visited = 0;
}


void BranchAndBound::printNode(std::ostream& os, const _NodeAttributes& a) {
	os << a.diff_col << "," << a.adj_uncol << endl;
	for (vector<Vertex>::const_iterator p = a.used_col.begin(); p
			!= a.used_col.end(); p++) {
		os << (*p) << " ";
	}
	os << endl;
}

Color BranchAndBound::chrom_number(Vertex Usize, Vertex first_u, Color B,
		Color K) {
	if (!p_ctrl->isTimeLeft())// && tmp.size() >= 1)
	{
		//(p_ctrl->getOutputStream())<<"- ";
		return B;

	}

	if (Usize == 1) {
		tot_visited++;
		register Vertex u = first_u;
		if (NodesData[u].diff_col < K) {
			//cout<<"1 "<<K<<endl;
			register Color col = 0;
			while ((col < K) && NodesData[u].used_col[col] > 0) {
				col++;
			}
			//assert(col<K);
			C[u] = col;
			// new best solution found
			p_ctrl->setCurrentCostXRLF(0, K, tot_visited);
			p_ctrl->IncrPartialIter(tot_visited);
			tot_visited = 0;
			best_C = C;
			return K;
		} else if ((K + 1) < B) {
			//cout<<"2 "<<K+1<<endl;
			C[u] = K;
			// new best solution found
			p_ctrl->setCurrentCostXRLF(0, K + 1, tot_visited);
			p_ctrl->IncrPartialIter(tot_visited);
			tot_visited = 0;
			best_C = C;
			return (K + 1);
		} else {
			//cout<<"3 "<<B<<endl;
			return B;
		}
	} else {
		//2.
		register Vertex u = VERTEX_MAX;
		register Color max_diff_col = 0;
		register Vertex max_adj_uncol = 0;

		for (register Vertex i = 0; i < N; i++) {
			if (U[i]) {
				if (NodesData[i].diff_col > max_diff_col) {
					u = i;
					max_diff_col = NodesData[i].diff_col;
					max_adj_uncol = NodesData[i].adj_uncol;
				} else if (NodesData[i].diff_col == max_diff_col) {
					if (NodesData[i].adj_uncol > max_adj_uncol) {
						u = i;
						max_adj_uncol = NodesData[i].adj_uncol;
					} else if (NodesData[i].adj_uncol == max_adj_uncol) {
						if (rg->ran0N(1)) {
							//cout<<"----->"<<endl;
							u = i;
						}
					}
				}
			}
		}

		// cout << " --> " << u << " " <<max_diff_col<< " "<<max_adj_uncol
		// <<" "<<Usize<<" B: "<<B<<" K: "<<K<< endl;
		//3.
		if (NodesData[u].diff_col == B - 1) {
			//ass.col[node]=col;
			return B;
		}
		// 4.
		for (Color j = 0; j < K; j++) {
			if (NodesData[u].used_col[j] == 0) {
				//color
				C[u] = j;
				U[u] = 0;

				//register vector<bool> was_the_first(N,0);
				register VertexSet::iterator pt;
				for (pt = p_in->adjacency_set[u].begin(); pt
						!= p_in->adjacency_set[u].end(); pt++) {
					NodesData[*pt].adj_uncol--;

					if (NodesData[*pt].used_col[j] == 0) {
						//was_the_first[(*pt)]=true;
						NodesData[*pt].used_col[j] = u + 1;
						NodesData[*pt].diff_col++;
					}
				}

				register Vertex next_first_u = first_u;
				if (u == first_u) {
					register Vertex i = u + 1;
					while (i < N && U[i] == 0)
						i++;
					next_first_u = i;
				}

				B = chrom_number(Usize - 1, next_first_u, B, K);

				//uncolor
				C[u] = N + 1;
				U[u] = 1;

				for (pt = p_in->adjacency_set[u].begin(); pt
						!= p_in->adjacency_set[u].end(); pt++) {
					NodesData[*pt].adj_uncol++;

					//if (was_the_first[(*pt)]==true)
					if (NodesData[*pt].used_col[j] == u + 1) {
						NodesData[*pt].used_col[j] = 0;
						NodesData[*pt].diff_col--;
					}
				}
			}
		}
		if (K < B - 1) {
			//cout<<"+";
			//color
			C[u] = K;
			U[u] = 0;

			register vector<bool> was_the_first(N, 0);
			register VertexSet::iterator pt;
			for (pt = p_in->adjacency_set[u].begin(); pt
					!= p_in->adjacency_set[u].end(); pt++) {
				NodesData[*pt].adj_uncol--;

				if (NodesData[*pt].used_col[K] == 0) {
					//was_the_first[(*pt)]=true;
					NodesData[*pt].used_col[K] = u + 1;
					NodesData[*pt].diff_col++;
				}
			}

			register Vertex next_first_u = first_u;
			if (u == first_u) {
				register Vertex i = u + 1;
				while (i < N && U[i] == 0)
					i++;
				next_first_u = i;
			}

			B = chrom_number(Usize - 1, next_first_u, B, K + 1);

			//uncolor
			C[u] = N + 1;
			U[u] = 1;

			for (pt = p_in->adjacency_set[u].begin(); pt
					!= p_in->adjacency_set[u].end(); pt++) {
				NodesData[*pt].adj_uncol++;

				//if (was_the_first[(*pt)]==true)
				if (NodesData[*pt].used_col[K] == u + 1) {
					NodesData[*pt].used_col[K] = 0;
					NodesData[*pt].diff_col--;
				}
			}
		}
	}
	return B;
}
