#include "Control.h"

char buffer[125];

Control::Control(Random *rnd, int argc, char** argv) {

	rg = rnd;
	parset = new Parameters();

	CALL_ABORT( parset->paramsetAddInt(
					"control/maxTry",
					"number of runs",
					&maxTry, FALSE, DEFAULT_NR_TRIAL,
					0, INT_MAX,
					NULL) );

	CALL_ABORT( parset->paramsetAddLongint(
					"solver/max_iter",
					"total number of iterations per run",
					&max_iter, FALSE, DEFAULT_TOT_MAX_ITER,
					0, LONG_MAX,
					NULL) );

	CALL_ABORT( parset->paramsetAddLongint(
						"solver/max_idle",
						"total number of iterations per run",
						&max_idle, FALSE, DEFAULT_TOT_MAX_ITER,
						0, LONG_MAX,
						NULL) );

	CALL_ABORT( parset->paramsetAddReal(
					"solver/timeLimit",
					"time limit per run",
					&timeLimit, FALSE, DEFAULT_TIME_LIMIT,
					0, DBL_MAX,
					NULL) );

#ifdef WIN32
	int c_time = (int) Timer::getCurrentTime();
#else
	int c_time = time(NULL);
#endif
	CALL_ABORT( parset->paramsetAddInt(
					"solver/seed",
					"seed changed incrementing by 1 at every trial",
					&seed, FALSE, c_time,
					0, INT_MAX,
					NULL) );

	//TODO: move elsewhere
	rg->set_seed(seed);

	CALL_ABORT( parset->paramsetAddInt(
					"solver/solver",
					"solver procedure",
					&solver, FALSE, DEFAULT_SOLVER,
					0, 100,
					NULL) );

	CALL_ABORT( parset->paramsetAddInt(
					"runner/runner",
					"runner procedure",
					&runner, FALSE, DEFAULT_RUNNER,
					0, 100,
					NULL) );

	CALL_ABORT( parset->paramsetAddInt(
					"cheur/init",
					"construction heuristic for initial solution",
					&init, FALSE, DEFAULT_CHEUR_INIT,
					0, 100,
					NULL) );

	CALL_ABORT( parset->paramsetAddInt(
					"cheur/decr",
					"construction heuristic for decreasing by one color",
					&decr, FALSE, DEFAULT_CHEUR_DECR,
					0, 100,
					NULL) );

	CALL_ABORT( parset->paramsetAddShortint(
					"solver/initk",
					"number of colors to start solving",
					&k, FALSE, DEFAULT_INIT_K,
					0, SINT_MAX,
					NULL) );

	CALL_ABORT( parset->paramsetAddShortint(
					"solver/finalk",
					"number of colors to stop",
					&finalK, FALSE, DEFAULT_FINAL_K,
					0, SINT_MAX,
					NULL) );

	CALL_ABORT( parset->paramsetAddShortint(
					"solver/lower_bound",
					"best known lower bound, used to improve preprocessing ",
					&lower_bound, FALSE, DEFAULT_LOWER_BOUND,
					0, SINT_MAX,
					NULL) );


	CALL_ABORT( parset->paramsetAddLongint(
					"runner/ils/lsiter",
					"number of iterations for surrogate local search",
					&ls_iter, FALSE, DEFAULT_LS_ITER,
					0, LLONG_MAX,
					NULL) );

	CALL_ABORT( parset->paramsetAddInt(
					"runner/ts/tltype",
					"type of tabu tenure",
					&tltype, FALSE, DEFAULT_TL_TYPE,
					1, 3,
					NULL) );

	CALL_ABORT( parset->paramsetAddInt(
					"runner/ts/invtype",
					"definition of inverse move to make tabu",
					&invtype, FALSE, DEFAULT_INV_TYPE,
					1, 5,
					NULL) );

	CALL_ABORT( parset->paramsetAddReal(
					"runner/ts/alpha",
					"multiplier factor of conflict set in tabu length",
					&alpha, FALSE, DEFAULT_ALPHA,
					0, DBL_MAX,
					NULL) );

	CALL_ABORT( parset->paramsetAddInt(
					"runner/ts/tl",
					"max tabu length ",
					&tl, FALSE, DEFAULT_TL,
					0, INT_MAX,
					NULL) );

	CALL_ABORT( parset->paramsetAddInt(
					"runner/ts/tl1",
					"left side of interval for tabu length ",
					&tl1, FALSE, DEFAULT_TL1,
					0, INT_MAX,
					NULL) );
	CALL_ABORT( parset->paramsetAddInt(
					"runner/ts/tl2",
					"right side of interval for tabu length ",
					&tl2, FALSE, DEFAULT_TL2,
					0, INT_MAX,
					NULL) );

	CALL_ABORT( parset->paramsetAddInt(
						"runner/xrlf/exactlim",
						"right side of interval for tabu length ",
						&exact_lim, FALSE, DEFAULT_EXACT_LIM,
						0, INT_MAX,
						NULL) );

	CALL_ABORT( parset->paramsetAddInt(
							"runner/xrlf/setlim",
							"right side of interval for tabu length ",
							&set_lim, FALSE, DEFAULT_SET_LIM,
							0, INT_MAX,
							NULL) );

	CALL_ABORT( parset->paramsetAddInt(
							"runner/xrlf/candnum",
							"right side of interval for tabu length ",
							&cand_num, FALSE, DEFAULT_CAND_NUM,
							0, INT_MAX,
							NULL) );

	CALL_ABORT( parset->paramsetAddInt(
							"runner/xrlf/trialnum",
							"right side of interval for tabu length ",
							&trial_num, FALSE, DEFAULT_TRIAL_NUM,
							0, INT_MAX,
							NULL) );


	CALL_ABORT( parset->paramsetAddInt(
					"solver/rrestart/max_restart",
					"num of restarts for random restart solver ",
					&max_restart, FALSE, DEFAULT_MAX_RESTARTS,
					0, INT_MAX,
					NULL) );

	CALL_ABORT( parset->paramsetAddBool(
					"control/print_statistics",
					"prints instance statistics and reduction",
					&print_statistics, FALSE, DEFAULT_PRINT_STATISTICS,
					NULL) );
	CALL_ABORT( parset->paramsetAddBool(
						"control/reduction",
						"whether to preprocess the graph ",
						&reduction, FALSE, DEFAULT_PREPROCESSING,
						NULL) );
	CALL_ABORT( parset->paramsetAddBool(
						"control/force_if_disconnected",
						"whether to preprocess the graph ",
						&force_if_disconnected, FALSE, DEFAULT_FORCE_IF_DISCONNECTED,
						NULL) );

	CALL_ABORT( parset->paramsetAddBool(
							"control/certificate",
							"returns a certificate of the upper bound found ",
							&certificate, FALSE, DEFAULT_PRINT_CERTIFICATE,
							NULL) );


	CALL_ABORT( parset->paramsetAddInt(
						"runner/sa/temperature",
						"the initial temperature (int) (suggested 10) ",
						&temperature, FALSE, DEFAULT_SA_TEMPERATURE,
						0, INT_MAX,
						NULL) );

	CALL_ABORT( parset->paramsetAddInt(
							"runner/sa/freezelim",
							"the freeze_lim value (int) (suggested 5) ",
							&freeze_lim, FALSE, DEFAULT_SA_FREEZE_LIM,
							0, INT_MAX,
							NULL) );

	CALL_ABORT( parset->paramsetAddReal(
					"runner/sa/cutoff",
					"the cutoff value (double) (suggested 0.10)",
					&cutoff, FALSE, DEFAULT_SA_CUTOFF,
					0, DBL_MAX,
					NULL) );

	CALL_ABORT( parset->paramsetAddReal(
						"runner/sa/minpercent",
						" the minpercent value (double) (suggested 0.02 - default INT_MAX = no embedded stopping criterion)",
						&minpercent, FALSE, DEFAULT_SA_MINPERCENT,
						0, DBL_MAX,
						NULL) );
	CALL_ABORT( parset->paramsetAddReal(
							"runner/sa/tempfactor",
							"the tempfactor value (double) ",
							&tempfactor, FALSE, DEFAULT_SA_TEMPFACTOR,
							0, DBL_MAX,
							NULL) );

	CALL_ABORT( parset->paramsetAddReal(
							"runner/sa/sizefactor",
							"the tempfactor value (double)",
							&sizefactor, FALSE, DEFAULT_SA_SIZEFACTOR,
							0, DBL_MAX,
							NULL) );
	CALL_ABORT( parset->paramsetAddReal(
								"runner/gls/lambda",
								"factor to weight the influence of penalties in GLS",
								&lambda, FALSE, DEFAULT_GLS_LAMBDA,
								0, DBL_MAX,
								NULL) );

	CALL_ABORT( parset->paramsetAddInt(
								"runner/gls/sidewalks",
								"number of sidewalks moves in GLS ",
								&sidewalks, FALSE, DEFAULT_GLS_SIDEWALKS,
								0, INT_MAX,
								NULL) );

	CALL_ABORT( parset->paramsetAddInt(
									"runner/ils/pert",
									"perturbation type: 0 sequential, 1 dsatur, 2 rebuild",
									&pert, FALSE, DEFAULT_ILS_PERTURB,
									0, INT_MAX,
									NULL) );

	CALL_ABORT( parset->paramsetAddReal(
									"runner/ils/gamma",
									"factor to determine the number of colors to remove in ILS",
									&gamma, FALSE, DEFAULT_ILS_GAMMA,
									0, DBL_MAX,
									NULL) );

	CALL_ABORT( parset->paramsetAddReal(
									"runner/ils/ilsprob",
									"acceptance criterion in ILS (0=Random Walk 1=AlwaysBest)",
									&ilsprob, FALSE, DEFAULT_ILS_ILSPROB,
									0, DBL_MAX,
									NULL) );

	CALL_ABORT( parset->paramsetAddReal(
									"runner/nov/prob",
									"probility p in Novelty+ ",
									&prob, FALSE, DEFAULT_NOV_PROB,
									0, DBL_MAX,
									NULL) );


	//parset->paramsetSetToDefault();

	SetDefaultParameters();
	ComParser(argc, argv);
	parset->paramsetWrite(NULL, TRUE, FALSE);
}

Control::~Control() {
	if (os != NULL && os != &cout)
		delete os;
}

void Control::SetDefaultParameters() {
	K_I = 0;

	nrTry = 0;

	//#ifdef WIN32
	//	seed = (int) Timer::getCurrentTime();
	//#else
	//	seed = time(NULL);
	//#endif

	param_set_file = DEFAULT_PARAM_SET_FILE;

}

void Control::Usage() {
	cerr << "\nGraph Coloring\n";
	cerr << "usage: gcp [- see below -] [dat_file] (given are short options)\n";
	cerr << "-i f\t inputfile (DIMACS format necessary)\n";
	cerr << "-o f\t outputfile \n";
	cerr << "-n #\t number of tests\n";
	cerr << "-t #\t maximum time\n";
	cerr << "-m #\t number of iterations for each trial\n";
	cerr << "                      \t if flag not present maxiter=LONG_MAX\n";
	cerr << "                      \t if flag without arg maxiter=10000*N\n";
	cerr << "                      \t if --max=A          maxiter=A\n";
	cerr << "-s #\t seed\n";

	cerr << "--solver #\t Solver: \n";
	cerr
			<< "                      \t 0 => Random Restart - initial k flag to be set!!!\n";
	cerr
			<< "                      \t 1 => Metaheuristic with k Fixed - initial k to be set\n";
	cerr << "                      \t 2 => Metaheuristic Descend\n";
	cerr << "                      \t 3 => MultiRunnerSolver\n";
	cerr
			<< "                      \t 4 => Hybrid Evolutionary Algorithm Fixed - initial k flag to be set!!!\n";
	cerr << "                      \t 6 => Penalty Function (Kempe Chains)\n";
	cerr << "                      \t 100 => Constructive Heuristics\n";
	cerr << "--neigh                \t 1 one ex\n";
	cerr
			<< "                      \t 45 Kempe Chains Simulated Annealing (see -sahelp)\n";
	cerr << "                      \t 50 Guided Local Search\n";
	cerr << "                      \t 60 Iterated Local Search\n";
	cerr << "                      \t 71 MinConflicts+\n";
	cerr << "                      \t 80\n";
	cerr << "                      \t 100 without explorer for heuristics\n";
	cerr << "--test #\t run test version (1) or not (0)\n";
	cerr << "--init #\t way of initializing a solution: \n";
	cerr
			<< "                      \t 0 => completely random - initial k flag to be set!!!\n";
	cerr
			<< "                      \t 1 => construction Greedy - initial k produced by CG\n";
	cerr
			<< "                      \t 2 => construction Greedy with fixed k - initial k flag to be set!!!\n";
	cerr
			<< "                      \t 3 => construction RLF - initial k produced by CG\n";
	cerr
			<< "                      \t 4 => construction RLF with fixed k - initial k flag to be set!!!\n";
	cerr
			<< "                      \t 5 => construction DSATUR - initial k produced by CG\n";
	cerr
			<< "                      \t 6 => construction DSATUR with fixed k - initial k flag to be set!!!\n";
	cerr << "                      \t 7 => Read from file\n";
	cerr << "                      \t 9 => construction Static RLFDSATUR\n";
	cerr << "f f\t name of the file where a starting solution is recorded.\n";
	cerr << "--k #\t initial coloring\n";
	cerr << "--lb #\t best lower bound known. It improves the preprocessing\n";

	cerr << "--decr #\t kind of decreasing:\n";
	cerr
			<< "                      \t 0 => (default) with random reassignament\n";
	cerr << "                      \t 2 => Constructive Greedy with k fixed\n";
	cerr << "                      \t 4 => with RLF reassignament\n";
	cerr << "                      \t 6 => with DSAT reassignament\n";

	cerr << "                      \t 10 => random bounded assignament\n";
	cerr << "                      \t 12 => Bounded Construction Greedy\n";
	cerr << "                      \t 14 => Bounded Construction RLF\n";
	cerr << "                      \t 16 => Bounded Construction DSAT\n";

	cerr << "--tltype #\t length policy:\n";
	cerr << "                      \t 1 => (default) adaptive policy\n";
	cerr
			<< "                      \t      requires -alpha <factor> -tl <ref_length>\n";
	cerr << "                      \t 2 => interval policy\n";
	cerr
			<< "                      \t      requires -tl1 <min_length> -tl2 <max_length>\n";
	cerr << "                      \t 3 => fixed length policy\n";
	cerr << "                      \t      requires -tl <length>\n";

	cerr << "--restart #\t n.of restart for Solver 0\n";

	cerr << "--ls_iter #\t used if solver [4,5] to limit the ts iterations\n\n";
	cerr << "--K #\t final coloring\n";
	cerr << "--lambda #\t factor to weight the influence of penalties in GLS\n";
	cerr
			<< "--gamma #\t factor to determine the number of colors to remove in ILS\n";
	cerr
			<< "--pert  #\t which perturbation to use: 0 sequential, 1 dsatur, 2 rebuild\n";
	cerr
			<< "--ilsprob #\t acceptance criterion in ILS (0=Random Walk 1=AlwaysBest)\n";
	cerr << "--tl #\t tabu lenght for MinConflicts\n";
	cerr << "--prob #\t probility p in Novelty+\n";

	cerr << "--stat #\t prints instance statistics and reduction\n";
	cerr
			<< "\nOptional arguments must NOT be separated from their option";
	cerr
			<< "character.  Therefore, if \"--max\" accepts an optional argument,\n";
	cerr
			<< "\"--max 10\" is considered as \"--max\" without argument followed by\n";
	cerr << "\"10\" which is a non-option command line argument.\n\n";
	cerr << "--morehelp  \t help: prints this information\n";

}

void Control::SAUsage() {
	cerr
			<< "Additional parameters for SA Johnsons implementation (see corresponding paper).\n";
	cerr << "Suggested values are also default values\n";
	cerr
			<< "--temperature               \t the initial temperature (int) (suggested 10)\n";
	cerr
			<< "--cutoff                    \t the cutoff value (double) (suggested 0.10)\n";
	cerr
			<< "--minpercent                \t the minpercent value (double) (suggested 0.02 ";
	cerr << "- default INT_MAX = no embedded stopping criterion)\n";
	cerr
			<< "--freeze_lim                \t the freeze_lim value (int) (suggested 5)\n";
	cerr << "--tempfactor                \t the tempfactor value (double) \n";
	cerr << "--sizefactor                \t the sizefactor value (double) \n\n";
	cerr << "Copy n paste: -temperature 10 -cutoff 0.1 -minpercent 0.02";
	cerr << " --freeze_lim 5 -tempfactor 1 -sizefactor 1\n";
}

void Control::informationsReport() {
	ifstream info;
	char line[MAXSTRLEN];
	char *trailer;
	(*os)
			<< "===============================================================\n";
	(*os) << "Current Machine: \n";

	info.open("/proc/cpuinfo", ios::in);
	if (info.is_open()) {
		while (!info.eof()) {
			info.getline(line, 100);
			if (info.eof())
				break;
			if (strstr(line, "processor") != NULL)
				(*os) << line << "\n";
			else if (strstr(line, "vendor_id") != NULL)
				(*os) << line << "\n";
			else if (strstr(line, "model name") != NULL)
				(*os) << line << "\n";
			else if (strstr(line, "cpu MHz") != NULL)
				(*os) << line << "\n";
			else if (strstr(line, "cache size") != NULL)
				(*os) << line << "\n";
			else if (strstr(line, "bogomips") != NULL)
				(*os) << line << "\n";
			else if (strstr(line, "EOF") != NULL)
				break;
			else
				continue;
		}
		info.close();
	} else {
		(*os)
				<< "Couldn`t find file /proc/cpuinfo for current machine type data\n";
	}
	(*os) << "\n";

	(*os) << "OS:\n";
	info.open("/etc/SuSE-release", ios::in);
	if (info.is_open()) {
		info.getline(line, 200);
		(*os) << line << "\n";
		info.close();
	} else {
		(*os) << "Couldn`t find SuSE version.\n";
	}
	(*os) << "\n";

	(*os) << "Compiler: " << endl;

	if (os == &cout)
		system("g++ -v");

	(*os) << "Flags:\n";
	info.open("Makefile", ios::in);
	if (info.is_open()) {
		while (!info.eof()) {
			info.getline(line, MAXSTRLEN);
			trailer = strstr(line, "CXXFLAGS =");
			if (trailer != NULL && strlen(trailer) > 0)
				(*os) << trailer << "\n";
		}
		info.close();
	} else {
		(*os) << "Couldn`t find Makefile \n";
	}

	double initialization = getTime();
	(*os) << "Initialization Time: " << (initialization < 0 ? 0.0
			: initialization) << "\n";
	(*os)
			<< "===============================================================\n";
	(*os) << "Begin problem " << basename(is) << "\n";

}

char * Control::readline(istream &infile) {
	size_t length;
	size_t i;

	//char *result = f
	//cout << "entra/n";
	if (infile.getline(buffer, 250)) {
		length = strlen(buffer);
		//if (length > 0 && buffer[length-1] != 0)
		//(*os_log)<<"Buffer size in getline() isn't large enough\n"<<buffer;
		//cout << buffer << endl;
		for (i = 0; i < length; ++i)
			if (buffer[i] != ' ' || buffer[i] != '\t')
				return buffer;
	} else
		return 0;
	return readline(infile);
}

char * Control::match(char *source, char *prefix) {

	size_t prefixLength = strlen(prefix);

	if (strncmp(source, prefix, prefixLength) != 0)
		return 0;
	source += prefixLength;
	while ((*source) == ' ' || (*source) == '\t')
		source++;
	if (*source == ':' || *source == '=')
		source++;
	while ((*source) == ' ' || (*source) == '\t')
		source++;

	return source;
}

void Control::resetTime() {
	(*os) << "Reset timer" << endl;
	timer.resetTime();
}

double Control::getTime() {
	return timer.elapsedTime();
}

void Control::beginTry() {
	rg->set_seed(seed++);
	++nrTry;
	(*os) << "Begin try " << nrTry << endl;
	resetTime();
	bestCost = FVALUE_MAX;
	best_k = COLOR_MAX;
	partial_Iter = 0;
}

void Control::endTry() {
	double time = getTime();
	(*os) << "exit iter " << partial_Iter;
	(*os) << " time ";
	os->flags(ios::fixed);
	(*os) << (time < 0 ? 0.0 : time) << endl;
	(*os) << "End try " << nrTry << endl;
	//cout<<bestCost<<"\t"<<seed-1<<endl;
}

void Control::setCurrentCost(fvalue cost, Color cl, Longint iter) {
	if (solver == 13) {
		double time = getTime();
		(*os) << "best " << (int) cost;
		(*os) << " col " << cl << " time ";
		os->flags(ios::fixed);
		(*os) << (time < 0 ? 0.0 : time);
		(*os) << " iter " << (int) iter + partial_Iter;
		(*os) << " par_iter " << (int) iter << endl;

	} else if (isTimeLeft()) {
		if (cl < best_k) {
			best_k = cl;
			//currentCost = cost;
			bestCost = cost;
			double time = getTime();
			(*os) << "best " << (int) bestCost;
			(*os) << " col " << best_k << " time ";
			os->flags(ios::fixed);
			(*os) << (time < 0 ? 0.0 : time);
			(*os) << " iter " << (int) iter + partial_Iter;
			(*os) << " par_iter " << (int) iter << endl;
		} else if (cl == best_k) {
			if (cost < bestCost) {
				bestCost = cost;
				double time = getTime();
				(*os) << "best " << (int) bestCost;
				(*os) << " col " << best_k << " time ";
				os->flags(ios::fixed);
				(*os) << (time < 0 ? 0.0 : time);
				(*os) << " iter " << (int) iter + partial_Iter;
				(*os) << " par_iter " << (int) iter << endl;
			}
		} else if (runner == 30) {

			double time = getTime();
			(*os) << "best " << (int) cost;
			(*os) << " col " << cl << " time ";
			os->flags(ios::fixed);
			(*os) << (time < 0 ? 0.0 : time);
			(*os) << " iter " << (int) iter + partial_Iter;
			(*os) << " par_iter " << (int) iter << endl;
		}
	}

}

void Control::setCurrentCostXRLF(fvalue cost, Color cl, Longint iter) {
	if (cl + K_I < best_k) {
	  best_k = cl + K_I;
		//currentCost = cost;
		bestCost = cost;
		double time = getTime();
		(*os) << "best " << (int) bestCost;
		(*os) << " col " << best_k << " time ";
		os->flags(ios::fixed);
		(*os) << (time < 0 ? 0.0 : time);
		(*os) << " iter " << (int) iter + partial_Iter;
		(*os) << " par_iter " << (int) iter << endl;
	}
}

void Control::setHEACrossovers(unsigned int cross, unsigned int restart) {
	(*os) << "crossovers " << cross << " restart " << restart << endl;
}

