#ifndef CONTROL_H
#define CONTROL_H

#ifdef WIN32
#pragma warning(disable:4786)
#endif

#include <string>
#include <map>
#include <iostream>
#include <cfloat>
#include "Definitions.h"
#include "Utilities.h"
#include "Parameters.h"
#include "Settings.h"

using namespace std;

#include "Timer.h"
#include "Random.h"
#include <map>

class Control {
private:
	//map< string, string > parameters;
	Timer timer;

	int nrTry, maxTry;
	double bestCost;

Longint  partial_Iter;
	int seed;
	double timeLimit;

	ostream *os;

	const char *ss;

public:

	Control(Random *rnd, int, char**); // construct a control object, given a set of command line options
	~Control();
	const char *is;
	const char *param_set_file;

	int init;
	int decr;

	Longint ls_iter;

	//  TABU LIST PARAMETERS
	double alpha;
	int tltype;
	int tl;
	int tl1;
	int tl2;
	int invtype;

	int runner;
	int max_restart;
	int solver;
	Longint max_iter;
	Longint max_idle;

	Bool print_statistics;
	Bool reduction;
	Bool certificate;
	string os_certificate;
	Bool force_if_disconnected;

	Color lower_bound;
	Color finalK;
	Color best_k;
	//SA Johnson's
	int temperature;
	int freeze_lim;
	double cutoff;
	double minpercent;
	double tempfactor;
	double sizefactor;
	//ILS
	double gamma;
	int pert;
	double ilsprob;
	//Nov
	double prob;

	//GLS
	double lambda;
	int sidewalks;

	//XRLF
	int set_lim;
	int cand_num;
	int trial_num;
	int exact_lim;

	Vertex real_size_solved;

	Color k;

	Color K_I;

	void Usage();
	void SAUsage();
	void ComParser(int argc, char** argv);
	int getSeed() {
		return seed;
	} // get the seed for the random number generator
	void resetTime(); // reset the timer
	double getTime(); // get the timer's time
	bool isTimeLeft() {
		return (getTime() < timeLimit);
	}
	; // check whether our run has time left
	int getNumTry() {
		return (nrTry);
	}
	;

	void beginTry(); // begin a run
	void endTry(); // end a run

	void setCurrentCost(fvalue cost, Color cl, Longint iter); // tell control the current cost of your solution
	void setCurrentCostXRLF(fvalue cost, Color cl, Longint iter); // tell control the current cost of your solution
	void SetColorsXRLF(Color c) {
		K_I = c;
	}
	;
	void setPartialIter(Longint iter) {
		partial_Iter = iter;
	}
	;
	void IncrPartialIter(Longint iter) {
		partial_Iter += iter;
	}
	;
	void setHEACrossovers(unsigned int cross, unsigned int restart);

	const char *getInputStream() {
		return is;
	} // get the input stream for constructing problems
	const char *getInputSolutionStream() {
		return ss;
	} // get the input stream for constructing problems
	ostream& getOutputStream() {
		return *os;
	} // get the output stream used for writing output data
	int getNumberOfTries() {
		return maxTry;
	} // get the max number of tries
	bool triesLeft() {
		return (nrTry < maxTry);
	} // check whether more tries need to be run
	double getTimeLimit() {
		return timeLimit;
	} // get the time limit of one run
	void informationsReport();
	char *readline(istream &infile);
	char *match(char *source, char *prefix);
	void SetDefaultParameters();
	const char *fullPathToInstance() {return is;};
	Random *rg;
	Parameters *parset;

};

#endif
