#include "Control.h"

char buffer[125];

Control::Control(Random *rnd, int argc, char** argv) {

	rg = rnd;
	parset = new Parameters();

	CALL_ABORT( parset->paramsetAddInt(
					"control/maxTry",
					"number of runs",
					&maxTry, FALSE, DEFAULT_NR_TRIAL,
					0, INT_MAX,
					NULL) );

	CALL_ABORT( parset->paramsetAddLongint(
					"solver/max_iter",
					"total number of iterations per run",
					&max_iter, FALSE, DEFAULT_TOT_MAX_ITER,
					0, LLONG_MAX,
					NULL) );

	CALL_ABORT( parset->paramsetAddReal(
					"solver/timeLimit",
					"time limit per run",
					&timeLimit, FALSE, DEFAULT_TIME_LIMIT,
					0, DBL_MAX,
					NULL) );

#ifdef WIN32
	int c_time = (int) Timer::getCurrentTime();
#else
	int c_time = time(NULL);
#endif
	CALL_ABORT( parset->paramsetAddInt(
					"solver/seed",
					"seed changed incrementing by 1 at every trial",
					&seed, FALSE, c_time,
					0, LLONG_MAX,
					NULL) );

	//TODO: move elsewhere
	rg->set_seed(seed);

	CALL_ABORT( parset->paramsetAddInt(
					"solver/solver",
					"solver procedure",
					&solver, FALSE, DEFAULT_SOLVER,
					0, 100,
					NULL) );

	CALL_ABORT( parset->paramsetAddInt(
						"runner/runner",
						"runner procedure",
						&runner, FALSE, DEFAULT_RUNNER,
						0, 100,
						NULL) );

	CALL_ABORT( parset->paramsetAddInt(
							"cheur/init",
							"construction heuristic for initial solution",
							&init, FALSE, DEFAULT_CHEUR_INIT,
							0, 100,
							NULL) );

	CALL_ABORT( parset->paramsetAddInt(
								"cheur/decr",
								"construction heuristic for decreasing by one color",
								&decr, FALSE, DEFAULT_CHEUR_DECR,
								0, 100,
								NULL) );

	CALL_ABORT( parset->paramsetAddInt(
									"solver/k",
									"number of colors to use",
									&k, FALSE, DEFAULT_CHEUR_DECR,
									0, 100,
									NULL) );

	//parset->paramsetSetToDefault();

	SetDefaultParameters();
	ComParser(argc, argv);
	parset->paramsetWrite(NULL, TRUE, FALSE);

	cerr << "k " << k << endl;
	cerr << "test_version " << test_version << endl;

	cerr << "ls_iter " << ls_iter << endl;
	cerr << "tltype " << tltype << endl;
	cerr << "invtype " << invtype << endl;
	cerr << "tl " << tl << endl;
	cerr << "tl1 " << tl1 << endl;
	cerr << "tl2 " << tl2 << endl;
	cerr << "alpha " << alpha << endl;
	cerr << "max_restart " << max_restart << endl;
	cerr << "lower_bound " << lower_bound << endl;
	cerr << "print_statistics " << print_statistics << endl;

	cerr << "temperature " << temperature << endl;
	cerr << "cutoff " << cutoff << endl;
	cerr << "minpercent " << minpercent << endl;
	cerr << "freeze_lim " << freeze_lim << endl;
	cerr << "tempfactor " << tempfactor << endl;
	cerr << "sizefactor " << sizefactor << endl;

	cerr << "lambda " << lambda << endl;
	cerr << "gamma " << gamma << endl;
	cerr << "tl " << tl << endl;
	cerr << "prob " << prob << endl;
	cerr << "ilsprob " << ilsprob << endl;

	cerr << "reduction " << reduction << endl;
	cerr << "verify " << certificate << endl;
	//exit(1);
	//cout<<os_certificate<<endl;
	// exit(1);
}

Control::~Control() {
	if (os != NULL && os != &cout)
		delete os;
}

void Control::SetDefaultParameters() {
	K_I = 0;

	nrTry = 0;

//#ifdef WIN32
//	seed = (int) Timer::getCurrentTime();
//#else
//	seed = time(NULL);
//#endif

	param_set_file = DEFAULT_PARAM_SET_FILE;
//	rg->set_seed(seed);
	k = 1000;

	test_version = 0;
	ls_iter = LONG_MAX;
	//
	alpha = 0.6;
	tltype = 0;
	tl = 10;
	tl1 = 0;
	tl2 = 0;
	invtype = INV_OLD;

	max_restart = 0;
	lower_bound = 0;
	print_statistics = false;

	// SA
	temperature = 10;
	cutoff = 0.1;
	minpercent = 0.02;
	freeze_lim = INT_MAX;//5;
	tempfactor = -1;
	sizefactor = -1;

	//GLS
	lambda = 0;
	sidewalks = 0;

	gamma = 0;
	prob = 0;
	ilsprob = 1;

	reduction = false;
	certificate = false;

}

void Control::Usage() {
	cerr << "\nGraph Coloring\n";
	cerr << "usage: gcp [- see below -] [dat_file] (given are short options)\n";
	cerr << "-i f\t inputfile (DIMACS format necessary)\n";
	cerr << "-o f\t outputfile \n";
	cerr << "-n #\t number of tests\n";
	cerr << "-t #\t maximum time\n";
	cerr << "-m #\t number of iterations for each trial\n";
	cerr << "                      \t if flag not present maxiter=LONG_MAX\n";
	cerr << "                      \t if flag without arg maxiter=10000*N\n";
	cerr << "                      \t if --max=A          maxiter=A\n";
	cerr << "-s #\t seed\n";

	cerr << "--solver #\t Solver: \n";
	cerr
			<< "                      \t 0 => Random Restart - initial k flag to be set!!!\n";
	cerr
			<< "                      \t 1 => Metaheuristic with k Fixed - initial k to be set\n";
	cerr << "                      \t 2 => Metaheuristic Descend\n";
	cerr << "                      \t 3 => MultiRunnerSolver\n";
	cerr
			<< "                      \t 4 => Hybrid Evolutionary Algorithm Fixed - initial k flag to be set!!!\n";
	cerr << "                      \t 6 => Penalty Function (Kempe Chains)\n";
	cerr << "                      \t 100 => Constructive Heuristics\n";
	cerr << "--neigh                \t 1 one ex\n";
	cerr
			<< "                      \t 45 Kempe Chains Simulated Annealing (see -sahelp)\n";
	cerr << "                      \t 50 Guided Local Search\n";
	cerr << "                      \t 60 Iterated Local Search\n";
	cerr << "                      \t 71 MinConflicts+\n";
	cerr << "                      \t 80\n";
	cerr << "                      \t 100 without explorer for heuristics\n";
	cerr << "--test #\t run test version (1) or not (0)\n";
	cerr << "--init #\t way of initializing a solution: \n";
	cerr
			<< "                      \t 0 => completely random - initial k flag to be set!!!\n";
	cerr
			<< "                      \t 1 => construction Greedy - initial k produced by CG\n";
	cerr
			<< "                      \t 2 => construction Greedy with fixed k - initial k flag to be set!!!\n";
	cerr
			<< "                      \t 3 => construction RLF - initial k produced by CG\n";
	cerr
			<< "                      \t 4 => construction RLF with fixed k - initial k flag to be set!!!\n";
	cerr
			<< "                      \t 5 => construction DSATUR - initial k produced by CG\n";
	cerr
			<< "                      \t 6 => construction DSATUR with fixed k - initial k flag to be set!!!\n";
	cerr << "                      \t 7 => Read from file\n";
	cerr << "                      \t 9 => construction Static RLFDSATUR\n";
	cerr << "f f\t name of the file where a starting solution is recorded.\n";
	cerr << "--k #\t initial coloring\n";
	cerr << "--lb #\t best lower bound known. It improves the preprocessing\n";

	cerr << "--decr #\t kind of decreasing:\n";
	cerr
			<< "                      \t 0 => (default) with random reassignament\n";
	cerr << "                      \t 2 => Constructive Greedy with k fixed\n";
	cerr << "                      \t 4 => with RLF reassignament\n";
	cerr << "                      \t 6 => with DSAT reassignament\n";

	cerr << "                      \t 10 => random bounded assignament\n";
	cerr << "                      \t 12 => Bounded Construction Greedy\n";
	cerr << "                      \t 14 => Bounded Construction RLF\n";
	cerr << "                      \t 16 => Bounded Construction DSAT\n";

	cerr << "--tltype #\t length policy:\n";
	cerr << "                      \t 1 => (default) adaptive policy\n";
	cerr
			<< "                      \t      requires -alpha <factor> -tl <ref_length>\n";
	cerr << "                      \t 2 => interval policy\n";
	cerr
			<< "                      \t      requires -tl1 <min_length> -tl2 <max_length>\n";
	cerr << "                      \t 3 => fixed length policy\n";
	cerr << "                      \t      requires -tl <length>\n";

	cerr << "--restart #\t n.of restart for Solver 0\n";

	cerr << "--ls_iter #\t used if solver [4,5] to limit the ts iterations\n\n";
	cerr << "--K #\t final coloring\n";
	cerr << "--lambda #\t factor to weight the influence of penalties in GLS\n";
	cerr
			<< "--gamma #\t factor to determine the number of colors to remove in ILS\n";
	cerr
			<< "--pert  #\t which perturbation to use: 0 sequential, 1 dsatur, 2 rebuild\n";
	cerr
			<< "--ilsprob #\t acceptance criterion in ILS (0=Random Walk 1=AlwaysBest)\n";
	cerr << "--tl #\t tabu lenght for MinConflicts\n";
	cerr << "--prob #\t probility p in Novelty+\n";

	cerr << "--stat #\t prints instance statistics and reduction\n";
	cerr
			<< "\nOptional arguments must be separated from their option by a \"=\"";
	cerr
			<< "character.  Therefore, if \"--max\" accepts an optional argument,\n";
	cerr
			<< "\"--max 10\" is considered as \"--max\" without argument followed by\n";
	cerr << "\"10\" which is a non-option command line argument.\n\n";
	cerr << "--morehelp  \t help: prints this information\n";

}

void Control::SAUsage() {
	cerr
			<< "Additional parameters for SA Johnsons implementation (see corresponding paper).\n";
	cerr << "Suggested values are also default values\n";
	cerr
			<< "--temperature               \t the initial temperature (int) (suggested 10)\n";
	cerr
			<< "--cutoff                    \t the cutoff value (double) (suggested 0.10)\n";
	cerr
			<< "--minpercent                \t the minpercent value (double) (suggested 0.02 ";
	cerr << "- default INT_MAX = no embedded stopping criterion)\n";
	cerr
			<< "--freeze_lim                \t the freeze_lim value (int) (suggested 5)\n";
	cerr << "--tempfactor                \t the tempfactor value (double) \n";
	cerr << "--sizefactor                \t the sizefactor value (double) \n\n";
	cerr << "Copy n paste: -temperature 10 -cutoff 0.1 -minpercent 0.02";
	cerr << " --freeze_lim 5 -tempfactor 1 -sizefactor 1\n";
}

void Control::informationsReport() {
	ifstream info;
	char line[MAXSTRLEN];
	char *trailer;
	(*os) << "Current Machine: \n";

	info.open("/proc/cpuinfo", ios::in);
	if (info.is_open()) {
		while (!info.eof()) {
			info.getline(line, 100);
			if (info.eof())
				break;
			if (strstr(line, "processor") != NULL)
				(*os) << line << "\n";
			else if (strstr(line, "vendor_id") != NULL)
				(*os) << line << "\n";
			else if (strstr(line, "model name") != NULL)
				(*os) << line << "\n";
			else if (strstr(line, "cpu MHz") != NULL)
				(*os) << line << "\n";
			else if (strstr(line, "cache size") != NULL)
				(*os) << line << "\n";
			else if (strstr(line, "bogomips") != NULL)
				(*os) << line << "\n";
			else if (strstr(line, "EOF") != NULL)
				break;
			else
				continue;
		}
		info.close();
	} else {
		(*os)
				<< "Couldn`t find file /proc/cpuinfo for current machine type data\n";
	}
	(*os) << "\n";

	(*os) << "OS:\n";
	info.open("/etc/SuSE-release", ios::in);
	if (info.is_open()) {
		info.getline(line, 200);
		(*os) << line << "\n";
		info.close();
	} else {
		(*os) << "Couldn`t find SuSE version.\n";
	}
	(*os) << "\n";

	(*os) << "Compiler: " << endl;

	if (os == &cout)
		system("g++ -v");

	(*os) << "Flags:\n";
	info.open("Makefile", ios::in);
	if (info.is_open()) {
		while (!info.eof()) {
			info.getline(line, MAXSTRLEN);
			trailer = strstr(line, "CXXFLAGS =");
			if (trailer != NULL && strlen(trailer) > 0)
				(*os) << trailer << "\n";
		}
		info.close();
	} else {
		(*os) << "Couldn`t find Makefile \n";
	}

	(*os) << "\n";
	(*os) << "Algorithm:\n";
	(*os) << "GCP\n";
	(*os) << "\n";
	(*os) << "Parameter-settings:" << endl;
	(*os) << "maxTry: " << maxTry << "\n";
	(*os) << "timeLimit: " << timeLimit << "\n";
	(*os) << "seed: " << seed << "\n";
	(*os) << "\n";
	double initialization = getTime();
	(*os) << "Initialization Time: " << (initialization < 0 ? 0.0
			: initialization) << "\n";
	(*os) << "\n";
	(*os) << "Begin problem " << basename(is) << "\n";

}

char * Control::readline(istream &infile) {
	size_t length;
	size_t i;

	//char *result = f
	//cout << "entra/n";
	if (infile.getline(buffer, 250)) {
		length = strlen(buffer);
		//if (length > 0 && buffer[length-1] != 0)
		//(*os_log)<<"Buffer size in getline() isn't large enough\n"<<buffer;
		//cout << buffer << endl;
		for (i = 0; i < length; ++i)
			if (buffer[i] != ' ' || buffer[i] != '\t')
				return buffer;
	} else
		return 0;
	return readline(infile);
}

char * Control::match(char *source, char *prefix) {

	size_t prefixLength = strlen(prefix);

	if (strncmp(source, prefix, prefixLength) != 0)
		return 0;
	source += prefixLength;
	while ((*source) == ' ' || (*source) == '\t')
		source++;
	if (*source == ':' || *source == '=')
		source++;
	while ((*source) == ' ' || (*source) == '\t')
		source++;

	return source;
}

void Control::resetTime() {
	timer.resetTime();
}

double Control::getTime() {
	return timer.elapsedTime();
}

void Control::beginTry() {
	rg->set_seed(seed++);
	++nrTry;
	(*os) << "begin try " << nrTry << endl;
	resetTime();
	bestCost = MAX_FVALUE;
	best_k = INT_MAX;
	partial_Iter = 0;
}

void Control::endTry() {
	double time = getTime();
	(*os) << "exit iter " << partial_Iter;
	(*os) << " time ";
	os->flags(ios::fixed);
	(*os) << (time < 0 ? 0.0 : time) << endl;
	(*os) << "end try " << nrTry << endl;
	//cout<<bestCost<<"\t"<<seed-1<<endl;
}

void Control::setCurrentCost(fvalue cost, Color cl, unsigned long iter) {
	if (solver == 13) {
		double time = getTime();
		(*os) << "best " << (int) cost;
		(*os) << " col " << cl << " time ";
		os->flags(ios::fixed);
		(*os) << (time < 0 ? 0.0 : time);
		(*os) << " iter " << (int) iter + partial_Iter;
		(*os) << " par_iter " << (int) iter << endl;

	} else if (isTimeLeft()) {
		if (cl < best_k) {
			best_k = cl;
			//currentCost = cost;
			bestCost = cost;
			double time = getTime();
			(*os) << "best " << (int) bestCost;
			(*os) << " col " << best_k << " time ";
			os->flags(ios::fixed);
			(*os) << (time < 0 ? 0.0 : time);
			(*os) << " iter " << (int) iter + partial_Iter;
			(*os) << " par_iter " << (int) iter << endl;
		} else if (cl == best_k) {
			if (cost < bestCost) {
				bestCost = cost;
				double time = getTime();
				(*os) << "best " << (int) bestCost;
				(*os) << " col " << best_k << " time ";
				os->flags(ios::fixed);
				(*os) << (time < 0 ? 0.0 : time);
				(*os) << " iter " << (int) iter + partial_Iter;
				(*os) << " par_iter " << (int) iter << endl;
			}
		} else if (runner == 30) {

			double time = getTime();
			(*os) << "best " << (int) cost;
			(*os) << " col " << cl << " time ";
			os->flags(ios::fixed);
			(*os) << (time < 0 ? 0.0 : time);
			(*os) << " iter " << (int) iter + partial_Iter;
			(*os) << " par_iter " << (int) iter << endl;
		}
	}

}

void Control::setCurrentCostXRLF(fvalue cost, Color cl, unsigned long iter) {
	if (cl + K_I < best_k) {
		best_k = cl + K_I;
		//currentCost = cost;
		bestCost = cost;
		double time = getTime();
		(*os) << "best " << (int) bestCost;
		(*os) << " col " << best_k << " time ";
		os->flags(ios::fixed);
		(*os) << (time < 0 ? 0.0 : time);
		(*os) << " iter " << (int) iter + partial_Iter;
		(*os) << " par_iter " << (int) iter << endl;
	}
}

void Control::setHEACrossovers(unsigned int cross, unsigned int restart) {
	(*os) << "crossovers " << cross << " restart " << restart << endl;
}

