/* GCP version 6.0 Copyright (C) 2003 Marco Chiarandini  
 Technische Universitaet Darmstadt
 email: machud@intellektik.informatik.tu-darmstadt.de

 This file is part of GCP.

 GCP is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.
 
 Foobar is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with GCP; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "Explorer.h"

#include <algorithm>
#include <cmath>
#include <iostream>
using namespace std;

void ColNeighborhoodExplorer::MakeMove(CompleteColoring& coloring, const ColMove& rc) {
	if (rc.type == ONE_EX) {
		//one-ex
		Color old_col;
		Vertex node = rc.seq[0];
		old_col = coloring.col[node];
		coloring.col[node] = rc.seq[1];

		coloring.partitions[old_col].erase(node);
		coloring.partitions[rc.seq[1]].insert(node);

		p_sm->updateDeltaDataOneEx(coloring, rc.seq[0], old_col, rc.seq[1]);

		coloring.cost += rc.delta;
		if (coloring.cost != 0)
			coloring.feasible = false;
		else
			coloring.feasible = true;
	} else if (rc.type == CYCLE_EX) {
		//cycle
		unsigned int cycle_length = rc.seq.size();
		Vertex node;
		Color last_col, old_col, new_col;

		last_col = coloring.Class(rc.seq[0]);

		for (unsigned int i = 0; i < cycle_length - 1; i++) {
			if (((Vertex) rc.seq[i]) < p_in->Vertices()) {
				node = rc.seq[i];
				old_col = coloring.col[node];
				if (i == cycle_length - 2)
					new_col = last_col;
				else
					new_col = coloring.Class(rc.seq[i + 1]);

				coloring.col[node] = new_col;
				//cout <<"-->"<< node <<" "<< coloring.col[node]<<endl;
				coloring.partitions[old_col].erase(node);
				coloring.partitions[new_col].insert(node);
				//cout<<node<<" "<<old_col<<" "<<new_col<<endl;
				p_sm->updateDeltaDataOneEx(coloring, node, old_col, new_col);
			}
		}

		coloring.cost += rc.delta;

		if (coloring.cost != 0)
			coloring.feasible = false;
		else
			coloring.feasible = true;
	} else if (rc.type == PATH_EX) {
		//path
		unsigned int cycle_length = rc.seq.size();
		Vertex node;
		Color last_col, old_col, new_col;

		last_col = coloring.col[rc.seq[0]];

		for (unsigned int i = 0; i < cycle_length - 1; i++) {
			node = rc.seq[i];
			old_col = coloring.col[node];
			new_col = coloring.col[rc.seq[i + 1]];

			coloring.col[node] = new_col;

			coloring.partitions[old_col].erase(node);
			coloring.partitions[new_col].insert(node);

			p_sm->updateDeltaDataOneEx(coloring, node, old_col, new_col);
		}
		coloring.cost += rc.delta;

		if (coloring.cost != 0)
			coloring.feasible = false;
		else
			coloring.feasible = true;
	} else if (rc.type == SWAP) //pairsExchange
	{
		//one-ex
		Color old_col_i, old_col_j, old_col_p, old_col_q;
		Vertex i = rc.seq[0];
		Vertex j = rc.seq[1];
		Vertex p = rc.seq[2];
		Vertex q = rc.seq[3];

		old_col_i = coloring.col[i];
		old_col_j = coloring.col[j];
		old_col_p = coloring.col[p];
		old_col_q = coloring.col[q];

		coloring.col[i] = old_col_p;
		coloring.partitions[old_col_i].erase(i);
		coloring.partitions[old_col_p].insert(i);
		p_sm->updateDeltaDataOneEx(coloring, i, old_col_i, old_col_p);

		coloring.col[p] = old_col_i;
		coloring.partitions[old_col_p].erase(p);
		coloring.partitions[old_col_i].insert(p);
		p_sm->updateDeltaDataOneEx(coloring, p, old_col_p, old_col_i);

		coloring.col[j] = old_col_q;
		coloring.partitions[old_col_j].erase(j);
		coloring.partitions[old_col_q].insert(j);
		p_sm->updateDeltaDataOneEx(coloring, j, old_col_j, old_col_q);

		coloring.col[q] = old_col_j;
		coloring.partitions[old_col_q].erase(q);
		coloring.partitions[old_col_j].insert(q);
		p_sm->updateDeltaDataOneEx(coloring, q, old_col_q, old_col_j);

		coloring.cost += rc.delta;
		if (coloring.cost != 0)
			coloring.feasible = false;
		else
			coloring.feasible = true;
	} else {
		cerr << rc;
		cerr << "Wrong move in Explorer.cpp\n";
		exit(EXIT_FAILURE);
	}
}

//------------------------------------------------------------
//--------------------COLORING Neighboorhood------------------
//------------------------------------------------------------

// random move builder
void ColOneExExplorer::RandomMove(const CompleteColoring& coloring, ColMove& rc) {
	rc.seq.clear();
	rc.seq.resize(2);
	const Graph &g = *p_in; // g is an alias for the input Graph
	Vertex v;
	rc.type = ONE_EX;
	v = rc.seq[0] = rg->ran0N(g.Vertices() - 1);
	while (coloring.conflicts[rc.seq[0]] == 0) {
		rc.seq[0] = (rc.seq[0] + 1) % g.Vertices();
		if (((Vertex) rc.seq[0]) == v)
			break;
	}
	if (((Vertex) rc.seq[0]) == v && coloring.conflicts[rc.seq[0]] == 0) {
		std::cerr << "Could not find any neighbor" << endl;
		std::cerr << "But the Conflict list has size: "
				<< coloring.conflicts_set.size() << endl;
		cerr << coloring;
		assert(false);
	}

	do
		rc.seq[1] = (Color) rg->ran0N((coloring.CurrColors() - 1));
	while (((Color) rc.seq[1]) == coloring.col[rc.seq[0]]);
	rc.seq.push_back(coloring.col[rc.seq[0]]);
}

// update the state according to the move 
/*
 void ColOneExExplorer::MakeMove(CompleteColoring& coloring, const ColMove& rc)
 {
 //const Graph &g = *p_in; // g is an alias for the input Graph
 // for the identical move
 Color old_col;
 assert(rc.type == 1);
 Vertex node = rc.seq[0];

 old_col = coloring.col[node];

 coloring.col[node] = rc.seq[1];

 coloring.partitions[old_col].erase(node);
 coloring.partitions[rc.seq[1]].insert(node);

 p_sm->updateDeltaMatrixOneEx(coloring, rc.seq[0], old_col, rc.seq[1]);
 //  p_sm->updateImpGraphOneEx(coloring, rc.seq[0], old_col, rc.seq[1]);

 coloring.cost += rc.delta;

 if (coloring.cost != 0)
 coloring.feasible = false;
 else
 coloring.feasible = true;
 }
 */

// compute the next move in the exploration of the neighborhood
void ColOneExExplorer::NextMove(const CompleteColoring& coloring, ColMove& rc) {
	rc.seq.clear();
	rc.seq.resize(2);
	const Graph &g = *p_in; // g is an alias for the input Graph

	Color old_col = coloring.col[rc.seq[0]];

	rc.seq[1] = (rc.seq[1] + 1) % coloring.CurrColors();

	// try the next conflicting node skipping the trivial move (i.e. nc = oc)
	if (((Color) rc.seq[1]) == old_col) {
		do
			rc.seq[0] = (rc.seq[0] + 1) % g.Vertices();
		while (!coloring.InConflicts(rc.seq[0]));
		rc.seq[1] = (coloring.col[rc.seq[0]] + 1) % coloring.CurrColors();
	}
	rc.delta = coloring.Delta(rc.seq[0], rc.seq[1]);
	//rc.seq.push_back(coloring.col[rc.seq[0]]);
}

void ColOneExExplorer::BestMove(const CompleteColoring &st, ColMove& mv) {
	//const Graph &g = *p_in; // g is an alias for the input Graph
	fvalue min = FVALUE_MAX;
	VertexSet::iterator pt;
	vector<pair<Vertex, Color> > candidates;
	fvalue delta;

	mv.seq.clear();
	mv.seq.resize(2);
	mv.type = ONE_EX;
	for (pt = st.conflicts_set.begin(); pt != st.conflicts_set.end(); pt++) {
		for (Color j = 0; j < st.CurrColors(); j++) {
			if (st.col[*pt] != j) {
				delta = st.Delta(*pt, j);
				if (delta == min) {
					pair<Vertex, Color> initPair;
					initPair.first = *pt;
					initPair.second = j;
					candidates.push_back(initPair);
				} else if (delta < min) {
					candidates.clear();
					pair<Vertex, Color> initPair;
					initPair.first = *pt;
					initPair.second = j;
					candidates.push_back(initPair);
					min = delta;
				}
			}
		}
	}
	if (min < 0) {
		int rnd = rg->ran0N(candidates.size() - 1);
		mv.seq[0] = candidates[rnd].first;
		mv.seq[1] = candidates[rnd].second;

		mv.delta = min;
	} else
		mv.type = NOT_DEFINED;
	//mv.seq.push_back(st.col[mv.seq[0]]);


}

//For Tabu Col
void ColOneExExplorer::BestMove(const CompleteColoring& coloring, ColMove& rc,
		TabuList &t_l, unsigned long iter, fvalue best_so_far) {
	//const Graph &g = *p_in; // g is an alias for the input Graph
	fvalue min = FVALUE_MAX;
	VertexSet::iterator pt;
	vector<pair<Vertex, Color> > candidates;
	fvalue delta;

	rc.seq.clear();
	rc.seq.resize(2);
	rc.type = ONE_EX;
	for (pt = coloring.conflicts_set.begin(); pt
			!= coloring.conflicts_set.end(); pt++) {
		for (Color j = 0; j < coloring.CurrColors(); j++) {
			if (coloring.col[*pt] != j) {
				delta = coloring.Delta(*pt, j);
				if (delta == min) {
					if ((!t_l.isTabu(*pt, j, iter)) || (coloring.cost
							+ delta < best_so_far)) {
						pair<Vertex, Color> initPair;
						initPair.first = *pt;
						initPair.second = j;
						candidates.push_back(initPair);
					}
				} else if (delta < min) {
					if ((!t_l.isTabu(*pt, j, iter)) || (coloring.cost
							+ delta < best_so_far)) {
						candidates.clear();
						pair<Vertex, Color> initPair;
						initPair.first = *pt;
						initPair.second = j;
						candidates.push_back(initPair);
						min = delta;
					}
				}
			}
		}
	}

	if (candidates.size() > 0) {
		int rnd = rg->ran0N(candidates.size() - 1);
		rc.seq[0] = candidates[rnd].first;
		rc.seq[1] = candidates[rnd].second;
		rc.delta = min;
		//rc.seq.push_back(coloring.col[rc.seq[0]]);
	} else { // select one at random waiting for some tabu to expire
		size_t size = rg->ran0N(coloring.conflicts_set.size() - 1);
		pt = coloring.conflicts_set.begin();
		size_t i = 0;
		while (i < size) {
			i++;
			pt++;
		}
		assert(*pt<p_in->Vertices());
		rc.seq[0] = *pt;
		do {
			rc.seq[1] = rg->ran0N(coloring.CurrColors() - 1);
		} while (((Color) rc.seq[1]) == coloring.col[rc.seq[0]]);
		rc.delta = coloring.Delta(rc.seq[0], rc.seq[1]);
	}
}

//For TabuHash:
void ColOneExExplorer::BestMove(const CompleteColoring& coloring, ColMove& rc,
		HashForTabu *ht, fvalue best_so_far) {
	//const Graph &g = *p_in; // g is an alias for the input Graph
	fvalue min = FVALUE_MAX;
	VertexSet::iterator pt;
	vector<pair<Vertex, Color> > candidates;
	fvalue delta;
	Longint hash_val;

	rc.seq.clear();
	rc.seq.resize(2);
	rc.type = ONE_EX;
	for (pt = coloring.conflicts_set.begin(); pt
			!= coloring.conflicts_set.end(); pt++) {
		for (Color j = 0; j < coloring.CurrColors(); j++) {
			if (coloring.col[*pt] != j) {
				delta = coloring.Delta(*pt, j);
				if (delta == min) {
					rc.seq[0] = *pt;
					rc.seq[1] = j;
					hash_val = ht->compute_hash_value_incrementally(coloring,
							rc);
					if ((!(ht->is_in_hash_table(hash_val)))
							|| (coloring.cost + delta < best_so_far)) {
						pair<Vertex, Color> initPair;
						initPair.first = *pt;
						initPair.second = j;
						candidates.push_back(initPair);
					}
				} else if (delta < min) {
					rc.seq[0] = *pt;
					rc.seq[1] = j;
					hash_val = ht->compute_hash_value_incrementally(coloring,
							rc);
					if ((!(ht->is_in_hash_table(hash_val)))
							|| (coloring.cost + delta < best_so_far)) {
						candidates.clear();
						pair<Vertex, Color> initPair;
						initPair.first = *pt;
						initPair.second = j;
						candidates.push_back(initPair);
						min = delta;
					}
				}
			}
		}
	}

	if (candidates.size() > 0) {
		int rnd = rg->ran0N(candidates.size() - 1);
		rc.seq[0] = candidates[rnd].first;
		rc.seq[1] = candidates[rnd].second;
		rc.delta = min;
		//rc.seq.push_back(coloring.col[rc.seq[0]]);
	} else {
		 size_t size = rg->ran0N(coloring.conflicts_set.size() - 1);
		pt = coloring.conflicts_set.begin();
		size_t i = 0;
		while (i < size) {
			i++;
			pt++;
		}
		rc.seq[0] = *pt;
		do {
			rc.seq[1] = rg->ran0N(coloring.CurrColors() - 1);
		} while (((Color) rc.seq[1]) == coloring.col[rc.seq[0]]);
		rc.delta = coloring.Delta(rc.seq[0], rc.seq[1]);
	}
}

void ColOneExExplorer::BestMove(const CompleteColoring& coloring, ColMove& rc,
		TabuList &t_l, vector<Vertex> dont_look, unsigned long iter,
		fvalue best_so_far) {
	//const Graph &g = *p_in; // g is an alias for the input Graph
	fvalue min = FVALUE_MAX;
	VertexSet::iterator pt;
	vector<pair<Vertex, Color> > candidates;
	fvalue delta;

	//cout<<"--> "<<coloring.conflicts_set.size()<<endl;
	rc.seq.clear();
	rc.seq.resize(2);
	rc.type = ONE_EX;
	for (pt = coloring.conflicts_set.begin(); pt
			!= coloring.conflicts_set.end(); pt++) {

		if (dont_look[(*pt)] != 1)
			continue;
		for (Color j = 0; j < coloring.CurrColors(); j++) {
			if (coloring.col[*pt] != j) {
				delta = coloring.Delta(*pt, j);
				if (delta == min) {
					if ((!t_l.isTabu(*pt, j, iter)) || (coloring.cost
							+ delta < best_so_far)) {
						pair<Vertex, Color> initPair;
						initPair.first = *pt;
						initPair.second = j;
						candidates.push_back(initPair);
					}
				} else if (delta < min) {
					if ((!t_l.isTabu(*pt, j, iter)) || (coloring.cost
							+ delta < best_so_far)) {
						candidates.clear();
						pair<Vertex, Color> initPair;
						initPair.first = *pt;
						initPair.second = j;
						candidates.push_back(initPair);
						min = delta;
					}
				}
			}
		}
	}

	if (candidates.size() > 0) {
		int rnd = rg->ran0N(candidates.size() - 1);
		rc.seq[0] = candidates[rnd].first;
		rc.seq[1] = candidates[rnd].second;
		rc.delta = min;
		//rc.seq.push_back(coloring.col[rc.seq[0]]);
	} else {
		do {
			rc.seq[0] = rg->ran0N(p_in->Vertices() - 1);
			rc.seq[1] = rg->ran0N(coloring.CurrColors() - 1);
		} while (((Color) rc.seq[1]) == coloring.col[rc.seq[0]]);
		rc.delta = coloring.Delta(rc.seq[0], rc.seq[1]);
	}
}

/*
 void ColOneExExplorer::BestMove(const CompleteColoring& coloring, ColMove& rc,
 TabuList &t_l, unsigned long iter,
 fvalue best_so_far)
 {    
 //const Graph &g = *p_in; // g is an alias for the input Graph
 fvalue min = FVALUE_MAX;
 //VertexSet::iterator pt;
 vector < pair<Vertex, Color> > candidates;
 fvalue delta;

 rc.seq.clear();
 rc.seq.resize(2);
 rc.type= 1;
 for (Vertex i = 0; i < p_in->Vertices(); i++)
 {
 for (Color j=0 ; j < coloring.k ; j++) {
 if (coloring.col[i]!=j)
 {
 delta = coloring.Delta(i,j);
 if (delta == min) {
 if ((!t_l.isTabu(i,j,iter))
 || (coloring.cost + delta < best_so_far))
 {
 pair<Vertex, Color> initPair;
 initPair.first = i ;
 initPair.second = j ;
 candidates.push_back(initPair);
 }
 }
 else if (delta < min)  {
 if ((!t_l.isTabu(i,j,iter))
 || (coloring.cost + delta < best_so_far))
 {
 candidates.clear();
 pair<Vertex, Color> initPair;
 initPair.first = i ;
 initPair.second = j ;
 candidates.push_back(initPair);
 min = delta;
 }
 }
 }
 }
 }
 if (candidates.size()>0)
 {
 int rnd = rg->ran0N(candidates.size()-1);
 rc.seq[0] = candidates[rnd].first;
 rc.seq[1] = candidates[rnd].second;
 rc.delta = min;
 //rc.seq.push_back(coloring.col[rc.seq[0]]);
 }
 else
 {
 do {
 rc.seq[0]=rg->ran0N(p_in->Vertices()-1);
 rc.seq[1]=rg->ran0N(coloring.k-1);
 } while (rc.seq[1]==coloring.col[rc.seq[0]]);
 rc.delta = coloring.Delta(rc.seq[0],rc.seq[1]);
 }

 }*/

//without Aspiration Criteria!!!
void ColOneExExplorer::BestMove(const CompleteColoring& coloring, ColMove& rc,
		TabuList &t_l, long unsigned iter) {
	//const Graph &g = *p_in; // g is an alias for the input Graph
	fvalue min = FVALUE_MAX;
	VertexSet::iterator pt;
	vector<pair<Vertex, Color> > candidates;
	fvalue delta;

	rc.seq.clear();
	rc.seq.resize(2);
	rc.type = ONE_EX;
	for (pt = coloring.conflicts_set.begin(); pt
			!= coloring.conflicts_set.end(); pt++) {
		for (Color j = 0; j < coloring.CurrColors(); j++) {
			if (coloring.col[*pt] != j) {
				if (!t_l.isTabu(*pt, j, iter)) {
					delta = coloring.Delta(*pt, j);
					if (delta == min) {
						pair<Vertex, Color> initPair;
						initPair.first = *pt;
						initPair.second = j;
						candidates.push_back(initPair);
					} else if (delta < min) {
						candidates.clear();
						pair<Vertex, Color> initPair;
						initPair.first = *pt;
						initPair.second = j;
						candidates.push_back(initPair);
						min = delta;
					}
				}
			}
		}
	}
	if (candidates.size() > 0) {
		int rnd = rg->ran0N(candidates.size() - 1);
		rc.seq[0] = candidates[rnd].first;
		rc.seq[1] = candidates[rnd].second;
		rc.delta = min;
		//rc.seq.push_back(coloring.col[rc.seq[0]]);
	} else {
		do {
			rc.seq[0] = rg->ran0N(p_in->Vertices() - 1);
			rc.seq[1] = rg->ran0N(coloring.CurrColors() - 1);
		} while ( ((Color) rc.seq[1]) == coloring.col[rc.seq[0]]);
		rc.delta = coloring.Delta(rc.seq[0], rc.seq[1]);
	}
}

//for Chains!
void ColOneExExplorer::BestMove(const CompleteColoring& coloring, Color prev_class,
		ColMove& rc, TabuList &t_l, unsigned long iter, fvalue best_so_far) {
	//const Graph &g = *p_in; // g is an alias for the input Graph
	fvalue min = FVALUE_MAX;
	VertexSet::iterator pt;
	vector<pair<Vertex, Color> > candidates;
	fvalue delta;

	rc.seq.clear();
	rc.seq.resize(2);
	rc.type = ONE_EX;
	for (pt = coloring.partitions[prev_class].begin(); pt
			!= coloring.partitions[prev_class].end(); pt++) {
		if (coloring.conflicts[*pt] > 0) {
			for (Color j = 0; j < coloring.CurrColors(); j++) {
				if (coloring.col[*pt] != j) {
					delta = coloring.Delta(*pt, j);
					if (delta == min) {
						if ((!t_l.isTabu(*pt, j, iter)) || (coloring.cost
								+ delta < best_so_far)) {
							pair<Vertex, Color> initPair;
							initPair.first = *pt;
							initPair.second = j;
							candidates.push_back(initPair);
						}
					} else if (delta < min) {
						if ((!t_l.isTabu(*pt, j, iter)) || (coloring.cost
								+ delta < best_so_far)) {
							candidates.clear();
							pair<Vertex, Color> initPair;
							initPair.first = *pt;
							initPair.second = j;
							candidates.push_back(initPair);
							min = delta;
						}
					}
				}
			}
		}
	}

	if (candidates.size() > 0) {
		int rnd = rg->ran0N(candidates.size() - 1);
		rc.seq[0] = candidates[rnd].first;
		rc.seq[1] = candidates[rnd].second;
		rc.delta = min;
		//rc.seq.push_back(coloring.col[rc.seq[0]]);
	} else {
		do {
			rc.seq[0] = rg->ran0N(p_in->Vertices() - 1);
			rc.seq[1] = rg->ran0N(coloring.CurrColors() - 1);
		} while ( ((Color) rc.seq[1]) == coloring.col[rc.seq[0]]);
		rc.delta = coloring.Delta(rc.seq[0], rc.seq[1]);
	}
}

//for GLS
/*
 void ColOneExExplorer::BestMove(const CompleteColoring &st, ColMove& mv, float &pen_delta)

 {
 //const Graph &g = *p_in; // g is an alias for the input Graph
 float min = FVALUE_MAX;
 VertexSet::iterator pt;
 vector < pair<Vertex, Color> > candidates;

 mv.seq.clear();
 mv.seq.resize(2);
 mv.type= 1;
 for (pt = st.conflicts_set.begin();
 pt != st.conflicts_set.end(); pt++)
 {
 for (Color j=0 ; j < st.CurrColors() ; j++) {
 if (st.col[*pt]!=j)
 {
 pen_delta = st.pen_delta[*pt][j];
 if (pen_delta == min) {
 pair<Vertex, Color> initPair;
 initPair.first = *pt ;
 initPair.second = j ;
 candidates.push_back(initPair);
 }
 else if (pen_delta < min)  {
 candidates.clear();
 pair<Vertex, Color> initPair;
 initPair.first = *pt ;
 initPair.second = j ;
 candidates.push_back(initPair);
 min = pen_delta;
 }
 }
 }
 }
 if (candidates.size()>0)
 {
 int rnd = rg->ran0N(candidates.size()-1);
 mv.seq[0] = candidates[rnd].first;
 mv.seq[1] = candidates[rnd].second;
 pen_delta = min;
 mv.delta = st.Delta(mv.seq[0],mv.seq[1]);
 }
 else
 mv.type = 0;
 //mv.seq.push_back(st.col[mv.seq[0]]);
 }
 */

void ColOneExExplorer::BestMove(const CompleteColoring &st, ColMove& mv,
		float &pen_delta)

{
	//const Graph &g = *p_in; // g is an alias for the input Graph
	float min = FVALUE_MAX;
	VertexSet::iterator pt;
	vector<pair<Vertex, Color> > candidates;

	mv.seq.clear();
	mv.seq.resize(2);
	mv.type = ONE_EX;
	for (pt = st.conflicts_set.begin(); pt != st.conflicts_set.end(); pt++) {
		for (Color j = 0; j < st.CurrColors(); j++) {
			if (st.col[*pt] != j) {
				pen_delta = st.PenDelta(*pt, j);//st.pen_delta[*pt][j];
				if (pen_delta == min) {
					pair<Vertex, Color> initPair;
					initPair.first = *pt;
					initPair.second = j;
					candidates.push_back(initPair);
				} else if (pen_delta < min) {
					candidates.clear();
					pair<Vertex, Color> initPair;
					initPair.first = *pt;
					initPair.second = j;
					candidates.push_back(initPair);
					min = pen_delta;
				}
			}
		}
	}
	if (candidates.size() > 0) {
		int rnd = rg->ran0N(candidates.size() - 1);
		mv.seq[0] = candidates[rnd].first;
		mv.seq[1] = candidates[rnd].second;
		pen_delta = min;
		mv.delta = st.Delta(mv.seq[0], mv.seq[1]);
	} else
		mv.type = NOT_DEFINED;
	//mv.seq.push_back(st.col[mv.seq[0]]);
}

//------------------------------------------------------------
//------------------------------------------------------------
//------------------------------------------------------------

