/* GCP version 6.0 Copyright (C) 2003 Marco Chiarandini  
 Technische Universitaet Darmstadt
 email: machud@intellektik.informatik.tu-darmstadt.de

 This file is part of GCP.

 GCP is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.
 
 Foobar is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with GCP; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef EXPLORER_H
#define EXPLORER_H
#include <vector>
#include "Utilities.h"
#include "Problem.h"
#include "Control.h"
#include "Random.h"
#include "Solution.h"
#include "Manager.h"
#include "Move.h"
#include "Hash.h"

using namespace std;

class ColNeighborhoodExplorer {
public:
	ColNeighborhoodExplorer(ColoringManager* sm, Random *rnd, Graph* pin = NULL) {
		p_sm = sm;
		p_in = pin;
		rg = rnd;
	}
	ColNeighborhoodExplorer(const ColNeighborhoodExplorer* ne) {
		p_sm = ne->p_sm;
		p_in = ne->p_in;
		rg = ne->rg;
	}
	/** Virtual destructor. */
	virtual ~ColNeighborhoodExplorer() {
	}
	;

	virtual void initAuxiliar(const Color K) {
	}
	;
	//void Print(ostream& os = cout) const;
	//virtual void RandomMove(const CompleteColoring&, ColMove&) {};
	virtual void MakeMove(Coloring&, const ColMove&) {
	}
	;
	virtual void MakeMove(CompleteColoring&, const ColMove&);
	virtual void MakeMove(PartialColoring&, const ColMove&);
	//virtual void MakeMove2(Coloring&, const ColMove&);
	//virtual void NextMove(const Coloring&, ColMove&) = 0;
	virtual void NextMove(const CompleteColoring&, ColMove&,
			long unsigned current_iter) {
	}
	;

	void RandomMove(const CompleteColoring&, ColMove&);
	void RandomMove(const PartialColoring&, ColMove&);

	virtual void BestMove(const CompleteColoring &st, ColMove& mv) {
	}
	;
	virtual void BestMove(const CompleteColoring& coloring, ColMove& rc,
			TabuList& t_l, unsigned long iter, fvalue best_so_far) {
	}
	;
	virtual void BestMove(const PartialColoring& coloring, ColMove& rc,
			TabuList& t_l, unsigned long iter, fvalue best_so_far) {
	}
	;

	//for RLFtabu
	virtual void BestMove(const CompleteColoring& coloring, ColMove& rc,
			HashForTabu *ht, fvalue best_so_far) {
	}
	;

	//for RLFtabu
	virtual void BestMove(const CompleteColoring& coloring, ColMove& rc,
			TabuList &t_l, vector<Vertex> dont_look, unsigned long iter,
			fvalue best_so_far) {
	}
	;

	//for Double runner
	virtual void BestMove(const CompleteColoring& coloring, ColMove& rc,
			TabuList& t_l, unsigned long iter, const ColMove &prec_move,
			fvalue best_so_far) {
	}
	;
	//without Aspiration Criterion
	virtual void BestMove(const CompleteColoring& coloring, ColMove& rc,
			TabuList& t_l, unsigned long iter) {
	}
	;
	virtual void BestMove(const CompleteColoring &st, ColMove& mv,
			float &pen_delta) {
	}
	;

protected:
	ColoringManager* p_sm; /**< A pointer to the attached
	 state manager. */
	Graph* p_in; /**< A pointer to the input object. */
	Random *rg;
};

/*------------------------------------------------------------
 ---------------- ONE-EX ----------------------------------------
 ------------------------------------------------------------*/

class ColOneExExplorer: public ColNeighborhoodExplorer {
public:
	ColOneExExplorer(ColoringManager* sm, Random *rnd, Graph* pin = NULL) :
		ColNeighborhoodExplorer(sm, rnd, pin) {
	}
	;
	ColOneExExplorer(const ColNeighborhoodExplorer* ne) :
		ColNeighborhoodExplorer(ne) {
	}
	;
	/** Virtual destructor. */
	virtual ~ColOneExExplorer() {
	}
	;
	void SetParameters(int depth, int labels);

	void RandomMove(const CompleteColoring&, ColMove&);

	//void MakeMove(Coloring&, const ColMove&);
	void NextMove(const CompleteColoring&, ColMove&);

	void BestMove(const CompleteColoring &st, ColMove& mv);
	void BestMove(const CompleteColoring& coloring, ColMove& rc, TabuList& t_l,
			unsigned long iter, fvalue best_so_far);
	// For TabuHash:
	void BestMove(const CompleteColoring& coloring, ColMove& rc,
			HashForTabu *ht, fvalue best_so_far);
	//without Aspiration Criterion
	void BestMove(const CompleteColoring& coloring, ColMove& rc, TabuList &t_l,
			unsigned long iter);
	//for Chains
	void BestMove(const CompleteColoring& coloring, Color prev_class,
			ColMove& rc, TabuList &t_l, unsigned long iter, fvalue best_so_far);
	void BestMove(const CompleteColoring &st, ColMove& mv, float &pen_delta);
	//for RLFtabu
	void BestMove(const CompleteColoring& coloring, ColMove& rc, TabuList &t_l,
			vector<Vertex> dont_look, unsigned long iter, fvalue best_so_far);

};

class ColISwapExplorer: public ColNeighborhoodExplorer {
public:
	ColISwapExplorer(ColoringManager* sm, Random *rnd, Graph* pin = NULL) :
		ColNeighborhoodExplorer(sm, rnd, pin) {
	}
	;
	ColISwapExplorer(const ColNeighborhoodExplorer* ne) :
		ColNeighborhoodExplorer(ne) {
	}
	;
	/** Virtual destructor. */
	virtual ~ColISwapExplorer() {
	}
	;
	void SetParameters(int depth, int labels);

	void RandomMove(const PartialColoring&, ColMove&);

	// For PartialTabuCol
	void BestMove(const PartialColoring& coloring, ColMove& rc, TabuList& t_l,
			unsigned long iter, fvalue best_so_far);
};

class KempeChains: public ColNeighborhoodExplorer {
public:
	KempeChains(ColoringManager* sm, Random *rnd, Graph* pin);
	KempeChains(const ColNeighborhoodExplorer* ne);
	/** Virtual destructor. */
	virtual ~KempeChains() {
	}
	void initAuxiliar(const Color);
	void MakeMove(CompleteColoring&, const ColMove&);
	//void RandomMove(const CompleteColoring&, ColMove&) {};
	//void MakeMove(CompleteColoring&, const ColMove&);
	void
			NextMove(const CompleteColoring&, ColMove&,
					long unsigned current_iter);

	void BestMove(const CompleteColoring &st, ColMove& mv);
	void BestMove(const CompleteColoring& coloring, ColMove& rc, TabuList& t_l,
			unsigned long iter);
	void BestMove(const CompleteColoring& coloring, ColMove& rc, TabuList& t_l,
			unsigned long iter, fvalue best_so_far);
	//without Aspiration Criterion
	//void BestMove(const CompleteColoring& coloring, ColMove& rc,
	//	   TabuList &t_l, unsigned long iter);
	//for Chains
	//  void BestMove(const CompleteColoring& coloring, Color prev_class,
	//			ColMove& rc,
	//			TabuList &t_l, unsigned long iter,
	//			fvalue best_so_far);

protected:

	int bfscomponent(const CompleteColoring &ass, Vertex v, Color clr1,
			Color clr2, int compnum);
	int bfscomponentWithTabu(const CompleteColoring &ass, Vertex v, Color clr1,
			Color clr2, int compnum);
	int CheckComponents(const CompleteColoring &ass, Color clr1, Color clr2,
			fvalue &delta);
	int CheckComponentsWithTabu(TabuList& t_l, const CompleteColoring &ass,
			Color clr1, Color clr2, fvalue &delta, unsigned long iter);
	int CheckComponentsSingle(const CompleteColoring &ass, Color clr1,
			Color clr2, fvalue &delta);
	void recordMembers(const CompleteColoring &ass, Color clr1, Color clr2,
			int component, vector<int> &members);
	bool isTabu(TabuList &tl, const CompleteColoring &ass, Color clr1,
			Color clr2, int component, unsigned long par_iter);
	fvalue deltaBestChains(const CompleteColoring &ass, Color clr1, Color clr2);

	bool recentlyFullKempe(Color c, Color d) {
		return (full_kempe[c][d] > last_mod[c] && full_kempe[c][d]
				> last_mod[d]);
	}

	vector<long int> last_mod;
	vector<vector<long int> > full_kempe;

	vector<Vertex> flagset;

	vector<Vertex> components;
	vector<Vertex> best_components;
	unsigned long iter;

};

class ColOneExMinConfExplorer: public ColNeighborhoodExplorer {
public:
	ColOneExMinConfExplorer(ColoringManager* sm, Random *rnd, Graph* pin = NULL) :
		ColNeighborhoodExplorer(sm, rnd, pin) {
	}
	;
	ColOneExMinConfExplorer(const ColNeighborhoodExplorer* ne) :
		ColNeighborhoodExplorer(ne) {
	}
	;
	/** Virtual destructor. */
	virtual ~ColOneExMinConfExplorer() {
	}

	void BestMove(const CompleteColoring& coloring, ColMove& rc, TabuList& t_l,
			unsigned long iter, fvalue best_so_far);

};

class ColOneExNoveltyExplorer: public ColNeighborhoodExplorer {
public:
	ColOneExNoveltyExplorer(ColoringManager* sm, Random *rnd, Graph* pin = NULL) :
		ColNeighborhoodExplorer(sm, rnd, pin) {
	}
	;
	ColOneExNoveltyExplorer(const ColNeighborhoodExplorer* ne) :
		ColNeighborhoodExplorer(ne) {
	}
	;
	/** Virtual destructor. */
	virtual ~ColOneExNoveltyExplorer() {
	}

	void BestMove(const CompleteColoring& coloring, ColMove& rc, TabuList& t_l,
			unsigned long iter, fvalue best_so_far);

	//void BestMove(const CompleteColoring& coloring, ColMove& rc);

};

/*------------------------------------------------------------
 ---------------- CYCLE_HASH-----------------------------------
 ------------------------------------------------------------*/

/* -- 2^16 = 65536, the maximum amount of memory allocated on 4 levels is 31*65536~2MB */
#define INITALLOC	262144 //65536
#define MEMLEVELS	6

class _DPState {
public:
	// _DPState();
	//  ~_DPState();
	int i, j; /* -- The nodes i and j in the state */
	fvalue fCost; /* -- Cost associated with state */
	unsigned long Set[4];/* -- This long integer stores the sets through which the cycle goes */
	short Stage; /* -- The stage of the DP */
	_DPState *pred; /* -- The predecessor in the optimal path */
	_DPState *pNextHash;/* -- Pointer to the next state structure in hash table */
};

/* -- The array that is used to store memory for states */
class _MemStruct {
public:
	_MemStruct();
	~_MemStruct();
	_DPState* DPStateAlloc();

	_DPState *LevelMem[MEMLEVELS];
	_DPState *CurrLevelMem;
	int nMemLevel, nMemUsage, nMemAvailable;
};

/* -- 997 is just a large prime */
#define HASHTABLESIZE	9967

class _HashTable {
private:
	_DPState *HashTable[HASHTABLESIZE];
public:
	_HashTable();
	~_HashTable() {
	}
	;
	void HashInsert(_DPState *pState);
	_DPState *HashFind(int i, unsigned long Set[], short stage);
};

//------------------------------------------------------------

class ColVLSNImprovExplorer: public virtual ColNeighborhoodExplorer {
public:
	ColVLSNImprovExplorer(ColoringManager* sm, Random *rnd, int state_limit,
			Graph* pin);
	ColVLSNImprovExplorer(const ColNeighborhoodExplorer* ne);
	/** Virtual destructor. */
	virtual ~ColVLSNImprovExplorer() {
	}

	void SetParameters(int state_limit);

	virtual void BestMove(const CompleteColoring &st, ColMove& mv);
	/*void BestMove(const CompleteColoring &st, ColMove& mv,
	 TabuList &tl, unsigned long iter);*/

protected:
	int STATE_LIMIT;
};

class ColVLSNExplorer: public virtual ColVLSNImprovExplorer,
		public virtual ColNeighborhoodExplorer {
public:
	ColVLSNExplorer(ColoringManager* sm, Random *rnd, int state_limit,
			Graph* pin) :
				ColNeighborhoodExplorer(sm, rnd, pin),
		ColVLSNImprovExplorer(sm, rnd, state_limit, pin) {

	}
	;
	virtual ~ColVLSNExplorer() {
	}
	void BestMove(const CompleteColoring &st, ColMove& mv);
	void BestMove(const CompleteColoring &st, ColMove& mv, TabuList& t_l,
			unsigned long iter, fvalue best_so_far);
};

class ColReducedMinimumHashCycleOrPathExplorer: public virtual ColVLSNImprovExplorer,
		public virtual ColNeighborhoodExplorer {
public:
	ColReducedMinimumHashCycleOrPathExplorer(ColoringManager* sm, Random *rnd,
			int state_limit, Graph* pin = NULL) :
		ColNeighborhoodExplorer(sm, rnd, pin), ColVLSNImprovExplorer(sm, rnd,
				state_limit, pin) {
	}
	;
	virtual ~ColReducedMinimumHashCycleOrPathExplorer() {
	}

	//void BestMove(const CompleteColoring &st, ColMove& mv);
	void BestMove(const CompleteColoring &st, ColMove& mv, TabuList &tl,
			unsigned long iter, fvalue best_so_far);
};

class ColRevRedMinOneExOrNegHashCycleOrPathExplorerII: public ColVLSNImprovExplorer,
		public virtual ColNeighborhoodExplorer {
public:
	ColRevRedMinOneExOrNegHashCycleOrPathExplorerII(ColoringManager* sm,
			Random *rnd, int state_limit, Graph* pin = NULL) :
		ColNeighborhoodExplorer(sm, rnd, pin), ColVLSNImprovExplorer(sm, rnd,
				state_limit, pin) {
	}
	;
	virtual ~ColRevRedMinOneExOrNegHashCycleOrPathExplorerII() {
	}

	void BestMove(const CompleteColoring &st, ColMove& mv);
	void BestMove(const CompleteColoring &st, ColMove& mv, TabuList &tl,
			unsigned long iter);
};

//------------------------------------------------------------
//------------------------------------------------------------
//------------------------------------------------------------

#endif

