/* GCP version 3.5 Copyright (C) 2003 Marco Chiarandini  
 Technische Universitaet Darmstadt
 email: marco@imada.sdu.dk

 This file is part of GCP.

 GCP is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.
 
 Foobar is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with GCP; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "Runners.h"
#include "TabuList.h"

void IteratedLocalSearch::Run(CompleteColoring &initial_ass) {
	InitializeRun();
	CompleteColoring current_state(initial_ass);

	long int max_idle_iterations_ls = LONG_MAX;
	Vertex moved;

	Vertex (ColoringManager::*pert)(CompleteColoring &ass, Color to_remove,
			TabuList &tl, long int iter);
	switch (p_ctrl->pert) {
	case 0:
		pert = &ColoringManager::perturbate_seq;
		break;
	case 1:
		pert = &ColoringManager::perturbate_dsat;
		break;
	case 2:
		pert = &ColoringManager::perturbate_adj;
		break;

	default:
		cerr << "wrong perturbation selected\n";
		exit(1);
	}

	Color col_to_remove = (Color) ((double) p_ctrl->gamma
			* current_state.CurrColors());
	TabuList tl(p_in->Vertices(), current_state.CurrColors(), rg, p_ctrl);
	max_idle_iterations_ls = p_in->Vertices() * current_state.CurrColors();

	LocalSearch(current_state, tl, max_idle_iterations_ls);
	if (current_state.cost < initial_ass.cost) {
		initial_ass = current_state;
		p_ctrl->setCurrentCost(current_state.cost, current_state.CurrColors(),
				number_of_iterations);
		idle_iterations = 0;
	}

	while ((current_state.cost != 0) && !(MaxIterationExpired())
			&& (idle_iterations < max_idle_iteration) && p_ctrl->isTimeLeft()) {
		CompleteColoring previous_local_optimum(current_state);
		//Perturbe
		if (col_to_remove == 2) {
			if (rg->next() < 0.5)
				moved = (p_sm->*pert)(current_state, 1, tl,
						number_of_iterations);
			else
				moved = (p_sm->*pert)(current_state, 3, tl,
						number_of_iterations);
		} else
			moved = (p_sm->*pert)(current_state, col_to_remove, tl,
					number_of_iterations);
#ifdef TEST_RUNNER
		cout << "Perturbation: color removed: " << col_to_remove << " "
		<< max_idle_iterations_ls << " " << current_state.cost;
#endif
		//apply local search
		max_idle_iterations_ls = moved * 5 * current_state.CurrColors();

		LocalSearch(current_state, tl, max_idle_iterations_ls);
#ifdef TEST_RUNNER
		cout << " " << current_state.cost << " " << p_sm->distanceLap(
				current_state, previous_local_optimum) << endl;
#endif
		if (current_state.cost < initial_ass.cost) {
			initial_ass = current_state;
			p_ctrl->setCurrentCost(current_state.cost,
					current_state.CurrColors(), number_of_iterations);
			idle_iterations = 0;
		}
		if (current_state.cost == 0)
			break;

		//Acceptance Criterion
		if (rg->next() < p_ctrl->ilsprob)
			current_state = initial_ass;

	}

	TerminateRun();
}


void ILS_adj::Run(CompleteColoring &initial_ass) {
	InitializeRun();
	CompleteColoring current_state(initial_ass);

	long int max_idle_iterations_ls = LONG_MAX;
	Vertex moved;

	Vertex (ColoringManager::*pert)(CompleteColoring &ass, Color to_remove,
			TabuList &tl, long int iter);
	pert = &ColoringManager::perturbate_adj;

	Color levels = (Color) p_ctrl->gamma;
	TabuList tl(p_in->Vertices(), current_state.CurrColors(), rg, p_ctrl);
	max_idle_iterations_ls = p_in->Vertices() * current_state.CurrColors();

	LocalSearch(current_state, tl, max_idle_iterations_ls);
	if (current_state.cost < initial_ass.cost) {
		initial_ass = current_state;
		p_ctrl->setCurrentCost(current_state.cost, current_state.CurrColors(),
				number_of_iterations);
		idle_iterations = 0;
	}

	while ((current_state.cost != 0) && !(MaxIterationExpired())
			&& (idle_iterations < max_idle_iteration) && p_ctrl->isTimeLeft()) {
		moved = (p_sm->*pert)(current_state, levels, tl, number_of_iterations);
		cout << moved << endl;
		//apply local search
		max_idle_iterations_ls = moved * 5 * current_state.CurrColors();

		LocalSearch(current_state, tl, max_idle_iterations_ls);
		if (current_state.cost < initial_ass.cost) {
			initial_ass = current_state;
			p_ctrl->setCurrentCost(current_state.cost,
					current_state.CurrColors(), number_of_iterations);
			idle_iterations = 0;
		}
		if (current_state.cost == 0)
			break;

		//Acceptance Criterion
		if (rg->next() < p_ctrl->ilsprob)
			current_state = initial_ass;

	}

	TerminateRun();
}

void IteratedLocalSearch::LocalSearch(CompleteColoring &initial_ass, TabuList &tl,
		Longint max_idle_iterations_ls) {
	CompleteColoring current_state(initial_ass);
	Longint idle_iterations_ls = 0;
	ColMove mv;
	while ((current_state.cost != 0) && !(MaxIterationExpired())
			&& (idle_iterations_ls < max_idle_iterations_ls)
			&& p_ctrl->isTimeLeft()) {
		UpdateIterationCounter();
		p_ne->BestMove(current_state, mv, tl, number_of_iterations,
				initial_ass.cost);
		tl.UpdateTabuList(current_state, mv, number_of_iterations);
		p_ne->MakeMove(current_state, mv);

		if (current_state.cost != 0)
			current_state.feasible = false;
		else
			current_state.feasible = true;
		if (current_state.cost < initial_ass.cost) {
			initial_ass = current_state;
			p_ctrl->setCurrentCost(current_state.cost,
					current_state.CurrColors(), number_of_iterations);
			idle_iterations_ls = 0;
		}
		idle_iterations++;
		idle_iterations_ls++;
	}
}
