
/*
        Title: Greedy Routines Source File.
        file: greedy.c
        does: greedy routine
        Source: Joseph Culberson's Coloring Page
                http://web.cs.ualberta.ca/~joe/Coloring/index.html
        Author: Joseph Culberson
        email: joe@cs.ualberta.ca
	Copyright (c) 1997 Joseph Culberson. All rights reserved.

	Redistribution and use in source and binary forms, with or without
	modification, are permitted provided that the following conditions
	are met:
	1. Redistributions of source code must retain the above copyright
   	notice, this list of conditions and the following disclaimer.
	2. Redistributions in binary form must reproduce the above copyright
   	notice, this list of conditions and the following disclaimer in the
   	documentation and/or other materials provided with the distribution.
	3. All advertising materials mentioning features or use of this 
	software must display the following acknowledgement:
     	This product includes software developed by J. Culberson at the
     	University of Alberta, Edmonton.
	4. Neither the name of the University nor the names of its contributors
   	may be used to endorse or promote products derived from this software
   	without specific prior written permission.

	THIS SOFTWARE IS PROVIDED BY THE CONTRIBUTORS ``AS IS'' AND ANY
	EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
	THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
	PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
	CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
	SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
	NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
	LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
	HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
	CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
	OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
	EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

	THIS SOFTWARE IS SUPPLIED WITHOUT ANY SUPPORT SERVICES. 

*/
#include "Explorer.h"


KempeChains::KempeChains(ColoringManager* sm, Random *rnd, 
	    Graph* pin = NULL):ColNeighborhoodExplorer(sm,rnd,pin) 
{
  flagset.resize(pin->Vertices(),0);
}

KempeChains::KempeChains(const ColNeighborhoodExplorer* ne):
  ColNeighborhoodExplorer(ne) 
{
  flagset.resize(p_in->Vertices(),0);
}

void KempeChains::initAuxiliar(const Color K)
{
  last_mod.resize(K+1);
  full_kempe.resize(K+1);
  for (Color i=0; i < K+1; i++)
    {
      full_kempe[i].resize(K+1);
    }
}

int  KempeChains::bfscomponent(const CompleteColoring &ass, Vertex v,Color clr1,Color clr2,
			       int compnum)
/*
	find the compoent on v (must be in clr2) and return
	1 if more of clr2 than clr1, else return 0
	flagset will be set to the color of the vertex
	for each vertex in the component.
*/
{
  Vertex visit = p_in->Vertices();

  	vector<Vertex> queue1;
	vector<Vertex> queue2;

	vector<Vertex>::iterator u;
	VertexSet::iterator pt;
	/* We can assume that the color classes are much smaller
		than nbrhoods, so traverse the color classes */

	queue2.push_back(v);
	//first2 = 0; last2 = 0; queue2[first2] = v;
	flagset[v] = compnum;

	Vertex cnt1 = 0;
	Vertex cnt2 = 0;

	while(queue2.size()) 
	  {
	    /* assume queue 1 is empty */
	    queue1.clear();
	    u=queue2.begin();
	    while(u!=queue2.end()) 
	      {
		cnt2++;
		pt = ass.partitions[clr1].begin();
		while (pt != ass.partitions[clr1].end()) 
		  {
		    if (p_in->Edge(*u,*pt)) 
		      {
			if (flagset[*pt] == visit) {
			  flagset[*pt] = compnum;
			  queue1.push_back(*pt);
			}
		      }
		    pt++;
		  }
		u++;
	      }
	    
	    /* now empty queue2 and search through queue1 */
	    queue2.clear();
	    u=queue1.begin();
	    while(u!=queue1.end()) 
	      {
		cnt1++;
		pt = ass.partitions[clr2].begin();
		while (pt != ass.partitions[clr2].end()) 
		  {
		    if (p_in->Edge(*u,*pt)) {
		      if (flagset[*pt] == visit) {
			flagset[*pt] = compnum;
			queue2.push_back(*pt);
		      }
		    }
		    pt++;
		  }
		u++;
	      }
	  }

	//cout<<cnt1<<" "<<cnt2<<" "<<ass.partitions[clr1].size()<<" "<<
	// ass.partitions[clr2].size()<<endl;
//   if (p_sm->p_ctrl->getTime()>30)
//     cout<<cnt1<<" "<<ass.partitions[clr1].size()<<" "<<cnt2<<" "<<ass.partitions[clr2].size()<<endl;

	if (cnt1 == ass.partitions[clr1].size() 
	    &&  cnt2 == ass.partitions[clr2].size())
	  full_kempe[clr1][clr2] = iter;
	

//  	if ((cnt2 != cnt1) && 
//  	    (cnt1 != ass.partitions[clr1].size() 
//  	     || cnt2 != ass.partitions[clr2].size()))
	  
	if ((cnt1 != ass.partitions[clr1].size()) 
	       || (cnt2 != ass.partitions[clr2].size()))
	    return 1; 
	  else return 0;
	

//  	//this was what used in the paper for WEA 2003
//   	if ((cnt2 > cnt1) && 
//  	    (cnt1 != ass.partitions[clr1].size() 
//  	     || cnt2 != ass.partitions[clr1].size())) 
//  	  return 1; 
//  	else return 0;
	
//  	//what was in Culberson. I believe this is right if the two colors can be 
//  	//independently greater or less than the other
//  	if (cnt2 > cnt1) return(1); else return(0);

}

	
			

int KempeChains::CheckComponents(const CompleteColoring &ass, Color clr1, Color clr2,
				 fvalue &delta)
  /* check for components between color classes 1 and 2;
	if any clr1 is smaller than clr2, then swap the class 
	and return a flag value of one, else f no swaps return zero
*/
/* To increase efficiency, we use a data structure to store
	   a list of vertices of each color.  If p = 0.5,
	   then with prob. p we get a conflict after first test,
	  etc. On average we look at 2 vertices per color before rejecting.
	*/
{
  VertexSet::iterator pt;

  Vertex visit;
  int compnum,swaps;
  Vertex cnt1,cnt2;
	
  /* initialize flagset as not visited */
  visit = p_in->Vertices();
  /* sometimes color 1 will not all be visited in bfs */

  //memset(flagset,0,(n+2)*sizeof(vertextype));
  components.clear();
 
  cnt1 = 0;

  pt = ass.partitions[clr1].begin();
  while (pt != ass.partitions[clr1].end()) {
    flagset[*pt] = visit;
    pt++; 
    cnt1++; /* for debugging */		
  }

  cnt2 = 0;
  pt = ass.partitions[clr2].begin();
  while (pt != ass.partitions[clr2].end()) {
    flagset[*pt] = visit;
    pt++;
    cnt2++; /* for debugging */
  }
	
  compnum = 0;
  swaps = 0;
  pt = ass.partitions[clr2].begin();
  while (pt != ass.partitions[clr2].end()) 
    {
      if (flagset[*pt] == visit) { // not visited yet
	components.push_back(bfscomponent(ass,*pt,clr1,clr2,compnum));
	if (components[compnum])
	  {
	    swaps = 1;
	    /**delta_fit = deltaChain(clr1, clr2);*/
	  }
	compnum++;
      } 
      pt++;
    }


  /*	if (compnum > 1) {
	printf("CheckComp: compnum=%d swaps=%d\n",compnum,swaps);
	printf("\tclr1 = %d clr2 = %d\n",clr1,clr2);
	fflush(stdout);
	}
	if (swaps)
	{
	if (compnum >1) {
	printf("------------------------------------------------------------\n");
	printf("call swap c1,c2 %d, %d\n",clr1,clr2);
	printf("Before:\n");
	printsets(clr1,clr2,flagset,components,compnum);

#ifdef DEBUGKEMPE
void printsets(
vertextype clr1, vertextype clr2, vertextype flagset[],
int components[], int compnum)
{
	vertextype p;

	printf("Components compnum=%d:\n",compnum);
	for(p=0;p<compnum;p++) {
		printf("\tp = %d components[p]=%1d\n",p,components[p]);
	}
	printf("Colorset 1 = %d\n",clr1);
	printf("vertex\t flagset\n");
	p = start[clr1];
	while (p <=n) {
		printf("%5d\t%5d\n",p,flagset[p]);
		p = next_ver[p];
	}
	printf("Colorset 2 = %d\n",clr2);
	printf("vertex\t flagset\n");
	p = start[clr2];
	while (p <=n) {
		printf("%5d\t%5d\n",p,flagset[p]);
		p = next_ver[p];
	}
}
#endif

	printf("before swap cnt1= %d cnt2 = %d\n",cnt1,cnt2);
	}
	printf("perform swap\n");
	Swap(clr1,clr2,flagset,components);
	if (compnum >1) {
	printf("After:\n");
	printsets(clr1,clr2,flagset,components,compnum);
	printf("------------------------------------------------------------\n");
	fflush(stdout);
	}
	}
  */
	
  if (swaps) 
    {
      best_components.clear();
      delta = deltaBestChains(ass, clr1, clr2);
      return true;
    } 
  else  
    return false;
}





 
void KempeChains::recordMembers(const CompleteColoring &ass, Color clr1,Color clr2,
		   int component,vector<int> &members)
{
  VertexSet::iterator pt;
  
  pt = ass.partitions[clr1].begin();
  while (pt != ass.partitions[clr1].end()) {
    if ((int) flagset[*pt]==component) {
      members.push_back(*pt);
	  } 
    pt++;
  } 
	
  pt = ass.partitions[clr2].begin();
  while (pt != ass.partitions[clr2].end()) {
    if ((int) flagset[*pt]==component) {
       members.push_back(*pt);
    }
    pt++;
  } 
}

bool KempeChains::isTabu(TabuList &t_l, const CompleteColoring &ass, Color clr1,Color clr2,
		   int component, unsigned long par_iter)
{
  VertexSet::iterator pt;
  bool is_tabu=false;
		
  pt = ass.partitions[clr1].begin();
  while (pt != ass.partitions[clr1].end()) {
    if ((int) flagset[*pt]==component) {
      if (t_l.isTabu(*pt,clr2,par_iter))
	is_tabu=true;
    }
    pt++;
  } 
	
  pt = ass.partitions[clr2].begin();
  while (pt != ass.partitions[clr2].end()) {
    if ((int) flagset[*pt]==component) {
      if (t_l.isTabu(*pt,clr1,par_iter))
	is_tabu=true;
    } 
    pt++;
  } 
  if (is_tabu)
    return true;
  else
    return false;
}


	 


fvalue KempeChains::deltaBestChains(const CompleteColoring &ass, Color clr1, Color clr2)
{
  Vertex l1;
  Vertex l2;
  int c1swap,c2swap,c1same,c2same;
	
  unsigned int n_comp;
	
  VertexSet::iterator pt;
	
  fvalue delta_parz, best_delta = FVALUE_MAX;

  for (n_comp=0; n_comp < components.size(); n_comp++)
	  {
	    if (components[n_comp])
	      {
		/* swap the colors that match the flag set */
		
		l1 = 0; l2 = 0;
		
		/* okay its dumb, but I just delete them all off then put them
		   all back on the appropriate lists */
		
		c1swap = 0;
		c1same = 0;
		pt = ass.partitions[clr1].begin();
		while (pt != ass.partitions[clr1].end()) 
		  {
		    if ( (flagset[*pt]==n_comp)) {
		      c1swap++;
		      l2++;
		    } else {
		      c1same++;
		      l1++;
		    }
		    pt++;
		} 
		
		c2swap =0;
		c2same=0;
		pt = ass.partitions[clr2].begin();
		while (pt != ass.partitions[clr2].end()) 
		  {
		    if ( (flagset[*pt]==n_comp)) {
		      c2swap++;
		      l1++;
		    } else {
		      c2same++;
		      l2++;
		    }
		    pt++;
		}
		
		delta_parz = -(l1*l1 + l2*l2
			       - ass.partitions[clr1].size() * ass.partitions[clr1].size()
			       - ass.partitions[clr2].size() * ass.partitions[clr2].size());
		/*printf("[%ld,%ld,%ld,%ld]",l1,l2,cardinality[clr1],cardinality[clr2]);*/
		/*printf(" {%ld} ",delta_parz);*/
		
		if (delta_parz == best_delta)
		  {
		    best_components.push_back(n_comp);
		  }
		else if (delta_parz < best_delta)
		  {
		    best_components.clear();
		    best_components.push_back(n_comp);
		    best_delta = delta_parz;
		  }
	      }
	  }

	
#ifdef SWAPOUT
	if ( ((c1swap > 1)  && (c1same >1)) &&
		((c2swap > 1)  && (c2same >1)) ) {
		printf("Real swap clr1 c1swap = %d c1same = %d\n",
			c1swap,c1same);
		printf("Real swap clr2 c2swap = %d c2same = %d\n",
			c2swap,c2same);
	}
#endif
	
	//printf(" best:%ld ",best_delta);
	return best_delta;
}



void KempeChains::MakeMove(CompleteColoring &ass, const ColMove &mv)
{

  assert(mv.type==KEMPE_EX);

  Color K = ass.CurrColors();
  
  Color clr1 = mv.seq[0];
  Color clr2 = mv.seq[1];

  last_mod[clr1] = iter;
  last_mod[clr2] = iter;

  for (unsigned int i = 2; i < mv.seq.size(); i++)
    {
      Vertex node = mv.seq[i];
      if (ass.col[node]==clr1)
	{
	  ass.col[node] = clr2;
	  ass.partitions[clr1].erase(node);
	  ass.partitions[clr2].insert(node);
	}
      else if (ass.col[node]==clr2)
	{
	  ass.col[node] = clr1;
	  ass.partitions[clr2].erase(node);
	  ass.partitions[clr1].insert(node);
	}
      else 
	{
	  cerr<<"Problems in making Kempe move"<<endl;
	  exit(1);
	}
    }

  ass.cost += mv.delta;
  
  /*ATTENTION: it lacks a procedure to control
    there is no color empty different from the two here considered!!*/
  
  Color tmp = K;

  if (clr1 == tmp || clr2 == tmp)
   {
     K++;
     ass.SetColors(K);
     ass.partitions.resize(K+1);
   }

  if (ass.partitions[clr1].size()==0)
    {
      if (clr1<tmp-1)
	{
	  ass.partitions[clr1]=ass.partitions[tmp-1];
	  ass.partitions[tmp-1].clear();
	  
	  VertexSet::iterator pt;
	  for (pt = ass.partitions[clr1].begin();
	       pt != ass.partitions[clr1].end();
	       pt++)
	    ass.col[*pt]=clr1;
	}
      K--;
      ass.SetColors(K);
      ass.partitions.resize(K+1);
    }
  else if (ass.partitions[clr2].size()==0)
    {
      if (clr2<tmp-1)
	{
	  ass.partitions[clr2]=ass.partitions[tmp-1];
	  ass.partitions[tmp-1].clear();
	  
	  VertexSet::iterator pt;
	  for (pt = ass.partitions[clr2].begin();
	       pt != ass.partitions[clr2].end();
	       pt++)
	    ass.col[*pt]=clr2;
	}
      K--;
      ass.SetColors(K);
      ass.partitions.resize(K+1);
    }
  
  if (K > tmp)
    {
      last_mod.resize(K+1);
      full_kempe.resize(K+1);
      for (Color i = 0; i < K+1; i++)
	{
	  full_kempe[i].resize(K+1);
	}
    }  
 //   cout<<"--->"<<K<<endl;
//    for (Color i=0; i < ass.partitions.size(); i++)
//      {
//        cout<<ass.partitions[i].size()<<" ";
//      }
//    cout<<endl;


}



int KempeChains::CheckComponentsSingle(const CompleteColoring &ass, Color clr1, Color clr2,
				       fvalue &delta)
  /* check for components between color classes 1 and 2;
	if any clr1 is smaller than clr2, then swap the class 
	and return a flag value of one, else f no swaps return zero
*/
/* To increase efficiency, we use a data structure to store
	   a list of vertices of each color.  If p = 0.5,
	   then with prob. p we get a conflict after first test,
	  etc. On average we look at 2 vertices per color before rejecting.
	*/
{
  VertexSet::iterator pt;
  Vertex node = p_in->Vertices();
  Vertex visit;
  int compnum,swaps;
  
  /* initialize flagset as not visited */
  visit = p_in->Vertices();
  /* sometimes color 1 will not all be visited in bfs */

  //memset(flagset,0,(n+2)*sizeof(vertextype));
  components.clear();
 
 
  pt = ass.partitions[clr1].begin();
  while (pt != ass.partitions[clr1].end()) {
    flagset[*pt] = visit;
    pt++; 
   }

  Vertex node_index = rg->ran0N(ass.partitions[clr2].size()-1);
  
  
  Vertex cnt2=0;
  pt = ass.partitions[clr2].begin();
  while (pt != ass.partitions[clr2].end()) {
    flagset[*pt] = visit;
    if (cnt2 == node_index)
      {
	node = *pt;
      }
    pt++;
    cnt2++;
  }
  
 
  compnum = 0;
  swaps = 0;

  if (flagset[node] == visit) { // not visited yet
    components.push_back(bfscomponent(ass,node,clr1,clr2,compnum));
    if (components[compnum])
      {
	swaps = 1;
	/**delta_fit = deltaChain(clr1, clr2);*/
      }
    compnum++;
  } 

  if (swaps) 
    {
      best_components.clear();
      delta = deltaBestChains(ass, clr1, clr2);
      return true;
    } 
  else  
    return false;
}

void KempeChains::NextMove(const CompleteColoring& coloring, ColMove& rc,
			   long unsigned current_iter)//First Improvement
{
  fvalue delta = FVALUE_MAX, min = FVALUE_MAX;
  
  long int count,rnd;
  Color c,d;
  Color K = coloring.CurrColors();
  //coloring.partitions.resize(K+1);
  
  vector<ColMove> chosen;
  iter = current_iter;
  count = 0;
  while (1)
    {
      do {
	d = rg->ran0N(K);
	c = rg->ran0N(K);
      } while ((c==d) || recentlyFullKempe(c,d));
      if (c<d)
	{
	  Color h=c;
	  c=d;
	  d=h;
	}
      // best_components.clear();
      if (CheckComponentsSingle(coloring,c,d,delta))
	{
	  if (delta == min) {
	    for (Vertex l=0;l<best_components.size();l++)
	      {
		ColMove tmp;
		tmp.type = KEMPE_EX;
		tmp.seq.push_back(c);
		tmp.seq.push_back(d);
		tmp.delta = delta;
		recordMembers(coloring,c,d,best_components[l],tmp.seq);
		chosen.push_back(tmp);
	      }
	  }
	  else if (delta < min)  {
	    for (Vertex l=0;l<best_components.size();l++)
	      {
		chosen.clear();
		ColMove tmp;
		tmp.type = KEMPE_EX;
		tmp.seq.push_back(c);
		tmp.seq.push_back(d);
		tmp.delta = delta;
		recordMembers(coloring,c,d,best_components[l],tmp.seq);
		chosen.push_back(tmp);
		min = delta;
	      }
	  }
	}
      
     
      if (chosen.size() == 1) {         
	rc = chosen[0];
	/*printf("1\n");*/
	break;
      }
      else if (chosen.size() > 1) {    
	rnd = rg->ran0N(chosen.size()-1);
	rc = chosen[rnd];
	/*printf("2\n");*/
	break;
      }
      else {     
	
	continue;
      }
    }

  assert(rc.type == 4);
}


//old one used for experiments of the paper!
/*
void KempeChains::NextMove(const CompleteColoring& coloring, ColMove& rc,
			   long unsigned current_iter)//First Improvement
{
  fvalue delta = FVALUE_MAX, min = FVALUE_MAX;
  
  long int count,rnd;
  Color c,d;
  Color K = coloring.CurrColors();
  //coloring.partitions.resize(K+1);
  
  vector<ColMove> chosen;
  iter = current_iter;
  count = 0;
  while (1)
    {
      do {
	d = rg->ran0N(K);
	c = rg->ran0N(K);
      } while ((c==d) || recentlyFullKempe(c,d));
      if (c<d)
	{
	  Color h=c;
	  c=d;
	  d=h;
	}
      // best_components.clear();
      if (CheckComponents(coloring,c,d,delta))
	{
	  if (delta == min) {
	    for (Vertex l=0;l<best_components.size();l++)
	      {
		ColMove tmp;
		tmp.type = 4;
		tmp.seq.push_back(c);
		tmp.seq.push_back(d);
		tmp.delta = delta;
		recordMembers(coloring,c,d,best_components[l],tmp.seq);
		chosen.push_back(tmp);
	      }
	  }
	  else if (delta < min)  {
	    for (Vertex l=0;l<best_components.size();l++)
	      {
		chosen.clear();
		ColMove tmp;
		tmp.type = 4;
		tmp.seq.push_back(c);
		tmp.seq.push_back(d);
		tmp.delta = delta;
		recordMembers(coloring,c,d,best_components[l],tmp.seq);
		chosen.push_back(tmp);
		min = delta;
	      }
	  }
	}
      
     
      if (chosen.size() == 1) {         
	rc = chosen[0];
	break;
      }
      else if (chosen.size() > 1) {    
	rnd = rg->ran0N(chosen.size()-1);
	rc = chosen[rnd];
	break;
      }
      else {     
	
	continue;
      }
    }

  assert(rc.type == 4);
}
*/



//This procedure does not have the check for the full chains!!!
void KempeChains::BestMove(const CompleteColoring& coloring, ColMove& mv)
{
  fvalue delta = FVALUE_MAX, min = FVALUE_MAX;
  
  long int count,rnd;
  count=0;	
  Color K = coloring.CurrColors();
  //  coloring.partitions.resize(K+1);
  
  vector<ColMove> chosen;
  
 
/*moves_already_tabu=0;*/
  for(Color c = 0; c < K+1 ; c++){
    for(Color d = 0; d < c ; d++){
      if (d!=c)
	  {
	    if (CheckComponents(coloring,c,d,delta))
	    {
	      /*printf("(%ld)\n",delta_fit);*/
	      /*if (delta_fit != 0)*/
		{
		  /*moves_already_tabu++;*/
		  if (delta == min) {
		    for (Vertex l=0;l<best_components.size();l++)
		      {
			ColMove tmp;
			    tmp.type = KEMPE_EX;
			    tmp.seq.push_back(c);
			    tmp.seq.push_back(d);
			    tmp.delta = delta;
			    recordMembers(coloring,c,d,best_components[l],tmp.seq);
			    chosen.push_back(tmp);
		      }
		  }
		  else if (delta < min)  {
		    for (Vertex l=0;l<best_components.size();l++)
		      {
			    chosen.clear();
			     ColMove tmp;
			     tmp.type = KEMPE_EX;
			    tmp.seq.push_back(c);
			    tmp.seq.push_back(d);
			    tmp.delta = delta;
			    recordMembers(coloring,c,d,best_components[l],tmp.seq);
			    chosen.push_back(tmp);
			    min = delta;
		      }
		  }
		}
	    }
	  }
    }
  }
	  
  if (min < 0)
    {
      if (chosen.size() == 1) {         
	mv = chosen[0];
	/*printf("1\n");*/
      }
      else if (chosen.size() > 1) {    
	rnd = rg->ran0N(chosen.size()-1);
	mv = chosen[rnd];
	/*printf("2\n");*/
      }
      else {     
	/*printf("%d ",moves_already_tabu);*/
	printf("no chains to be choosen\n");
	/*	    break;*//*or .....*/
	exit(1);

      }
    }
  else
    mv.type = NOT_DEFINED;
  

}



void KempeChains::BestMove(const CompleteColoring& coloring, ColMove& mv,
			   TabuList& t_l, unsigned long current_iter,
			   fvalue best_so_far)
{
  fvalue delta = FVALUE_MAX, min = FVALUE_MAX;
  
  long int count,rnd;

  Color K = coloring.CurrColors();
  //coloring.partitions.resize(K+1);
  
  vector<ColMove> chosen;
  iter = current_iter;
  count = 0;
  /*moves_already_tabu=0;*/

  for(Color c = 0; c < K+1 ; c++){
    for(Color d = 0;d < c ; d++){
      if ((d!=c) && !recentlyFullKempe(c,d))
	{
	  
	  if (CheckComponents(coloring,c,d,delta))
	    {
	      //printf("(%ld)\n",delta);
	      /*if (delta_fit != 0)*/
		{
		  /*moves_already_tabu++;*/
		  if (delta == min) {
		    for (Vertex l=0;l<best_components.size();l++)
		      {
			if (((coloring.cost + delta) < best_so_far)
			    || (!isTabu(t_l,coloring,c,d,best_components[l],current_iter)))
			  {
			    ColMove tmp;
			    tmp.type = KEMPE_EX;
			    tmp.seq.push_back(c);
			    tmp.seq.push_back(d);
			    tmp.delta = delta;
			    recordMembers(coloring,c,d,best_components[l],tmp.seq);
			    chosen.push_back(tmp);
			  }
		      }
		  }
		  else if (delta < min)  {
		    for (Vertex l=0;l<best_components.size();l++)
		      {
			if (((coloring.cost + delta) < best_so_far)
			    || (!isTabu(t_l,coloring,c,d,best_components[l],current_iter)))
			  {
			    chosen.clear();
			     ColMove tmp;
			     tmp.type = KEMPE_EX;
			    tmp.seq.push_back(c);
			    tmp.seq.push_back(d);
			    tmp.delta = delta;
			    recordMembers(coloring,c,d,best_components[l],tmp.seq);
			    chosen.push_back(tmp);
			    min = delta;
			  }
		      }
		  }
		}
	    }
	  }
    }
  }
	  
  
  if (chosen.size() == 1) {         
    mv = chosen[0];
    /*printf("1\n");*/
  }
  else if (chosen.size() > 1) {    
    rnd = rg->ran0N(chosen.size()-1);
    mv = chosen[rnd];
    /*printf("2\n");*/
  }
  else {     
    /*printf("%d ",moves_already_tabu);*/
    printf("no chains to be chosen\n");
    /*	    break;*//*or .....*/
    exit(1);
  }
  //cout<<"kempe chain "<<min<<endl;

}


int  KempeChains::bfscomponentWithTabu(const CompleteColoring &ass, Vertex v,Color clr1,Color clr2,
			       int compnum)
/*
	find the compoent on v (must be in clr2) and return
	1 if more of clr2 than clr1, else return 0
	flagset will be set to the color of the vertex
	for each vertex in the component.
*/
{
  Vertex visit = p_in->Vertices();

  	vector<Vertex> queue1;
	vector<Vertex> queue2;
	vector<Vertex> chain;

	vector<Vertex>::iterator u;
	VertexSet::iterator pt;
	/* We can assume that the color classes are much smaller
		than nbrhoods, so traverse the color classes */

	queue2.push_back(v);
	chain.push_back(v);
	//first2 = 0; last2 = 0; queue2[first2] = v;
	flagset[v] = compnum;

	Vertex cnt1 = 0;
	Vertex cnt2 = 0;

	while(queue2.size()) 
	  {
	    /* assume queue 1 is empty */
	    queue1.clear();
	    u=queue2.begin();
	    while(u!=queue2.end()) 
	      {
		cnt2++;
		pt = ass.partitions[clr1].begin();
		while (pt != ass.partitions[clr1].end()) 
		  {
		    if (p_in->Edge(*u,*pt)) 
		      {
			if (flagset[*pt] == visit) {
			  flagset[*pt] = compnum;
			  queue1.push_back(*pt);
			  chain.push_back(*pt);
			}
			else if (flagset[*pt] == visit+1)
			  {
			    for (u = chain.begin();u!=chain.end();u++)
			      flagset[*u] = visit;
			    return 0;
			  }
			  
		      }
		    pt++;
		  }
		u++;
	      }
	    
	    /* now empty queue2 and search through queue1 */
	    queue2.clear();
	    u=queue1.begin();
	    while(u!=queue1.end()) 
	      {
		cnt1++;
		pt = ass.partitions[clr2].begin();
		while (pt != ass.partitions[clr2].end()) 
		  {
		    if (p_in->Edge(*u,*pt)) {
		      if (flagset[*pt] == visit) {
			flagset[*pt] = compnum;
			queue2.push_back(*pt);
			chain.push_back(*pt);
		      }
		      else if (flagset[*pt] == visit+1)
			  {
			    for (u = chain.begin();u!=chain.end();u++)
			      flagset[*u] = visit;
			    return 0;
			  }
		    }
		    pt++;
		  }
		u++;
	      }
	  }

	//cout<<cnt1<<" "<<cnt2<<" "<<ass.partitions[clr1].size()<<" "<<
	// ass.partitions[clr2].size()<<endl;

	if (cnt1 == ass.partitions[clr1].size() 
	    &&  cnt2 == ass.partitions[clr2].size())
	  full_kempe[clr1][clr2] = iter;
	
	if ((cnt1 != ass.partitions[clr1].size()) 
	       || (cnt2 != ass.partitions[clr2].size()))
	  return 1; 
	else return 0;
	

//  	//this was what used in the paper for WEA 2003
//   	if ((cnt2 > cnt1) && 
//  	    (cnt1 != ass.partitions[clr1].size() 
//  	     || cnt2 != ass.partitions[clr1].size())) 
//  	  return 1; 
//  	else return 0;
	
//  	//what was in Culberson. I believe this is right if the two colors can be 
//  	//independently greater or less than the other
//  	if (cnt2 > cnt1) return(1); else return(0);

}


int KempeChains::CheckComponentsWithTabu(TabuList& t_l, const CompleteColoring &ass, Color clr1, Color clr2,
				 fvalue &delta, unsigned long current_iter)
  /* check for components between color classes 1 and 2;
	if any clr1 is smaller than clr2, then swap the class 
	and return a flag value of one, else f no swaps return zero
*/
/* To increase efficiency, we use a data structure to store
	   a list of vertices of each color.  If p = 0.5,
	   then with prob. p we get a conflict after first test,
	  etc. On average we look at 2 vertices per color before rejecting.
	*/
{
  VertexSet::iterator pt;

  Vertex visit;
  int compnum,swaps;
  Vertex cnt1,cnt2;
	
  /* initialize flagset as not visited */
  visit = p_in->Vertices();
  /* sometimes color 1 will not all be visited in bfs */

  //memset(flagset,0,(n+2)*sizeof(vertextype));
  components.clear();
 
  cnt1 = 0;

  pt = ass.partitions[clr1].begin();
  while (pt != ass.partitions[clr1].end()) {
     if (t_l.isTabu(*pt,clr2,current_iter))
      flagset[*pt] = visit+1;
    else
      flagset[*pt] = visit;
    pt++; 
    cnt1++; /* for debugging */		
  }

  cnt2 = 0;
  pt = ass.partitions[clr2].begin();
  while (pt != ass.partitions[clr2].end()) {
    if (t_l.isTabu(*pt,clr1,current_iter))
      flagset[*pt] = visit+1;
    else
      flagset[*pt] = visit;
    pt++;
    cnt2++; /* for debugging */
  }
	
  compnum = 0;
  swaps = 0;
  pt = ass.partitions[clr2].begin();
  while (pt != ass.partitions[clr2].end()) 
    {
      if (flagset[*pt] == visit) { // not visited yet & not tabu
	components.push_back(bfscomponentWithTabu(ass,*pt,clr1,clr2,compnum));
	if (components[compnum])
	  {
	    swaps = 1;
	    /**delta_fit = deltaChain(clr1, clr2);*/
	  }
	compnum++;
      } 
      pt++;
    }


  /*	if (compnum > 1) {
	printf("CheckComp: compnum=%d swaps=%d\n",compnum,swaps);
	printf("\tclr1 = %d clr2 = %d\n",clr1,clr2);
	fflush(stdout);
	}
	if (swaps)
	{
	if (compnum >1) {
	printf("------------------------------------------------------------\n");
	printf("call swap c1,c2 %d, %d\n",clr1,clr2);
	printf("Before:\n");
	printsets(clr1,clr2,flagset,components,compnum);

#ifdef DEBUGKEMPE
void printsets(
vertextype clr1, vertextype clr2, vertextype flagset[],
int components[], int compnum)
{
	vertextype p;

	printf("Components compnum=%d:\n",compnum);
	for(p=0;p<compnum;p++) {
		printf("\tp = %d components[p]=%1d\n",p,components[p]);
	}
	printf("Colorset 1 = %d\n",clr1);
	printf("vertex\t flagset\n");
	p = start[clr1];
	while (p <=n) {
		printf("%5d\t%5d\n",p,flagset[p]);
		p = next_ver[p];
	}
	printf("Colorset 2 = %d\n",clr2);
	printf("vertex\t flagset\n");
	p = start[clr2];
	while (p <=n) {
		printf("%5d\t%5d\n",p,flagset[p]);
		p = next_ver[p];
	}
}
#endif

	printf("before swap cnt1= %d cnt2 = %d\n",cnt1,cnt2);
	}
	printf("perform swap\n");
	Swap(clr1,clr2,flagset,components);
	if (compnum >1) {
	printf("After:\n");
	printsets(clr1,clr2,flagset,components,compnum);
	printf("------------------------------------------------------------\n");
	fflush(stdout);
	}
	}
  */
	
  if (swaps) 
    {
      best_components.clear();
      delta = deltaBestChains(ass, clr1, clr2);
      return true;
    } 
  else  
    return false;
}


//Tabu Search without Aspiration Criterion
void KempeChains::BestMove(const CompleteColoring& coloring, ColMove& mv,
			   TabuList& t_l, unsigned long current_iter)
			   
{
  fvalue delta = FVALUE_MAX, min = FVALUE_MAX;
  
  long int count,rnd;

  Color K = coloring.CurrColors();
  //coloring.partitions.resize(K+1);
  
  vector<ColMove> chosen;
  iter = current_iter;
  count = 0;
  /*moves_already_tabu=0;*/

  for(Color c = 0; c < K+1 ; c++){
    for(Color d = 0;d < c ; d++){
      if ((d!=c) && !recentlyFullKempe(c,d))
	{
	  if (CheckComponentsWithTabu(t_l,coloring,c,d,delta,current_iter))
	    {
		  if (delta == min) {
		    for (Vertex l=0;l<best_components.size();l++)
		      {
			ColMove tmp;
			tmp.type = KEMPE_EX;
			tmp.seq.push_back(c);
			tmp.seq.push_back(d);
			tmp.delta = delta;
			recordMembers(coloring,c,d,best_components[l],tmp.seq);
			chosen.push_back(tmp);
		      }
		  }
		  else if (delta < min)  {
		    for (Vertex l=0;l<best_components.size();l++)
		      {
			chosen.clear();
			ColMove tmp;
			tmp.type = KEMPE_EX;
			tmp.seq.push_back(c);
			tmp.seq.push_back(d);
			tmp.delta = delta;
			recordMembers(coloring,c,d,best_components[l],tmp.seq);
			chosen.push_back(tmp);
			min = delta;
		      }
		  }
	    }
	}
    }
  }
	  
  
  if (chosen.size() == 1) {         
    mv = chosen[0];
    /*printf("1\n");*/
  }
  else if (chosen.size() > 1) {    
    rnd = rg->ran0N(chosen.size()-1);
    mv = chosen[rnd];
    /*printf("2\n");*/
  }
  else {     
    /*printf("%d ",moves_already_tabu);*/
    printf("no chains to be chosen\n");
    /*	    break;*//*or .....*/
    exit(1);
  }
  //cout<<"kempe chain "<<min<<endl;

}
	


