/*
 * Manager.cpp
 *
 *  Created on: Jul 22, 2010
 *      Author: marco
 */
#include "Manager.h"

ColoringManager::ColoringManager(Random *rnd, Graph* pin = NULL, Control *ctrl =
		NULL) {
	p_in = pin;
	rg = rnd;
	p_ctrl = ctrl;
}

void ColoringManager::updateColoringData(Coloring &ass) {
	const Graph &g = *p_in; // g is an alias for the input Graph

	// first build the group list
	// and the edge list

	/*sceglie il ciclo piu conveniente su colore o nodi adiacenti*/
	//approx_col = 2*N/ass.k;
	//approx_adj = N*(p_in->AvDensity());

	if ((float) (g.Vertices() / ass.CurrColors()) < (float) (g.Vertices()
			* (p_in->Density())))
		ass.loop_on_partition = true;
	else
		ass.loop_on_partition = false;

	ass.cost = 0;
	ass.partitions.resize(ass.CurrColors());
	for (Color c = 0; c < ass.CurrColors(); c++) {
		ass.partitions[c].clear();
	}
	// rebuild the conflicting list
	// and the delta values
	// Update the improvement graph
	// this is of the order N^2!!
	ass.adj_x_color.clear();
	ass.adj_x_color.resize(ass.CurrColors() * g.Vertices(), 0);

}

void ColoringManager::updateColoringData(Coloring &ass, int representation) {
	if (representation == COMPLETE_REPR)
		updateColoringData(static_cast<CompleteColoring &> (ass));
	else if (representation == PARTIAL_REPR)
		updateColoringData(static_cast<PartialColoring &> (ass));
}
;

void ColoringManager::updateColoringData(CompleteColoring &ass) {
	const Graph &g = *p_in; // g is an alias for the input Graph

	// first build the group list
	// and the edge list
	/*sceglie il ciclo piu conveniente su colore o nodi adiacenti*/
	//approx_col = 2*N/ass.k;
	//approx_adj = N*(p_in->AvDensity());

	if ((float) (g.Vertices() / ass.CurrColors()) < (float) (g.Vertices()
			* (p_in->Density())))
		ass.loop_on_partition = true;
	else
		ass.loop_on_partition = false;

	ass.cost = 0;
	ass.conflicts.resize(g.Vertices());
	ass.conflicts_set.clear();
#ifdef CONFLICT_EDGES
	ass.conflict_edges.clear();
#endif
	ass.partitions.resize(ass.CurrColors());
	for (Color c = 0; c < ass.CurrColors(); c++) {
		ass.partitions[c].clear();
	}
	// rebuild the conflicting list
	// and the delta values
	// Update the improvement graph
	// this is of the order N^2!!
	ass.adj_x_color.clear();
	ass.adj_x_color.resize(ass.CurrColors() * g.Vertices(), 0);
	for (Vertex i = 0; i < g.Vertices(); i++) {
		ass.conflicts[i] = 0;
		ass.partitions[ass.col[i]].insert(i);
		VertexSet::const_iterator pt;
		for (pt = g.adjacency_set[i].begin(); pt != g.adjacency_set[i].end(); pt++) {
			ass.AdjColor(i, ass.col[*pt])++;
		}
	}
	for (Vertex i = 0; i < g.Vertices(); i++) {
		//  	for (Color c = 0; c < ass.k; c++)
		//  	{
		//  	  if (ass.col[i]!=c)
		//  	    ass.Delta(i, c)= ass.AdjColor(i,c) - ass.AdjColor(i,ass.col[i]);
		//  	  else
		//  	    ass.Delta(i,c) = MAX_FVALUE;
		//  	}
		for (Vertex j = i; j < g.Vertices(); j++) {
			if (g.Edge(i, j) && (ass.col[i] == ass.col[j])) {
				ass.cost++;
				ass.conflicts_set.insert(i);
				ass.conflicts[i]++;
				ass.conflicts_set.insert(j);
				ass.conflicts[j]++;
#ifdef CONFLICT_EDGES
				ass.conflict_edges.insert(pair<Vertex, Vertex> (i, j));
#endif
			}
			//  	  if (ass.col[i]!=ass.col[j])
			//  	    {
			//  	      ass.impGraph[i][j]=0;
			//  	      ass.impGraph[j][i]=0;

			//  	      ass.impGraph[i][j] = ass.AdjColor(i,ass.col[j])
			//  		- ass.AdjColor(j,ass.col[j]) - g.Edge(i,j);
			//  	      ass.impGraph[j][i] = ass.AdjColor(j,ass.col[i])
			//  		- ass.AdjColor(i,ass.col[i]) - g.Edge(i,j);
			//  	    }
			//  	  else
			//  	    {
			//  	      ass.impGraph[i][j] = MAX_FVALUE;
			//  	      ass.impGraph[j][i] = MAX_FVALUE;
			//  	    }

		}
	}
	if (ass.cost != 0)
		ass.feasible = false;
	else
		ass.feasible = true;
}

void ColoringManager::updateColoringData(PartialColoring &ass) {
	const Graph &g = *p_in; // g is an alias for the input Graph

	// first build the group list
	// and the edge list

	// TODO: not used
	//select how to most convienient way to examine adjacent vertices
	//approx_col = 2*N/ass.k;
	//approx_adj = N*(p_in->AvDensity());

	if ((float) (g.Vertices() / ass.CurrColors()) < (float) (g.Vertices()
			* (p_in->Density())))
		ass.loop_on_partition = true;
	else
		ass.loop_on_partition = false;

	ass.cost = 0;
	ass.partitions.resize(ass.CurrColors());
	ass.impasse_class.clear();
	for (Color c = 0; c < ass.CurrColors(); c++) {
		ass.partitions[c].clear();
	}
	// rebuild the conflicting list
	// and the delta values
	// Update the improvement graph
	// this is of the order N^2!!
	ass.adj_x_color.clear();
	ass.adj_x_color.resize(ass.CurrColors() * g.Vertices(), 0);
	for (Vertex i = 0; i < g.Vertices(); i++) {
		if (isNotAColor(ass.col[i])) {
			ass.impasse_class.insert(i);
			ass.cost++;
		} else {
			ass.partitions[ass.col[i]].insert(i);
		}
		VertexSet::const_iterator pt;
		for (pt = g.adjacency_set[i].begin(); pt != g.adjacency_set[i].end(); pt++) {
			if (!isNotAColor(ass.col[*pt]))
				ass.AdjColor(i, ass.col[*pt])++;
		}
	}

	if (ass.cost != 0)
		ass.feasible = false;
	else
		ass.feasible = true;
}

fvalue ColoringManager::computeFitness(CompleteColoring &ass) {
	ass.cost = 0;
	for (Vertex v = 0; v < p_in->Vertices(); v++) {
		for (Vertex w = v + 1; w < p_in->Vertices(); w++) {
			if (p_in->Edge(v, w) && (ass.col[v] == ass.col[w])) {
				ass.cost++;
			}
		}
	}
	return ass.cost;
}

fvalue ColoringManager::computeFitness(PartialColoring &ass) {
	ass.cost = 0;
	for (Vertex v = 0; v < p_in->Vertices(); v++) {
		if (isNotAColor(ass.col[v])) {
			ass.cost++;
		}
	}
	return ass.cost;
}

void ColoringManager::updateDeltaDataOneEx(CompleteColoring &ass, Vertex var,
		Color old_col, Color new_col) {
	// cout<<"-->"<<var<<"["<<old_col<<","<<new_col<<"]"<<endl;
	ass.conflicts[var] = ass.AdjColor(var, new_col);
	//+=ass.Delta(var,new_col);

	if (ass.conflicts[var] == 0)
		ass.conflicts_set.erase(var);
	else if (ass.conflicts[var])
		ass.conflicts_set.insert(var);

	//    ass.Delta(var,old_col) = - ass.Delta(var,new_col);

	//     for (Color j=0;j< ass.k;j++)
	//        {
	//          if (j!=old_col && j!=new_col)
	//    	{
	//    	  ass.Delta(var,j) -= ass.Delta(var,new_col);
	//    	}
	//        }
	//     ass.Delta(var,new_col) = MAX_FVALUE;

	VertexSet::const_iterator pt;
	for (pt = p_in->adjacency_set[var].begin(); pt
			!= p_in->adjacency_set[var].end(); pt++) {
		ass.AdjColor((*pt), old_col)--;
		ass.AdjColor((*pt), new_col)++;
		//
		//	ass.Delta((*pt),old_col)--;
		//	ass.Delta((*pt),new_col)++;
		//
		if (ass.col[(*pt)] == new_col) {
			//	  ass.Delta((*pt),old_col)--;
			ass.conflicts[(*pt)]++;
			ass.conflicts_set.insert(*pt);
#ifdef CONFLICT_EDGES
			if (var < *pt) {
				ass.conflict_edges.insert(pair<Vertex, Vertex> (var, *pt));
			} else {
				ass.conflict_edges.insert(pair<Vertex, Vertex> (*pt, var));
			}
#endif
			//ass.Delta((*pt),new_col)=MAX_FVALUE;
			//ass.Delta((*pt),old_col)--;
		}
		if (ass.col[(*pt)] == old_col) {
			ass.conflicts[(*pt)]--;
			if (ass.conflicts[(*pt)] == 0)
				ass.conflicts_set.erase(*pt);
#ifdef CONFLICT_EDGES
			//set<pair<Vertex, Vertex > >::iterator exp;
			if (var < *pt) {
				//	  cout<<var<<" "<<*pt<<" ";
				//exp = ass.conflict_edges.find(pair<Vertex,Vertex> (var,*pt));
				//if (exp==ass.conflict_edges.end())
				//{
				//   cout<<"no "<<endl;
				//}
				//else
				//  cout<<"["<<(*exp).first<<","<<(*exp).second<<"]"<<endl;
				ass.conflict_edges.erase(pair<Vertex, Vertex> (var, *pt));
			} else {
				//cout<<*pt<<" "<<var<<" ";
				//exp = ass.conflict_edges.find(pair<Vertex,Vertex> (*pt, var));
				// if (exp==ass.conflict_edges.end())
				//	  {
				//	      cout<<"no "<<endl;
				//	  }
				//else
				//   cout<<"["<<(*exp).first<<","<<(*exp).second<<"]"<<endl;
				ass.conflict_edges.erase(pair<Vertex, Vertex> (*pt, var));
			}
#endif
			//ass.Delta((*pt),old_col)=MAX_FVALUE;
			//ass.Delta((*pt),new_col)++;
		}
	}
}

void ColoringManager::updateDeltaDataISwap(PartialColoring &ass, Vertex var,
		Color old_col, Color new_col) {

	if (isNotAColor(old_col)) { // the vertex leaving impasse class
		VertexSet::const_iterator pt;
		for (pt = p_in->adjacency_set[var].begin(); pt
				!= p_in->adjacency_set[var].end(); pt++) {
			ass.AdjColor((*pt), new_col)++;
		}
	}
	if (isNotAColor(new_col)) { // the vertex entering impasse class
		VertexSet::const_iterator pt;
		for (pt = p_in->adjacency_set[var].begin(); pt
				!= p_in->adjacency_set[var].end(); pt++) {
			ass.AdjColor((*pt), old_col)--;
		}
	}
}

void ColoringManager::decrease_colors(Coloring &ass, int representation) {
	p_ctrl->getOutputStream() << "Decrease_colors\n";
	Color removed_col = (Color) rg->ran0N(ass.CurrColors() - 1);

	for (Vertex i = 0; i < p_in->Vertices(); i++) {
		if (ass.col[i] > removed_col) {
			ass.col[i] = ass.col[i] - 1;
		} else if (ass.col[i] == removed_col) {
			if (representation == COMPLETE_REPR)
				ass.col[i] = rg->ran0N(ass.CurrColors() - 2);
			else if (representation == PARTIAL_REPR)
				ass.col[i] = NOT_A_COLOR;
		}
	}
	ass.SetColors(ass.CurrColors() - 1);
	updateColoringData(ass, representation);
	p_ctrl->getOutputStream() << "Done\n";
}

//######################################################################
//######################################################################
//                        ColManWithPenalties
//######################################################################
//######################################################################


ColManWithPenalties::ColManWithPenalties(Random *rnd, Graph* pin = NULL,
		Control *ctrl = NULL) :
	ColoringManager(rnd, pin, ctrl) {
	penalties.resize(p_in->Vertices());
	for (unsigned int i = 0; i < p_in->Vertices(); i++) {
		penalties[i].resize(p_in->Vertices(), 0);
	}
}

void ColManWithPenalties::reinit_data() {
	for (unsigned int i = 0; i < p_in->Vertices(); i++) {
		for (unsigned int j = 0; j < p_in->Vertices(); j++) {
			penalties[i][j] = 0;
		}
	}
}

void ColManWithPenalties::scale_penalties() {
	for (unsigned int i = 0; i < p_in->Vertices(); i++) {
		for (unsigned int j = 0; j < p_in->Vertices(); j++) {
			penalties[i][j] *= (float) 4 / 5;
		}
	}
}

void ColManWithPenalties::updateColoringData(CompleteColoring &ass) {

	const Graph &g = *p_in; // g is an alias for the input Graph
	float lambda = p_ctrl->lambda;
	// first build the group list
	// and the edge list

	/*sceglie il ciclo piu conveniente su colore o nodi adiacenti*/
	//approx_col = 2*N/ass.k;
	//approx_adj = N*(p_in->AvDensity());

	if ((float) (g.Vertices() / ass.CurrColors()) < (float) (g.Vertices()
			* (p_in->Density())))
		ass.loop_on_partition = true;
	else
		ass.loop_on_partition = false;

	ass.cost = 0;
	ass.conflicts.resize(g.Vertices());
	ass.conflicts_set.clear();
	//ass.activation_bit.resize(g.Vertices())
	ass.partitions.resize(ass.CurrColors());
	//  ass.pen_delta.resize(g.Vertices());
	for (Color c = 0; c < ass.CurrColors(); c++) {
		ass.partitions[c].clear();
	}
	ass.pen_adj_x_color.clear();
	ass.pen_adj_x_color.resize(ass.CurrColors() * g.Vertices(), 0);
	ass.adj_x_color.clear();
	ass.adj_x_color.resize(ass.CurrColors() * g.Vertices(), 0);
	//cout << "here"<<endl;
	for (Vertex i = 0; i < g.Vertices(); i++) {
		//activation_bit[i] = false;
		//  ass.pen_delta[i].resize(ass.CurrColors(),0);
		ass.conflicts[i] = 0;
		ass.partitions[ass.col[i]].insert(i);
		VertexSet::const_iterator pt;
		for (pt = g.adjacency_set[i].begin(); pt != g.adjacency_set[i].end(); pt++) {
			ass.PenAdjColor(i, ass.col[*pt]) += (1 + (lambda
					* penalties[i][*pt]));
			ass.AdjColor(i, ass.col[*pt])++;
		}
	}
	//------------------------------------------------------------
	//    for (Vertex i=0; i<g.Vertices(); i++)
	//     {
	//       VertexSet::const_iterator pt;
	//       for (Color c = 0; c < ass.CurrColors(); c++)
	// 	{
	// 	  ass.pen_delta[i][c]=0;
	// 	  if (c!=ass.col[i])
	// 	    for (pt = g.adjacency_set[i].begin(); pt != g.adjacency_set[i].end(); pt++)
	// 	      {
	// 		if (ass.col[*pt]==c)
	// 		  ass.pen_delta[i][c] += (1 + (lambda * penalties[i][*pt]));
	// 		if (ass.col[*pt]==ass.col[i])
	// 		  ass.pen_delta[i][c] -= (1 + (lambda * penalties[i][*pt]));

	// 	      }

	// 	}
	//     }
	//------------------------------------------------------------
	//cout << "here2"<<endl;
	for (Vertex i = 0; i < g.Vertices(); i++) {

		for (Vertex j = i; j < g.Vertices(); j++) {
			if (g.Edge(i, j) && (ass.col[i] == ass.col[j])) {
				ass.cost++;
				ass.conflicts_set.insert(i);
				ass.conflicts[i]++;
				ass.conflicts_set.insert(j);
				ass.conflicts[j]++;
			}
		}
	}
	if (ass.cost != 0)
		ass.feasible = false;
	else
		ass.feasible = true;

}

void ColManWithPenalties::updateDeltaDataOneEx(CompleteColoring &ass,
		Vertex var, Color old_col, Color new_col) {
	float lambda = p_ctrl->lambda;
	ass.conflicts[var] = ass.AdjColor(var, new_col);
	if (ass.conflicts[var] == 0)
		ass.conflicts_set.erase(var);
	else if (ass.conflicts[var])
		ass.conflicts_set.insert(var);
	//------------------------------------------------------------
	//   ass.pen_delta[var][old_col] = - ass.pen_delta[var][new_col];
	//   for (Color c=0; c < ass.CurrColors(); c++)
	//     {
	//       if (c!=old_col && c!=new_col)
	// 	{
	// 	  ass.pen_delta[var][c]= ass.pen_delta[var][c]-ass.pen_delta[var][new_col];
	// 	}
	//     }
	//   ass.pen_delta[var][new_col] = 0;
	//------------------------------------------------------------


	VertexSet::const_iterator pt;
	for (pt = p_in->adjacency_set[var].begin(); pt
			!= p_in->adjacency_set[var].end(); pt++) {
		ass.AdjColor((*pt), old_col)--;
		ass.AdjColor((*pt), new_col)++;
		ass.PenAdjColor((*pt), old_col) -= (1 + (lambda * penalties[var][*pt]));
		ass.PenAdjColor((*pt), new_col) += (1 + (lambda * penalties[var][*pt]));
		if (ass.col[*pt] == new_col) {
			ass.conflicts[*pt]++;
			ass.conflicts_set.insert(*pt);
			//------------------------------------------------------------
			// 	  for (Color c=0; c < ass.CurrColors(); c++)
			// 	    {
			// 	      if (c != new_col)
			// 		ass.pen_delta[*pt][c] -= (1 + (lambda * penalties[var][*pt]));
			// 	    }
			//------------------------------------------------------------
		}
		if (ass.col[(*pt)] == old_col) {
			ass.conflicts[(*pt)]--;
			if (ass.conflicts[(*pt)] == 0)
				ass.conflicts_set.erase(*pt);
			//------------------------------------------------------------
			// 	  for (Color c=0; c < ass.CurrColors(); c++)
			// 	    {
			// 	      if (c != old_col)
			// 		ass.pen_delta[*pt][c] += (1 + (lambda * penalties[var][*pt]));
			// 	    }
			//------------------------------------------------------------
		}
		//------------------------------------------------------------
		//       if (ass.col[(*pt)]!=old_col)
		// 	ass.pen_delta[*pt][old_col] -= (1 + (lambda * penalties[var][*pt]));
		//       if (ass.col[(*pt)]!=new_col)
		// 	ass.pen_delta[*pt][new_col] += (1 + (lambda * penalties[var][*pt]));
		//------------------------------------------------------------
	}

}

void ColManWithPenalties::update_data(CompleteColoring &col) {
	float util;
	float max;
	vector < pair<Vertex, Vertex> > max_util_features;

	vector<pair<Vertex, Vertex> >::iterator p;
	max = -1;

	vector<pair<Vertex, Vertex> >::iterator pt;
	for (pt = p_in->edges.begin(); pt != p_in->edges.end(); pt++) {
		if (col.col[(*pt).first] == col.col[(*pt).second]) {
			util = ((float) 1) / (1 + penalties[(*pt).first][(*pt).second]);
			if (util == max) {
				max_util_features.push_back(*pt);
			} else if (util > max) {
				max_util_features.clear();
				max_util_features.push_back(*pt);
				max = util;
			}
		}
	}

	for (p = max_util_features.begin(); p != max_util_features.end(); p++) {

		penalties[(*p).second][(*p).first]++;
		penalties[(*p).first][(*p).second]++;

		if (col.col[(*p).first] == col.col[(*p).second]) {
			col.PenAdjColor((*p).first, col.col[(*p).first]) += p_ctrl->lambda;
			col.PenAdjColor((*p).second, col.col[(*p).second])
					+= p_ctrl->lambda;
		} else {
			col.PenAdjColor((*p).second, col.col[(*p).first]) += p_ctrl->lambda;
			col.PenAdjColor((*p).first, col.col[(*p).second]) += p_ctrl->lambda;
		}

		//------------------------------------------------------------
		//       if (col.col[(*p).first] == col.col[(*p).second])
		// 	{
		// 	  for (Color c = 0; c < col.CurrColors(); c++)
		// 	    {
		// 	      if (c!=col.col[(*p).first])
		// 		{
		// 		  col.pen_delta[(*p).first][c] -= (p_ctrl->lambda);
		// 		  col.pen_delta[(*p).second][c] -= (p_ctrl->lambda);
		// 		}
		// 	    }
		// 	}
		//       else
		// 	{
		// 	  col.pen_delta[(*p).second][col.col[(*p).first]] += p_ctrl->lambda;
		// 	  col.pen_delta[(*p).first][col.col[(*p).second]] += p_ctrl->lambda;
		// 	}
	}
}

void ColManWithPenalties::print_data() {
	ofstream pen_os("tmp/gls-pen_profile.dat");
	for (Vertex i = 0; i < p_in->Vertices(); i++)
		for (Vertex j = 0; j < p_in->Vertices(); j++)
			pen_os << penalties[i][j] << " ";
	pen_os << endl;
	pen_os.close();
}

void ColManPenaltyFunction::updateColoringData(CompleteColoring &ass)  {
	if ((float) (p_in->Vertices() / ass.CurrColors())
			< (float) (p_in->Vertices() * (p_in->Density())))
		ass.loop_on_partition = true;
	else
		ass.loop_on_partition = false;

	ass.cost = 0;
	ass.conflicts.resize(p_in->Vertices());
	ass.conflicts_set.clear();
	ass.partitions.resize(ass.CurrColors());
	for (Color c = 0; c < ass.CurrColors(); c++) {
		ass.partitions[c].clear();
	}
	// rebuild the conflicting list
	// and the delta values
	// Update the improvement graph
	// this is of the order N^2!!
	ass.adj_x_color.clear();
	//ass.adj_x_color.resize(ass.CurrColors()*g.Vertices(),0);
	for (Vertex i = 0; i < p_in->Vertices(); i++) {
		ass.conflicts[i] = 0;
		ass.partitions[ass.col[i]].insert(i);
	}

	for (Color i = 0; i < ass.CurrColors(); i++) {
		ass.cost -= ass.partitions[i].size() * ass.partitions[i].size();
	}
	//ass.feasible = true;
}
/*
fvalue ColManPenaltyFunction::computeFitness(CompleteColoring &ass) {
	ass.cost = 0;
	for (Color c = 0; c < ass.CurrColors(); c++) {
		ass.cost += ass.partitions[c].size() * ass.partitions[c].size();
	}
	return ass.cost;
}
*/

