/*
 * Manager.h
 *
 *  Created on: Jul 22, 2010
 *      Author: marco
 */

#ifndef MANAGER_H_
#define MANAGER_H_

#include "Solution.h"
#include "TabuList.h"

class ColoringManager {
public:
	ColoringManager(Random *rnd, Graph* pin, Control *ctrl);
	virtual ~ColoringManager() {
	}
	;
	//void UpdateColoringData(vector<Vertex>) const;
	virtual fvalue computeFitness(CompleteColoring &ass);
	virtual fvalue computeFitness(PartialColoring &ass);

	virtual void updateColoringData(Coloring &ass);
	virtual void updateColoringData(Coloring &ass, int representation);
	virtual void updateColoringData(CompleteColoring&);
	virtual void updateColoringData(PartialColoring&);

	void randomAssignment(Coloring&, Color num_col);

	void constructiveGreedy(Coloring&, Color max_col, int representation);
	void constructiveGreedyExtension(Coloring&, int representation);

	void constructiveRecursiveLargestFirst(Coloring& ass, Color max_col,
			int representation);
	void constructiveRecursiveLargestFirstReverse(Coloring&, Color max_col,
			int representation);
	void constructiveRecursiveLargestFirstStatic(Coloring&, Color max_col,
			int representation);

	void constructiveDSATUR(Coloring& ass, Color max_col, int representation);
	void constructiveDSATURExtension(Coloring& ass, Color K, int representation);

	void readStartingSolution(Coloring& ass, Color num_col);

	void decrease_colors(Coloring &ass, int representation);
	virtual void updateDeltaDataOneEx(CompleteColoring &ass, Vertex var,
			Color old_value, Color value);
	void updateDeltaDataISwap(PartialColoring &ass, Vertex var, Color old_col,
			Color new_col);

	Vertex distanceLap(Coloring &a, Coloring &b);

	Vertex perturbate_seq(CompleteColoring &ass, Color to_remove, TabuList &tl,
			long int iter);
	Vertex perturbate_dsat(CompleteColoring &ass, Color to_remove, TabuList &tl,
			long int iter);
	Vertex perturbate_best_moves(CompleteColoring &ass, Color to_remove, TabuList &tl,
			long int iter);
	Vertex perturbate_adj(CompleteColoring &ass, Color to_remove, TabuList &tl,
			long int iter);

	virtual void update_data(Coloring &col) {
	}
	;
	virtual void reinit_data() {
	}
	;
	virtual void scale_penalties() {
	}
	;
	virtual void print_data() {
	}
	;
	void color(Coloring &ass, Vertex var, Color level, Vertex &changed);
	void uncolor(Coloring &ass, Vertex var, Color level);
	double p_smooth;
protected:
	void SetInput(Graph* pin) {
		p_in = pin;
	}
	Graph *p_in;
	Random* rg;
	Control *p_ctrl;
};

class ColManPenaltyFunction: public ColoringManager {
public:
	ColManPenaltyFunction(Random *rnd, Graph* pin, Control *ctrl) :
		ColoringManager(rnd, pin, ctrl) {
	}
	;
	~ColManPenaltyFunction() {
	}
	;
	void updateColoringData(Coloring &ass) const {
			updateColoringData(static_cast<CompleteColoring &> (ass));
		};
	void updateColoringData(Coloring& ass, int representation) {
		if (representation == COMPLETE_REPR)
					updateColoringData(static_cast<CompleteColoring &> (ass));
				else {
					cerr<<"Something wrong in the representation passed to ColManWithPenalties"<<endl;
					exit(EXIT_FAILURE);
				}
	};
	void updateColoringData(CompleteColoring&);

//	fvalue computeFitness(CompleteColoring &ass);
};

class ColManWithPenalties: public ColoringManager {
public:
	ColManWithPenalties(Random *rnd, Graph* pin, Control *ctrl);
	~ColManWithPenalties() {
	}
	;
	//fvalue computeFitness(Coloring &ass);
	void updateColoringData(Coloring &ass) {
		updateColoringData(static_cast<CompleteColoring &> (ass));
	};
	void updateColoringData(Coloring &ass, int representation) {
		if (representation == COMPLETE_REPR)
			updateColoringData(static_cast<CompleteColoring &> (ass));
		else {
			cerr<<"Something wrong in the representation passed to ColManWithPenalties"<<endl;
			exit(EXIT_FAILURE);
		}
	};
	void updateColoringData(CompleteColoring&);

	void updateDeltaDataOneEx(CompleteColoring &ass, Vertex var, Color old_col,
			Color new_col);
	void update_data(CompleteColoring &col);
	void reinit_data();
	void print_data();
	void scale_penalties();
protected:
	vector<vector<float> > penalties;
};





class BranchAndBound:public ColoringManager
{
 public:
  BranchAndBound(Random *rnd, Graph* pin,Control *ctrl);
  BranchAndBound(Random *rnd, Graph* pin,Control *ctrl, Coloring &initial);
  ~BranchAndBound() {};
   Color chrom_number(Vertex Usize, Vertex first_u, Color B, Color K);

   vector<Vertex> best_C;
 private:


   struct _NodeAttributes {
// friend std::ostream& operator<<(std::ostream&, const NodeAttributes&);
//public:
// NodeAttributes() {node = 0; diff_col=0; adj_uncol=0; used_col.clear();};
// NodeAttributes(Vertex n, vector< bool > u_c, Color d, Vertex a )
//         {node = n; diff_col = d; adj_uncol=a; used_col = u_c;};
//
  //Vertex node;
     vector<Vertex> used_col;
     Color diff_col;
     Vertex adj_uncol;
   };

   void printNode(ostream& os, const _NodeAttributes& a);

   vector< _NodeAttributes > NodesData;
   vector<Vertex> U;
   vector< Vertex > C;
   Vertex N;
   long int tot_visited;
};


#endif /* MANAGER_H_ */
