/*
 * ParamSet.cpp
 *
 *  Created on: Jul 22, 2010
 *      Author: marco
 */
#include "Parameters.h"

PARAM* Parameters::hashtableRetrieve(const char * key /**< key to retrieve */
) {
	std::map<const char *, PARAM *, strCmp>::iterator iter = paramset.find(key);
	if (iter != paramset.end())
		return (iter->second);
	else
		return NULL;
}

/** sets Bool parameter according to the value of the given string */
RETCODE Parameters::paramParseBool(PARAM* param, /**< parameter */
char* valuestr /**< value in string format (may be modified during parse) */
) {
	assert(param != NULL);
	assert(param->paramtype == PARAMTYPE_BOOL);
	assert(valuestr != NULL);

	if (strcasecmp(valuestr, "TRUE") == 0) {
		CALL ( paramSetBool(param, TRUE) );
	} else if (strcasecmp(valuestr, "FALSE") == 0) {
		paramSetBool(param, FALSE);
	} else {
		printf("invalid parameter value <%s> for Bool parameter <%s>\n",
				valuestr, param->name);
		return PARSEERROR;
	}

	return OKAY;
}

/** sets int parameter according to the value of the given string */
RETCODE Parameters::paramParseInt(PARAM* param, /**< parameter */
char* valuestr /**< value in string format (may be modified during parse) */
) {
	int value;

	assert(param != NULL);
	assert(param->paramtype == PARAMTYPE_INT);
	assert(valuestr != NULL);

	if (sscanf(valuestr, "%d", &value) == 1) {
		paramSetInt(param, value);
	} else {
		printf("invalid parameter value <%s> for int parameter <%s>\n",
				valuestr, param->name);
		return PARSEERROR;
	}

	return OKAY;
}

/** sets Longint parameter according to the value of the given string */
RETCODE Parameters::paramParseLongint(PARAM* param, /**< parameter */
char* valuestr /**< value in string format (may be modified during parse) */
) {
	Longint value;

	assert(param != NULL);
	assert(param->paramtype == PARAMTYPE_LONGINT);
	assert(valuestr != NULL);

	if (sscanf(valuestr, "%"LONGINT_FORMAT, &value) == 1) {
		paramSetLongint(param, value);
	} else {
		printf("invalid parameter value <%s> for Longint parameter <%s>\n",
				valuestr, param->name);
		return PARSEERROR;
	}

	return OKAY;
}

/** sets Real parameter according to the value of the given string */
RETCODE Parameters::paramParseReal(PARAM* param, /**< parameter */
char* valuestr /**< value in string format (may be modified during parse) */
) {
	Real value;
	assert(param != NULL);
	assert(param->paramtype == PARAMTYPE_REAL);
	assert(valuestr != NULL);

	if (sscanf(valuestr, "%"REAL_FORMAT, &value) == 1) {
		paramSetReal(param, value);
	} else {
		printf("invalid parameter value <%s> for Real parameter <%s>\n",
				valuestr, param->name);
		return PARSEERROR;
	}

	return OKAY;
}

/** sets Char parameter according to the value of the given string */
RETCODE Parameters::paramParseChar(PARAM* param, /**< parameter */
char* valuestr /**< value in string format (may be modified during parse) */
) {
	char value;

	assert(param != NULL);
	assert(param->paramtype == PARAMTYPE_CHAR);
	assert(valuestr != NULL);

	if (sscanf(valuestr, "%c", &value) == 1) {
		paramSetChar(param, value);
	} else {
		printf("invalid parameter value <%s> for char parameter <%s>\n",
				valuestr, param->name);
		return PARSEERROR;
	}

	return OKAY;
}

/** sets String parameter according to the value of the given string */
RETCODE Parameters::paramParseString(PARAM* param, /**< parameter */
char* valuestr /**< value in string format (may be modified during parse) */
) {
	unsigned int len;

	assert(param != NULL);
	assert(param->paramtype == PARAMTYPE_STRING);
	assert(valuestr != NULL);

	/* check for quotes */
	len = strlen(valuestr);
	if (len <= 1 || valuestr[0] != '"' || valuestr[len - 1] != '"') {
		printf(
				"invalid parameter value <%s> for string parameter <%s> (string has to be in double quotes)\n",
				valuestr, param->name);
		return PARSEERROR;
	}

	/* remove the quotes */
	valuestr[len - 1] = '\0';
	valuestr++;
	paramSetString(param, valuestr);

	return OKAY;
}

/** returns value of Bool parameter */
Bool paramGetBool(PARAM* param /**< parameter */
) {
	assert(param != NULL);
	assert(param->paramtype == PARAMTYPE_BOOL);

	if (param->data.boolparam.valueptr != NULL)
		return *param->data.boolparam.valueptr;
	else
		return param->data.boolparam.curvalue;
}

/** returns default value of Bool parameter */
Bool paramGetBoolDefault(PARAM* param /**< parameter */
) {
	assert(param != NULL);
	assert(param->paramtype == PARAMTYPE_BOOL);

	return param->data.boolparam.defaultvalue;
}

/** returns value of int parameter */
int paramGetInt(PARAM* param /**< parameter */
) {
	assert(param != NULL);
	assert(param->paramtype == PARAMTYPE_INT);

	if (param->data.intparam.valueptr != NULL)
		return *param->data.intparam.valueptr;
	else
		return param->data.intparam.curvalue;
}

/** returns minimal value of int parameter */
int paramGetIntMin(PARAM* param /**< parameter */
) {
	assert(param != NULL);
	assert(param->paramtype == PARAMTYPE_INT);

	return param->data.intparam.minvalue;
}

/** returns maximal value of int parameter */
int paramGetIntMax(PARAM* param /**< parameter */
) {
	assert(param != NULL);
	assert(param->paramtype == PARAMTYPE_INT);

	return param->data.intparam.maxvalue;
}

/** returns default value of int parameter */
int paramGetIntDefault(PARAM* param /**< parameter */
) {
	assert(param != NULL);
	assert(param->paramtype == PARAMTYPE_INT);

	return param->data.intparam.defaultvalue;
}

/** returns value of Longint parameter */
Longint paramGetLongint(PARAM* param /**< parameter */
) {
	assert(param != NULL);
	assert(param->paramtype == PARAMTYPE_LONGINT);

	if (param->data.longintparam.valueptr != NULL)
		return *param->data.longintparam.valueptr;
	else
		return param->data.longintparam.curvalue;
}

/** returns minimal value of longint parameter */
Longint paramGetLongintMin(PARAM* param /**< parameter */
) {
	assert(param != NULL);
	assert(param->paramtype == PARAMTYPE_LONGINT);

	return param->data.longintparam.minvalue;
}

/** returns maximal value of longint parameter */
Longint paramGetLongintMax(PARAM* param /**< parameter */
) {
	assert(param != NULL);
	assert(param->paramtype == PARAMTYPE_LONGINT);

	return param->data.longintparam.maxvalue;
}

/** returns default value of Longint parameter */
Longint paramGetLongintDefault(PARAM* param /**< parameter */
) {
	assert(param != NULL);
	assert(param->paramtype == PARAMTYPE_LONGINT);

	return param->data.longintparam.defaultvalue;
}

/** returns value of Real parameter */
Real paramGetReal(PARAM* param /**< parameter */
) {
	assert(param != NULL);
	assert(param->paramtype == PARAMTYPE_REAL);

	if (param->data.realparam.valueptr != NULL)
		return *param->data.realparam.valueptr;
	else
		return param->data.realparam.curvalue;
}

/** returns minimal value of real parameter */
Real paramGetRealMin(PARAM* param /**< parameter */
) {
	assert(param != NULL);
	assert(param->paramtype == PARAMTYPE_REAL);

	return param->data.realparam.minvalue;
}

/** returns maximal value of real parameter */
Real paramGetRealMax(PARAM* param /**< parameter */
) {
	assert(param != NULL);
	assert(param->paramtype == PARAMTYPE_REAL);

	return param->data.realparam.maxvalue;
}

/** returns default value of Real parameter */
Real paramGetRealDefault(PARAM* param /**< parameter */
) {
	assert(param != NULL);
	assert(param->paramtype == PARAMTYPE_REAL);

	return param->data.realparam.defaultvalue;
}

/** returns value of char parameter */
char paramGetChar(PARAM* param /**< parameter */
) {
	assert(param != NULL);
	assert(param->paramtype == PARAMTYPE_CHAR);

	if (param->data.charparam.valueptr != NULL)
		return *param->data.charparam.valueptr;
	else
		return param->data.charparam.curvalue;
}

/** returns default value of char parameter */
char paramGetCharDefault(PARAM* param /**< parameter */
) {
	assert(param != NULL);
	assert(param->paramtype == PARAMTYPE_CHAR);

	return param->data.charparam.defaultvalue;
}

/** returns value of string parameter */
char* paramGetString(PARAM* param /**< parameter */
) {
	assert(param != NULL);
	assert(param->paramtype == PARAMTYPE_STRING);

	if (param->data.stringparam.valueptr != NULL)
		return *param->data.stringparam.valueptr;
	else
		return param->data.stringparam.curvalue;
}

/** returns default value of String parameter */
char* paramGetStringDefault(PARAM* param /**< parameter */
) {
	assert(param != NULL);
	assert(param->paramtype == PARAMTYPE_STRING);

	return param->data.stringparam.defaultvalue;
}

/** returns whether the parameter is on its default setting */
Bool paramIsDefault(PARAM* param /**< parameter */
) {
	assert(param != NULL);

	switch (param->paramtype) {
	case PARAMTYPE_BOOL:
		return (paramGetBool(param) == paramGetBoolDefault(param));

	case PARAMTYPE_INT:
		return (paramGetInt(param) == paramGetIntDefault(param));

	case PARAMTYPE_LONGINT:
		return (paramGetLongint(param) == paramGetLongintDefault(param));

	case PARAMTYPE_REAL:
		return EPSZ(paramGetReal(param) - paramGetRealDefault(param), 1e-16);

	case PARAMTYPE_CHAR:
		return (paramGetChar(param) == paramGetCharDefault(param));

	case PARAMTYPE_STRING:
		return (strcmp(paramGetString(param), paramGetStringDefault(param))
				== 0);

	default:
		printf("unknown parameter type\n");
		ABORT();
		return FALSE; /*lint !e527*/
	}
}

/*
 * Parameter set methods
 */

/** creates parameter set */
RETCODE Parameters::paramsetCreate() {
	//assert(paramset != NULL);

	//ALLOC( BMSallocMemory(paramset) );

	paramset.clear();

	return OKAY;
}

/** creates a parameter with name and description, does not set the type specific parameter values themselves */
RETCODE Parameters::paramCreate(PARAM** param, /**< pointer to the parameter */
const char* name, /**< name of the parameter */
const char* desc, /**< description of the parameter */
PARAMDATA* paramdata /**< locally defined parameter specific data */
) {
	assert(param != NULL);
	assert(name != NULL);
	assert(desc != NULL);

	duplicateMemoryArray(&((*param)->name),name,strlen(name)+1);
	duplicateMemoryArray(&(*param)->desc,desc,strlen(desc)+1);

	(*param)->paramdata = paramdata;

	return OKAY;
}


RETCODE Parameters::paramsetDelete(
) {
	std::map<const char *, PARAM*, strCmp>::iterator iter;
	for (iter = paramset.begin(); iter != paramset.end(); ++iter) {
		paramDelete(&(iter->second));
	}

	return OKAY;
}

RETCODE Parameters::paramDelete(PARAM** param /**< pointer to the parameter */
) {

	assert(param != NULL);
	   assert(*param != NULL);

	   switch( (*param)->paramtype )
	   {
	   case PARAMTYPE_BOOL:
	   case PARAMTYPE_INT:
	   case PARAMTYPE_LONGINT:
	   case PARAMTYPE_REAL:
	      break;
	   case PARAMTYPE_CHAR:
	      delete[] (*param)->data.charparam.allowedvalues;
	      break;
	   case PARAMTYPE_STRING:
	      delete[] (*param)->data.stringparam.defaultvalue;
	      if( (*param)->data.stringparam.valueptr == NULL )
	      {
	         delete[] (*param)->data.stringparam.curvalue;
	      }
	      else
	      {
	         delete[] (*param)->data.stringparam.valueptr;
	      }
	      break;
	   default:
	      fprintf(stderr,"invalid parameter type\n");
	      ABORT();
	   }

	delete[] (*param)->name;
	delete[] (*param)->desc;

	delete *param;
	return OKAY;
}

/** creates a Bool parameter, and sets its value to default */
RETCODE Parameters::paramCreateBool(PARAM** param, /**< pointer to the parameter */
const char* name, /**< name of the parameter */
const char* desc, /**< description of the parameter */
Bool* valueptr, /**< pointer to store the current parameter value, or NULL */
Bool isadvanced, /**< is this parameter an advanced parameter? */
Bool defaultvalue, /**< default value of the parameter */
PARAMDATA* paramdata /**< locally defined parameter specific data */
) {
	assert(param != NULL);
	assert(name != NULL);

	CALL( paramCreate(param, name, desc, paramdata) );

	(*param)->paramtype = PARAMTYPE_BOOL;
	(*param)->data.boolparam.valueptr = valueptr;
	(*param)->isadvanced = isadvanced;
	(*param)->data.boolparam.defaultvalue = defaultvalue;

	CALL( paramSetBool(*param, defaultvalue) );

	return OKAY;
}

/** creates a int parameter, and sets its value to default */
RETCODE Parameters::paramCreateInt(PARAM** param, /**< pointer to the parameter */
const char* name, /**< name of the parameter */
const char* desc, /**< description of the parameter */
int* valueptr, /**< pointer to store the current parameter value, or NULL */
Bool isadvanced, /**< is this parameter an advanced parameter? */
int defaultvalue, /**< default value of the parameter */
int minvalue, /**< minimum value for parameter */
int maxvalue, /**< maximum value for parameter */
PARAMDATA* paramdata /**< locally defined parameter specific data */
) {
	assert(param != NULL);
	assert(name != NULL);

	CALL( paramCreate(param, name, desc, paramdata) );

	(*param)->paramtype = PARAMTYPE_INT;
	(*param)->data.intparam.valueptr = valueptr;
	(*param)->isadvanced = isadvanced;
	(*param)->data.intparam.defaultvalue = defaultvalue;
	(*param)->data.intparam.minvalue = minvalue;
	(*param)->data.intparam.maxvalue = maxvalue;

	CALL( paramSetInt(*param, defaultvalue) );

	return OKAY;
}



/** creates a int parameter, and sets its value to default */
RETCODE Parameters::paramCreateShortint(PARAM** param, /**< pointer to the parameter */
const char* name, /**< name of the parameter */
const char* desc, /**< description of the parameter */
Shortint* valueptr, /**< pointer to store the current parameter value, or NULL */
Bool isadvanced, /**< is this parameter an advanced parameter? */
Shortint defaultvalue, /**< default value of the parameter */
Shortint minvalue, /**< minimum value for parameter */
Shortint maxvalue, /**< maximum value for parameter */
PARAMDATA* paramdata /**< locally defined parameter specific data */
) {
	assert(param != NULL);
	assert(name != NULL);

	CALL( paramCreate(param, name, desc, paramdata) );

	(*param)->paramtype = PARAMTYPE_INT;
	(*param)->data.shortintparam.valueptr = valueptr;
	(*param)->isadvanced = isadvanced;
	(*param)->data.shortintparam.defaultvalue = defaultvalue;
	(*param)->data.shortintparam.minvalue = minvalue;
	(*param)->data.shortintparam.maxvalue = maxvalue;

	CALL( paramSetShortint(*param, defaultvalue) );

	return OKAY;
}


/** creates a Longint parameter, and sets its value to default */
RETCODE Parameters::paramCreateLongint(PARAM** param, /**< pointer to the parameter */
const char* name, /**< name of the parameter */
const char* desc, /**< description of the parameter */
Longint* valueptr, /**< pointer to store the current parameter value, or NULL */
Bool isadvanced, /**< is this parameter an advanced parameter? */
Longint defaultvalue, /**< default value of the parameter */
Longint minvalue, /**< minimum value for parameter */
Longint maxvalue, /**< maximum value for parameter */
PARAMDATA* paramdata /**< locally defined parameter specific data */
) {
	assert(param != NULL);
	assert(name != NULL);

	CALL( paramCreate(param, name, desc, paramdata) );

	(*param)->paramtype = PARAMTYPE_LONGINT;
	(*param)->data.longintparam.valueptr = valueptr;
	(*param)->isadvanced = isadvanced;
	(*param)->data.longintparam.defaultvalue = defaultvalue;
	(*param)->data.longintparam.minvalue = minvalue;
	(*param)->data.longintparam.maxvalue = maxvalue;

	CALL( paramSetLongint(*param, defaultvalue) );

	return OKAY;
}

/** creates a Real parameter, and sets its value to default */
RETCODE Parameters::paramCreateReal(PARAM** param, /**< pointer to the parameter */
const char* name, /**< name of the parameter */
const char* desc, /**< description of the parameter */
Real* valueptr, /**< pointer to store the current parameter value, or NULL */
Bool isadvanced, /**< is this parameter an advanced parameter? */
Real defaultvalue, /**< default value of the parameter */
Real minvalue, /**< minimum value for parameter */
Real maxvalue, /**< maximum value for parameter */
PARAMDATA* paramdata /**< locally defined parameter specific data */
) {
	assert(param != NULL);
	assert(name != NULL);

	CALL( paramCreate(param, name, desc, paramdata) );

	(*param)->paramtype = PARAMTYPE_REAL;
	(*param)->data.realparam.valueptr = valueptr;
	(*param)->isadvanced = isadvanced;
	(*param)->data.realparam.defaultvalue = defaultvalue;
	(*param)->data.realparam.minvalue = minvalue;
	(*param)->data.realparam.maxvalue = maxvalue;

	CALL( paramSetReal(*param, defaultvalue) );

	return OKAY;
}

/** creates a char parameter, and sets its value to default */
RETCODE Parameters::paramCreateChar(PARAM** param, /**< pointer to the parameter */
const char* name, /**< name of the parameter */
const char* desc, /**< description of the parameter */
char* valueptr, /**< pointer to store the current parameter value, or NULL */
Bool isadvanced, /**< is this parameter an advanced parameter? */
char defaultvalue, /**< default value of the parameter */
const char* allowedvalues, /**< array with possible parameter values, or NULL if not restricted */
PARAMDATA* paramdata /**< locally defined parameter specific data */
) {
	assert(param != NULL);
	assert(name != NULL);

	CALL( paramCreate(param, name, desc, paramdata) );

	(*param)->paramtype = PARAMTYPE_CHAR;
	(*param)->data.charparam.valueptr = valueptr;
	(*param)->isadvanced = isadvanced;
	(*param)->data.charparam.defaultvalue = defaultvalue;
	if (allowedvalues != NULL) {
		duplicateMemoryArray(&(*param)->data.charparam.allowedvalues, allowedvalues, strlen(
				allowedvalues) + 1);
	} else
		(*param)->data.charparam.allowedvalues = NULL;

	CALL( paramSetChar(*param, defaultvalue) );

	return OKAY;
}

/** creates a string parameter, and sets its value to default */
RETCODE Parameters::paramCreateString(PARAM** param, /**< pointer to the parameter */
const char* name, /**< name of the parameter */
const char* desc, /**< description of the parameter */
char** valueptr, /**< pointer to store the current parameter value, or NULL */
Bool isadvanced, /**< is this parameter an advanced parameter? */
const char* defaultvalue, /**< default value of the parameter */
PARAMDATA* paramdata /**< locally defined parameter specific data */
) {
	assert(param != NULL);
	assert(name != NULL);
	assert(valueptr == NULL || *valueptr == NULL);
	assert(defaultvalue != NULL);

	CALL( paramCreate(param, name, desc, paramdata) );

	(*param)->paramtype = PARAMTYPE_STRING;
	(*param)->data.stringparam.valueptr = valueptr;
	(*param)->isadvanced = isadvanced;
	duplicateMemoryArray(&(*param)->data.stringparam.defaultvalue, defaultvalue, strlen(
			defaultvalue) + 1);
	(*param)->data.stringparam.curvalue = NULL;

	CALL( paramSetString(*param, defaultvalue) );

	return OKAY;
}

/** sets value of int parameter */
RETCODE Parameters::paramSetInt(PARAM* param, /**< parameter */
int value /**< new value of the parameter */
) {
	assert(param != NULL);

	/* check, if value is possible for the parameter */
	CALL_QUIET( paramCheckInt(param, value) );

	/* set the parameter's current value */
	if (param->data.intparam.valueptr != NULL)
		*param->data.intparam.valueptr = value;
	else
		param->data.intparam.curvalue = value;

	return OKAY;
}


/** sets value of int parameter */
RETCODE Parameters::paramSetShortint(PARAM* param, /**< parameter */
Shortint value /**< new value of the parameter */
) {
	assert(param != NULL);

	/* check, if value is possible for the parameter */
	CALL_QUIET( paramCheckInt(param, value) );

	/* set the parameter's current value */
	if (param->data.intparam.valueptr != NULL)
		*param->data.intparam.valueptr = value;
	else
		param->data.intparam.curvalue = value;

	return OKAY;
}


/** sets value of Longint parameter */
RETCODE Parameters::paramSetLongint(PARAM* param, /**< parameter */
Longint value /**< new value of the parameter */
) {
	assert(param != NULL);

	/* check, if value is possible for the parameter */
	CALL_QUIET( paramCheckLongint(param, value) );

	/* set the parameter's current value */
	if (param->data.longintparam.valueptr != NULL)
		*param->data.longintparam.valueptr = value;
	else
		param->data.longintparam.curvalue = value;

	return OKAY;
}

/** sets value of Real parameter */
RETCODE Parameters::paramSetReal(PARAM* param, /**< parameter */
Real value /**< new value of the parameter */
) {
	assert(param != NULL);

	/* check, if value is possible for the parameter */
	value = MAX(value, REAL_MIN);
	value = MIN(value, REAL_MAX);
	CALL_QUIET( paramCheckReal(param, value) );

	/* set the parameter's current value */
	if (param->data.realparam.valueptr != NULL)
		*param->data.realparam.valueptr = value;
	else
		param->data.realparam.curvalue = value;

	return OKAY;
}

/** sets value of char parameter */
RETCODE Parameters::paramSetChar(PARAM* param, /**< parameter */
char value /**< new value of the parameter */
) {
	assert(param != NULL);

	/* check, if value is possible for the parameter */
	CALL_QUIET( paramCheckChar(param, value) );

	/* set the parameter's current value */
	if (param->data.charparam.valueptr != NULL)
		*param->data.charparam.valueptr = value;
	else
		param->data.charparam.curvalue = value;

	return OKAY;
}

/** sets value of string parameter */
RETCODE Parameters::paramSetString(PARAM* param, /**< parameter */
const char* value /**< new value of the parameter */
) {
	assert(param != NULL);

	/* check, if value is possible for the parameter */
	CALL_QUIET( paramCheckString(param, value) );

	/* set the parameter's current value */
	if (param->data.stringparam.valueptr != NULL) {
		strncpy(*(param->data.stringparam.valueptr), value, strlen(value) + 1);
	} else {
		strncpy(param->data.stringparam.curvalue, value, strlen(value) + 1);
	}

	return OKAY;
}

/** adds parameter to the parameter set */
RETCODE Parameters::paramsetAdd(PARAM* param /**< parameter to add */
) {
	assert(param != NULL);
	/* insert the parameter name to the hash table */

	//paramWrite(param, std::cout, TRUE, FALSE);

	//paramset.insert(std::pair<const char *, PARAM *>(param->name, param));

	paramset[param->name]=param;
	return OKAY;
}

/** creates a Bool parameter, sets it to its default value, and adds it to the parameter set */
RETCODE Parameters::paramsetAddBool(const char* name, /**< name of the parameter */
const char* desc, /**< description of the parameter */
Bool* valueptr, /**< pointer to store the current parameter value, or NULL */
Bool isadvanced, /**< is this parameter an advanced parameter? */
Bool defaultvalue, /**< default value of the parameter */
PARAMDATA* paramdata /**< locally defined parameter specific data */
) {
	PARAM* param=new PARAM;

	/* create the parameter */
	CALL( paramCreateBool(&param, name, desc, valueptr, isadvanced, defaultvalue, paramdata) );

	/* add parameter to the parameter set */
	CALL( paramsetAdd(param) );

	return OKAY;
}

/** creates a int parameter, sets it to its default value, and adds it to the parameter set */
RETCODE Parameters::paramsetAddInt(const char* name, /**< name of the parameter */
const char* desc, /**< description of the parameter */
Int* valueptr, /**< pointer to store the current parameter value, or NULL */
Bool isadvanced, /**< is this parameter an advanced parameter? */
int defaultvalue, /**< default value of the parameter */
int minvalue, /**< minimum value for parameter */
int maxvalue, /**< maximum value for parameter */
PARAMDATA* paramdata /**< locally defined parameter specific data */
) {
	PARAM* param = new PARAM;

	/* create the parameter */
	CALL( paramCreateInt(&param, name, desc, valueptr, isadvanced, defaultvalue, minvalue, maxvalue,
					paramdata) );

	/* add parameter to the parameter set */
	CALL( paramsetAdd(param) );

	return OKAY;
}



/** creates a int parameter, sets it to its default value, and adds it to the parameter set */
RETCODE Parameters::paramsetAddShortint(const char* name, /**< name of the parameter */
const char* desc, /**< description of the parameter */
Shortint* valueptr, /**< pointer to store the current parameter value, or NULL */
Bool isadvanced, /**< is this parameter an advanced parameter? */
Shortint defaultvalue, /**< default value of the parameter */
Shortint minvalue, /**< minimum value for parameter */
Shortint maxvalue, /**< maximum value for parameter */
PARAMDATA* paramdata /**< locally defined parameter specific data */
) {
	PARAM* param = new PARAM;

	/* create the parameter */
	CALL( paramCreateShortint(&param, name, desc, valueptr, isadvanced, defaultvalue, minvalue, maxvalue,
					paramdata) );

	/* add parameter to the parameter set */
	CALL( paramsetAdd(param) );

	return OKAY;
}



/** creates a Longint parameter, sets it to its default value, and adds it to the parameter set */
RETCODE Parameters::paramsetAddLongint(const char* name, /**< name of the parameter */
const char* desc, /**< description of the parameter */
Longint* valueptr, /**< pointer to store the current parameter value, or NULL */
Bool isadvanced, /**< is this parameter an advanced parameter? */
Longint defaultvalue, /**< default value of the parameter */
Longint minvalue, /**< minimum value for parameter */
Longint maxvalue, /**< maximum value for parameter */
PARAMDATA* paramdata /**< locally defined parameter specific data */
) {
	PARAM* param=new PARAM;

	/* create the parameter */
	CALL( paramCreateLongint(&param, name, desc, valueptr, isadvanced, defaultvalue, minvalue, maxvalue,
					paramdata) );

	/* add parameter to the parameter set */
	CALL( paramsetAdd(param) );

	return OKAY;
}

/** creates a Real parameter, sets it to its default value, and adds it to the parameter set */
RETCODE Parameters::paramsetAddReal(const char* name, /**< name of the parameter */
const char* desc, /**< description of the parameter */
Real* valueptr, /**< pointer to store the current parameter value, or NULL */
Bool isadvanced, /**< is this parameter an advanced parameter? */
Real defaultvalue, /**< default value of the parameter */
Real minvalue, /**< minimum value for parameter */
Real maxvalue, /**< maximum value for parameter */
PARAMDATA* paramdata /**< locally defined parameter specific data */
) {
	PARAM* param=new PARAM;

	/* create the parameter */
	CALL( paramCreateReal(&param, name, desc, valueptr, isadvanced, defaultvalue, minvalue, maxvalue,
					paramdata) );

	/* add parameter to the parameter set */
	CALL( paramsetAdd(param) );

	return OKAY;
}

/** creates a char parameter, sets it to its default value, and adds it to the parameter set */
RETCODE Parameters::paramsetAddChar(const char* name, /**< name of the parameter */
const char* desc, /**< description of the parameter */
char* valueptr, /**< pointer to store the current parameter value, or NULL */
Bool isadvanced, /**< is this parameter an advanced parameter? */
char defaultvalue, /**< default value of the parameter */
const char* allowedvalues, /**< array with possible parameter values, or NULL if not restricted */
PARAMDATA* paramdata /**< locally defined parameter specific data */
) {
	PARAM* param=new PARAM;

	/* create the parameter */
	CALL( paramCreateChar(&param, name, desc, valueptr, isadvanced, defaultvalue, allowedvalues,
					paramdata) );

	/* add parameter to the parameter set */
	CALL( paramsetAdd(param) );

	return OKAY;
}

/** creates a string parameter, sets it to its default value, and adds it to the parameter set */
RETCODE Parameters::paramsetAddString(const char* name, /**< name of the parameter */
const char* desc, /**< description of the parameter */
char** valueptr, /**< pointer to store the current parameter value, or NULL */
Bool isadvanced, /**< is this parameter an advanced parameter? */
const char* defaultvalue, /**< default value of the parameter */
PARAMDATA* paramdata /**< locally defined parameter specific data */
) {
	PARAM* param=new PARAM;

	/* create the parameter */
	CALL( paramCreateString(&param, name, desc, valueptr, isadvanced, defaultvalue, paramdata) );

	/* add parameter to the parameter set */
	CALL( paramsetAdd(param) );

	return OKAY;
}

/** returns the name of the given paramter type */
const char* Parameters::paramtypeGetName(PARAMTYPE paramtype /**< type of parameter */
) {
	const char* paramtypename[] = { "Bool", /* PARAMTYPE_BOOL    = 0 */
	"int", /* PARAMTYPE_INT     = 1 */
	"Longint", /* PARAMTYPE_LONGINT = 2 */
	"Real", /* PARAMTYPE_REAL    = 3 */
	"char", /* PARAMTYPE_CHAR    = 4 */
	"string" /* PARAMTYPE_STRING  = 5 */
	};

	return paramtypename[(int) paramtype];
}

/** gets the value of an existing Bool parameter */
RETCODE Parameters::paramsetGetBool(const char* name, /**< name of the parameter */
Bool* value /**< pointer to store the parameter */
) {
	PARAM* param;

	assert(value != NULL);

	/* retrieve parameter from hash table */
	param = (PARAM*) hashtableRetrieve(name);
	if (param == NULL) {
		printf("parameter <%s> unknown\n", name);
		return PARAMETERUNKNOWN;
	}
	if (param->paramtype != PARAMTYPE_BOOL) {
		printf(
				"wrong parameter type - parameter <%s> has type <%s> instead of <%s>\n",
				name, paramtypeGetName(param->paramtype), paramtypeGetName(
						PARAMTYPE_BOOL));
		return PARAMETERWRONGTYPE;
	}

	/* get the parameter's current value */
	*value = paramGetBool(param);

	return OKAY;
}

/** gets the value of an existing int parameter */
RETCODE Parameters::paramsetGetInt(const char* name, /**< name of the parameter */
int* value /**< pointer to store the parameter */
) {
	PARAM* param;

	assert(value != NULL);

	/* retrieve parameter from hash table */
	param = (PARAM*) hashtableRetrieve(name);
	if (param == NULL) {
		printf("parameter <%s> unknown\n", name);
		return PARAMETERUNKNOWN;
	}
	if (param->paramtype != PARAMTYPE_INT) {
		printf(
				"wrong parameter type - parameter <%s> has type <%s> instead of <%s>\n",
				name, paramtypeGetName(param->paramtype), paramtypeGetName(
						PARAMTYPE_INT));
		return PARAMETERWRONGTYPE;
	}

	/* get the parameter's current value */
	*value = paramGetInt(param);

	return OKAY;
}

/** gets the value of an existing Longint parameter */
RETCODE Parameters::paramsetGetLongint(const char* name, /**< name of the parameter */
Longint* value /**< pointer to store the parameter */
) {
	PARAM* param;

	assert(value != NULL);

	/* retrieve parameter from hash table */
	param = (PARAM*) hashtableRetrieve(name);
	if (param == NULL) {
		printf("parameter <%s> unknown\n", name);
		return PARAMETERUNKNOWN;
	}
	if (param->paramtype != PARAMTYPE_LONGINT) {
		printf(
				"wrong parameter type - parameter <%s> has type <%s> instead of <%s>\n",
				name, paramtypeGetName(param->paramtype), paramtypeGetName(
						PARAMTYPE_LONGINT));
		return PARAMETERWRONGTYPE;
	}

	/* get the parameter's current value */
	*value = paramGetLongint(param);

	return OKAY;
}

/** gets the value of an existing Real parameter */
RETCODE Parameters::paramsetGetReal(const char* name, /**< name of the parameter */
Real* value /**< pointer to store the parameter */
) {
	PARAM* param;

	assert(value != NULL);

	/* retrieve parameter from hash table */
	param = (PARAM*) hashtableRetrieve(name);
	if (param == NULL) {
		printf("parameter <%s> unknown\n", name);
		return PARAMETERUNKNOWN;
	}
	if (param->paramtype != PARAMTYPE_REAL) {
		printf(
				"wrong parameter type - parameter <%s> has type <%s> instead of <%s>\n",
				name, paramtypeGetName(param->paramtype), paramtypeGetName(
						PARAMTYPE_REAL));
		return PARAMETERWRONGTYPE;
	}

	/* get the parameter's current value */
	*value = paramGetReal(param);

	return OKAY;
}

/** gets the value of an existing char parameter */
RETCODE Parameters::paramsetGetChar(const char* name, /**< name of the parameter */
char* value /**< pointer to store the parameter */
) {
	PARAM* param;

	assert(value != NULL);

	/* retrieve parameter from hash table */
	param = (PARAM*) hashtableRetrieve(name);
	if (param == NULL) {
		printf("parameter <%s> unknown\n", name);
		return PARAMETERUNKNOWN;
	}
	if (param->paramtype != PARAMTYPE_CHAR) {
		printf(
				"wrong parameter type - parameter <%s> has type <%s> instead of <%s>\n",
				name, paramtypeGetName(param->paramtype), paramtypeGetName(
						PARAMTYPE_CHAR));
		return PARAMETERWRONGTYPE;
	}

	/* get the parameter's current value */
	*value = paramGetChar(param);

	return OKAY;
}

/** gets the value of an existing string parameter */
RETCODE Parameters::paramsetGetString(const char* name, /**< name of the parameter */
char** value /**< pointer to store the parameter */
) {
	PARAM* param;

	assert(value != NULL);

	/* retrieve parameter from hash table */
	param = (PARAM*) hashtableRetrieve(name);
	if (param == NULL) {
		printf("parameter <%s> unknown\n", name);
		return PARAMETERUNKNOWN;
	}
	if (param->paramtype != PARAMTYPE_STRING) {
		printf(
				"wrong parameter type - parameter <%s> has type <%s> instead of <%s>\n",
				name, paramtypeGetName(param->paramtype), paramtypeGetName(
						PARAMTYPE_STRING));
		return PARAMETERWRONGTYPE;
	}

	/* get the parameter's current value */
	*value = paramGetString(param);

	return OKAY;
}

/** installs default values for all parameters */
RETCODE Parameters::paramsetSetToDefault() {
	std::map<const char *, PARAM*, strCmp>::iterator iter;
	for (iter = paramset.begin(); iter != paramset.end(); ++iter) {
		CALL( paramSetToDefault(iter->second) );
	}

	return OKAY;
}

/** sets the parameter to its default setting */
RETCODE Parameters::paramSetToDefault(PARAM* param /**< parameter */
) {
	assert(param != NULL);

	switch (param->paramtype) {
	case PARAMTYPE_BOOL:
		CALL( paramSetBool(param, paramGetBoolDefault(param)) );
		break;

	case PARAMTYPE_INT:
		CALL( paramSetInt(param, paramGetIntDefault(param)) );
		break;

	case PARAMTYPE_LONGINT:
		CALL( paramSetLongint(param, paramGetLongintDefault(param)) );
		break;

	case PARAMTYPE_REAL:
		CALL( paramSetReal(param, paramGetRealDefault(param)) );
		break;

	case PARAMTYPE_CHAR:
		CALL( paramSetChar(param, paramGetCharDefault(param)) );
		break;

	case PARAMTYPE_STRING:
		CALL( paramSetString(param, paramGetStringDefault(param)) );
		break;

	default:
		fprintf(stderr, "unknown parameter type\n");
		return INVALIDDATA;
	}

	return OKAY;
}

/** changes the value of an existing Bool parameter */
RETCODE Parameters::paramsetSetBool(const char* name, /**< name of the parameter */
Bool value /**< new value of the parameter */
) {
	PARAM* param;

	/* retrieve parameter from hash table */
	param = (PARAM*) hashtableRetrieve(name);
	if (param == NULL) {
		printf("parameter <%s> unknown\n", name);
		return PARAMETERUNKNOWN;
	}
	if (param->paramtype != PARAMTYPE_BOOL) {
		printf(
				"wrong parameter type - parameter <%s> has type <%s> instead of <%s>\n",
				name, paramtypeGetName(param->paramtype), paramtypeGetName(
						PARAMTYPE_BOOL));
		return PARAMETERWRONGTYPE;
	}
	/* set the parameter's current value */
	CALL( paramSetBool(param, value) );

	return OKAY;
}
/** sets value of Bool parameter */
RETCODE Parameters::paramSetBool(PARAM* param, Bool value /**< new value of the parameter */
) {
	assert(param != NULL);

	/* check, if value is possible for the parameter */
	//_CALL_QUIET( paramCheckBool(param, value) );

	/* set the parameter's current value */
	if (param->data.boolparam.valueptr != NULL)
		*param->data.boolparam.valueptr = value;
	else
		param->data.boolparam.curvalue = value;

	return OKAY;
}
/** changes the value of an existing int parameter */
RETCODE Parameters::paramsetSetInt(const char* name, /**< name of the parameter */
int value /**< new value of the parameter */
) {
	PARAM* param;

	/* retrieve parameter from hash table */
	param = (PARAM*) hashtableRetrieve(name);
	if (param == NULL) {
		printf("parameter <%s> unknown\n", name);
		return PARAMETERUNKNOWN;
	}
	if (param->paramtype != PARAMTYPE_INT) {
		printf(
				"wrong parameter type - parameter <%s> has type <%s> instead of <%s>\n",
				name, paramtypeGetName(param->paramtype), paramtypeGetName(
						PARAMTYPE_INT));
		return PARAMETERWRONGTYPE;
	}

	/* set the parameter's current value */
	CALL( paramSetInt(param, value) );

	return OKAY;
}

/** changes the value of an existing Longint parameter */
RETCODE Parameters::paramsetSetLongint(const char* name, /**< name of the parameter */
Longint value /**< new value of the parameter */
) {
	PARAM* param;

	/* retrieve parameter from hash table */
	param = (PARAM*) hashtableRetrieve(name);
	if (param == NULL) {
		printf("parameter <%s> unknown\n", name);
		return PARAMETERUNKNOWN;
	}
	if (param->paramtype != PARAMTYPE_LONGINT) {
		printf(
				"wrong parameter type - parameter <%s> has type <%s> instead of <%s>\n",
				name, paramtypeGetName(param->paramtype), paramtypeGetName(
						PARAMTYPE_LONGINT));
		return PARAMETERWRONGTYPE;
	}

	/* set the parameter's current value */
	CALL( paramSetLongint(param, value) );

	return OKAY;
}

/** changes the value of an existing Real parameter */
RETCODE Parameters::paramsetSetReal(const char* name, /**< name of the parameter */
Real value /**< new value of the parameter */
) {
	PARAM* param;

	/* retrieve parameter from hash table */
	param = (PARAM*) hashtableRetrieve(name);
	if (param == NULL) {
		printf("parameter <%s> unknown\n", name);
		return PARAMETERUNKNOWN;
	}
	if (param->paramtype != PARAMTYPE_REAL) {
		printf(
				"wrong parameter type - parameter <%s> has type <%s> instead of <%s>\n",
				name, paramtypeGetName(param->paramtype), paramtypeGetName(
						PARAMTYPE_REAL));
		return PARAMETERWRONGTYPE;
	}

	/* set the parameter's current value */
	CALL( paramSetReal(param, value) );

	return OKAY;
}

/** changes the value of an existing char parameter */
RETCODE Parameters::paramsetSetChar(const char* name, /**< name of the parameter */
char value /**< new value of the parameter */
) {
	PARAM* param;
	/* retrieve parameter from hash table */
	param = (PARAM*) hashtableRetrieve(name);
	if (param == NULL) {
		printf("parameter <%s> unknown\n", name);
		return PARAMETERUNKNOWN;
	}
	if (param->paramtype != PARAMTYPE_CHAR) {
		fprintf(
				stderr,
				"wrong parameter type - parameter <%s> has type <%s> instead of <%s>\n",
				name, paramtypeGetName(param->paramtype), paramtypeGetName(
						PARAMTYPE_CHAR));
		return PARAMETERWRONGTYPE;
	}

	/* set the parameter's current value */
	paramSetChar(param, value);

	return OKAY;
}

/** changes the value of an existing string parameter */
RETCODE Parameters::paramsetSetString(const char* name, /**< name of the parameter */
const char* value /**< new value of the parameter */
) {
	PARAM* param;

	/* retrieve parameter from hash table */
	param = (PARAM*) hashtableRetrieve(name);
	if (param == NULL) {
		printf("parameter <%s> unknown\n", name);
		return PARAMETERUNKNOWN;
	}
	if (param->paramtype != PARAMTYPE_STRING) {
		printf(
				"wrong parameter type - parameter <%s> has type <%s> instead of <%s>\n",
				name, paramtypeGetName(param->paramtype), paramtypeGetName(
						PARAMTYPE_STRING));
		return PARAMETERWRONGTYPE;
	}

	/* set the parameter's current value */
	CALL( paramSetString(param, value) );

	return OKAY;
}

/** parses a parameter file line "paramname = paramvalue" and sets parameter accordingly */
RETCODE Parameters::paramsetParse(char* line /**< line to parse (is modified during parse, but not freed) */
) {
	PARAM* param;
	char* paramname;
	char* paramvaluestr;
	char* lastquote;
	Bool quoted;

	assert(line != NULL);

	/* find the start of the parameter name */
	while (*line == ' ' || *line == '\t' || *line == '\r')
		line++;
	if (*line == '\0' || *line == '\n' || *line == '#')
		return OKAY;
	paramname = line;

	/* find the end of the parameter name */
	while (*line != ' ' && *line != '\t' && *line != '\r' && *line != '\n'
			&& *line != '#' && *line != '\0' && *line != '=')
		line++;
	if (*line == '=') {
		*line = '\0';
		line++;
	} else {
		*line = '\0';
		line++;

		/* search for the '=' char in the line */
		while (*line == ' ' || *line == '\t' || *line == '\r')
			line++;
		if (*line != '=') {
			printf("character '=' was expected after the parameter name\n");
			return PARSEERROR;
		}
		line++;
	}

	/* find the start of the parameter value string */
	while (*line == ' ' || *line == '\t' || *line == '\r')
		line++;
	if (*line == '\0' || *line == '\n' || *line == '#') {
		printf("parameter value is missing\n");
		return PARSEERROR;
	}
	paramvaluestr = line;

	/* find the end of the parameter value string */
	quoted = (*paramvaluestr == '"');
	lastquote = NULL;
	while ((quoted || (*line != ' ' && *line != '\t' && *line != '\r' && *line
			!= '\n' && *line != '#')) && *line != '\0') {
		if (*line == '"')
			lastquote = line;
		line++;
	}
	if (lastquote != NULL)
		line = lastquote + 1;
	if (*line == '#')
		*line = '\0';
	else if (*line != '\0') {
		/* check, if the rest of the line is clean */
		*line = '\0';
		line++;
		while (*line == ' ' || *line == '\t' || *line == '\r')
			line++;
		if (*line != '\0' && *line != '\n' && *line != '#') {
			printf("additional characters after parameter value\n");
			return PARSEERROR;
		}
	}

	/* retrieve parameter from hash table */
	param = (PARAM*) hashtableRetrieve(paramname);
	if (param == NULL) {
		printf("unknown parameter <%s>\n", paramname);
		return OKAY;
	}

	/* set parameter's value */
	switch (param->paramtype) {
	case PARAMTYPE_BOOL:
		paramParseBool(param, paramvaluestr);
		break;
	case PARAMTYPE_INT:
		CALL( paramParseInt(param, paramvaluestr) );
		break;
	case PARAMTYPE_LONGINT:
		CALL( paramParseLongint(param, paramvaluestr) );
		break;
	case PARAMTYPE_REAL:
		CALL( paramParseReal(param, paramvaluestr) );
		break;
	case PARAMTYPE_CHAR:
		CALL( paramParseChar(param, paramvaluestr) );
		break;
	case PARAMTYPE_STRING:
		CALL( paramParseString(param, paramvaluestr) );
		break;
	default:
		printf("unknown parameter type\n");
		return INVALIDDATA;
	}

	return OKAY;
}

/** reads parameters from a file */
void Parameters::paramsetRead(
//paramset*        paramset,           /**< parameter set */
		//set*             set,                /**< global  settings */
		const char* filename /**< file name */
) {
	std::ifstream is;
	char line[1024];
	int lineno;
	RETCODE retcode;

	//assert(paramset != NULL);
	assert(filename != NULL);

	/* open the file for reading */
	is.open(filename);
	if (!is.is_open()) {
		fprintf(stderr, "cannot open file <%s> for reading\n", filename);
		exit(1);
	}

	/*read the parameters from the file */
	lineno = 0;
	retcode = OKAY;
	while (is.getline(line, sizeof(line)) && retcode == OKAY) {
		lineno++;
		retcode = paramsetParse(//paramset, set,
				line);
	}

	/* close input file */
	is.close();

	if (retcode == PARSEERROR) {
		printf("input error in file <%s> line %d\n", filename, lineno);
	}
}

/** writes all parameters in the parameter set to a file */
RETCODE Parameters::paramsetWrite(const char* filename, /**< file name, or NULL for stdout */
bool comments, /**< should parameter descriptions be written as comments? */
bool onlychanged /**< should only the parameters been written, that are changed from default? */
) {
	std::ofstream file;

	//assert(paramset != NULL);

	/* open the file for writing */
	if (filename != NULL) {
		file.open(filename);
		if (!file.is_open()) {
			printf("cannot open file <%s> for writing\n", filename);
			exit(1);
		}
	}

	/*write the parameters to the file */
	std::map<const char *, PARAM*, strCmp>::iterator iter;
	for (iter = paramset.begin(); iter != paramset.end(); ++iter) {
		if (filename != NULL)
			paramWrite(iter->second, file, comments, onlychanged);
		else
		  paramWrite(iter->second, std::cout, comments, onlychanged);
	}

	/* close output file */
	if (filename != NULL) {
		assert(file.is_open());
		file.close();
	}

	return OKAY;
}

/** writes the parameter to a file */
RETCODE Parameters::paramWrite(PARAM* param, /**< parameter */
std::ostream &file, /**< file stream to write parameter to, or NULL for stdout  */
Bool comments, /**< should parameter descriptions be written as comments? */
Bool onlychanged /**< should only the parameters been written, that are changed from default? */
) {
	char buf[MAXSTRLEN];
	assert(param != NULL);

	/* write parameters at default values only, if the onlychanged flag is not set */
	if (onlychanged && paramIsDefault(param))
		return OKAY;

	/* write parameter description, bounds, and defaults as comments */
	if (comments) {
		file << "# " << param->desc << std::endl;
		switch (param->paramtype) {
		case PARAMTYPE_BOOL:
			sprintf(buf, "# [type: bool, range: {TRUE,FALSE}, default: %s]\n",
					param->data.boolparam.defaultvalue ? "TRUE" : "FALSE");
			file << buf;
			break;
		case PARAMTYPE_INT:
			sprintf(buf, "# [type: int, range: [%d,%d], default: %d]\n",
					param->data.intparam.minvalue,
					param->data.intparam.maxvalue,
				param->data.intparam.defaultvalue);
			file << buf;
			break;
		case PARAMTYPE_LONGINT:
			sprintf(
					buf,
					"# [type: longint, range: [%"LONGINT_FORMAT",%"LONGINT_FORMAT"], default: %"LONGINT_FORMAT"]\n",
					param->data.longintparam.minvalue,
					param->data.longintparam.maxvalue,
					param->data.longintparam.defaultvalue);
			file << buf;
			break;
		case PARAMTYPE_REAL:
			sprintf(buf,
					"# [type: real, range: [%.15g,%.15g], default: %.15g]\n",
					param->data.realparam.minvalue,
					param->data.realparam.maxvalue,
					param->data.realparam.defaultvalue);
			file << buf;
			break;
		case PARAMTYPE_CHAR:
			sprintf(
					buf,
					"# [type: char, range: {%s}, default: %c]\n",
					param->data.charparam.allowedvalues != NULL ? param->data.charparam.allowedvalues
							: "all chars", param->data.charparam.defaultvalue);
			file << buf;
			break;
		case PARAMTYPE_STRING:
			sprintf(buf, "# [type: string, default: \"%s\"]\n",
					param->data.stringparam.defaultvalue);
			file << buf;
			break;
		default:
			printf("unknown parameter type\n");
			return INVALIDDATA;
		}
	}

	/* write parameter value */
	sprintf(buf, "%s = ", param->name);
	file << buf;
	switch (param->paramtype) {
	case PARAMTYPE_BOOL:
		sprintf(buf, "%s\n", paramGetBool(param) ? "TRUE" : "FALSE");
		file << buf;
		break;
	case PARAMTYPE_INT:
		sprintf(buf, "%d\n", paramGetInt(param));
		file << buf;
		break;
	case PARAMTYPE_LONGINT:
		sprintf(buf, "%"LONGINT_FORMAT"\n", paramGetLongint(param));
		file << buf;
		break;
	case PARAMTYPE_REAL:
		sprintf(buf, "%.15g\n", paramGetReal(param));
		file << buf;
		break;
	case PARAMTYPE_CHAR:
		sprintf(buf, "%c\n", paramGetChar(param));
		file << buf;
		break;
	case PARAMTYPE_STRING:
		sprintf(buf, "\"%s\"\n", paramGetString(param));
		file << buf;
		break;
	default:
		printf("unknown parameter type\n");
		return INVALIDDATA;
	}

	if (comments) {
		sprintf(buf, "\n");
		file << buf;
	}

	return OKAY;
}
