/*
 * PartialCol.cpp
 *
 *  Created on: Aug 14, 2010
 *      Author: marco
 */

#include <iostream>
#include "Runners.h"
#include "Explorer.h"

using namespace std;

/*
 * I. Blchliger, N. Zufferey (2008). A graph coloring heuristic using partial solutions and a reactive tabu scheme. Computers & Operations Research, vol. 35, no. 3, pp. 960-975
 */
void ColoringPartialCol::Run(PartialColoring &initial_ass) {
	p_ctrl->getOutputStream() << "PartialCol\n";
	InitializeRun();
	PartialColoring current_state(initial_ass);
	ColMove mv;
	bool tl_created = false;

	if (p_tl == NULL) {
		p_tl = new TabuList();
		tl_created = true;
	}
	p_tl->Init(p_in->Vertices(), current_state.CurrColors(), rg, p_ctrl);

	while ((current_state.cost != 0) && !(MaxIterationExpired())
			&& (idle_iterations < max_idle_iteration) && p_ctrl->isTimeLeft()) {

		UpdateIterationCounter();
		p_ne->BestMove(current_state, mv, *p_tl, number_of_iterations,
				initial_ass.cost);

#ifdef TEST_RUNNER
		PartialColoring test_state(current_state);
		//cout << test_state<<endl;
		assert(test_state == current_state);
		cout << mv << endl;
#endif
		p_tl->UpdateTabuList(current_state, mv, number_of_iterations);
		p_ne->MakeMove(current_state, mv);

#ifdef TEST_RUNNER
		p_ne->MakeMove(test_state, mv);
		p_sm->updateColoringData(test_state);
		assert(test_state == current_state);
#endif
		if (current_state.cost != 0)
			current_state.feasible = false;
		else
			current_state.feasible = true;

		if (current_state.cost < initial_ass.cost) {
			initial_ass = current_state;
			p_ctrl->setCurrentCost(current_state.cost,
					current_state.CurrColors(), number_of_iterations);
			idle_iterations = 0;

		}
		idle_iterations++;

	}
	TerminateRun();
	if (tl_created) {
		delete p_tl;
		p_tl = NULL;
	}
	p_ctrl->getOutputStream() << "Done (PartialCol)\n";
}

//------------------------------------------------------------
//--------------------COLORING Neighboorhood------------------
//------------------------------------------------------------

void ColISwapExplorer::RandomMove(const PartialColoring& coloring,
		ColMove& rc) {
	rc.seq.clear();
	rc.seq.resize(2);
	int v;
	rc.type = I_SWAP;

	assert(coloring.impasse_class.size()!=0);

	v = rg->ran0N(coloring.impasse_class.size() - 1);
	VertexSet::const_iterator pt;
	pt=coloring.impasse_class.begin();
	for (int i = 0; i < v; i++)
		pt++;
	rc.seq[0] = *pt;

	do
		rc.seq[1] = (Color) rg->ran0N((coloring.CurrColors() - 1));
	while ( ((Color) rc.seq[1]) == coloring.col[rc.seq[0]]);
	rc.seq.push_back(coloring.col[rc.seq[0]]);

	// retrieve vertices adj to vertex to move to impasse class
	for (pt = p_in->adjacency_set[rc.seq[0]].begin(); pt
			!= p_in->adjacency_set[rc.seq[0]].end(); pt++) {
		if (coloring.col[*pt] == rc.seq[1]) {
			rc.seq.push_back(*pt);
		}
	}
	rc.seq.push_back(-1);

}

//For PartialCol
void ColISwapExplorer::BestMove(const PartialColoring& coloring, ColMove& rc,
		TabuList &t_l, unsigned long iter, fvalue best_so_far) {
	fvalue min = FVALUE_MAX;
	VertexSet::const_iterator pt;
	vector < pair<Vertex, Color> > candidates;
	fvalue delta;

	rc.seq.clear();
	rc.seq.resize(2);
	rc.type = I_SWAP;
	for (pt = coloring.impasse_class.begin(); pt
			!= coloring.impasse_class.end(); pt++) {
		for (Color j = 0; j < coloring.CurrColors(); j++) {
			delta = coloring.Delta(*pt, j);
			if (delta == min) {
				if ((!t_l.isTabu(*pt, j, iter)) || (coloring.cost + delta
						< best_so_far)) {
					pair < Vertex, Color > initPair;
					initPair.first = *pt;
					initPair.second = j;
					candidates.push_back(initPair);
				}
			} else if (delta < min) {
				if ((!t_l.isTabu(*pt, j, iter)) || (coloring.cost + delta
						< best_so_far)) {
					candidates.clear();
					pair < Vertex, Color > initPair;
					initPair.first = *pt;
					initPair.second = j;
					candidates.push_back(initPair);
					min = delta;
				}
			}
		}
	}

	if (candidates.size() > 0) {
		int rnd = rg->ran0N(candidates.size() - 1);
		rc.seq[0] = candidates[rnd].first;
		rc.seq[1] = candidates[rnd].second;
		rc.delta = min;
		//rc.seq.push_back(coloring.col[rc.seq[0]]);
	} else { // select one at random waiting for some tabu to expire
		size_t size = rg->ran0N(coloring.impasse_class.size() - 1);
		pt = coloring.impasse_class.begin();
		for (size_t i = 0; i < size; i++) {
			pt++;
		}
		rc.seq[0] = *pt;
		rc.seq[1] = rg->ran0N(coloring.CurrColors() - 1);
		rc.delta = coloring.Delta(rc.seq[0], rc.seq[1]);
	}
	Vertex node = rc.seq[0];
	Color col = rc.seq[1];

	// retrieve vertices adj to vertex to move to impasse class
	for (pt = p_in->adjacency_set[node].begin(); pt
			!= p_in->adjacency_set[node].end(); pt++) {
		if (coloring.col[*pt] == col) {
			rc.seq.push_back(*pt);
		}
	}
	rc.seq.push_back(-1);
}

void ColNeighborhoodExplorer::MakeMove(PartialColoring& coloring,
		const ColMove& rc) {
	if (rc.type == I_SWAP) {
		Vertex node = rc.seq[0];
		Color col = rc.seq[1];
		// put vertex in class
		coloring.col[node] = col;
		coloring.impasse_class.erase(node);
		coloring.partitions[col].insert(node);
		p_sm->updateDeltaDataISwap(coloring, rc.seq[0], NOT_A_COLOR, rc.seq[1]);
		// move vertices adj to vertex to impasse class
		int i = 2;
		while (rc.seq[i] != -1) {
			coloring.col[rc.seq[i]] = NOT_A_COLOR;//coloring.CurrColors();
			coloring.partitions[col].erase(rc.seq[i]);
			coloring.impasse_class.insert(rc.seq[i]);
			p_sm->updateDeltaDataISwap(coloring, rc.seq[i], col, NOT_A_COLOR);
			i++;
		}

		coloring.cost += rc.delta;
		if (coloring.cost != 0)
			coloring.feasible = false;
		else
			coloring.feasible = true;
	} else {
		cerr << rc;
		cerr << "Wrong move in Explorer.cpp\n";
		exit(EXIT_FAILURE);
	}
}
