/* GCP version 3.5 Copyright (C) 2003 Marco Chiarandini  
 Technische Universitaet Darmstadt
 email: machud@intellektik.informatik.tu-darmstadt.de

 This file is part of GCP.

 GCP is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.
 
 Foobar is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with GCP; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef PROBLEM_H
#define PROBLEM_H

#include<cstdlib>
#include <vector>
#include <map>
#include <iostream>
#include <string>
#include "Definitions.h"
#include "Utilities.h"
#include "bin2asc.h"

using namespace std;

// Input class

class Graph {
	friend std::ostream& operator<<(std::ostream&, const Graph&);
public:
	Graph();
	/* this method loads a DIMACS coded graph coloring instance */
	void Load(const char *id, bool print_statistics);
	unsigned int Vertices() const {
		return num_of_vertices;
	}
	unsigned int Edges() const {
		return num_of_edges;
	}
	virtual ~Graph();
	bool Edge(const Vertex& v, const Vertex& w) const {
		return adjacency_matrix[v][w] != 0;
	}
	float Density() const {
		return density;
	}
	float AvDegree() const {
		return av_degree;
	}

	void detectDisconnectivity(Bool force_if_disconnected);

	vector<VertexSet> adjacency_set;
	vector<Vertex> degree;
	//virtual void Preprocess(const Color lower_bound) {};

	vector<pair<Vertex, Vertex> > edges;
protected:
	Vertex num_of_vertices;
	uint32_t num_of_edges;
	uint32_t declared_edges;
	Vertex max_degree;
	Vertex min_degree;

	float density;
	float av_degree;
	unsigned char **adjacency_matrix;
private:
	void read_DIMACS_line(char *buf);
};

#endif

