/* GCP version 3.5 Copyright (C) 2003 Marco Chiarandini
 Technische Universitaet Darmstadt
 email: machud@intellektik.informatik.tu-darmstadt.de

 This file is part of GCP.

 GCP is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 Foobar is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with GCP; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <algorithm>
#include <cmath>
#include <iostream>
#include <sstream>
#include <list>

#include "Solution.h"
#include "Manager.h"
#include "Move.h"


void ColoringManager::constructiveRecursiveLargestFirst(Coloring& ass, Color max_col, int representation) {
	p_ctrl->getOutputStream() << "ConstructiveRecursiveLargestFirst... ";
	Color num_col = 0;
	Vertex v;

	list<Vertex> U, V;
	list<Vertex>::iterator pt, vpt;

	VertexSet::iterator p;

	vector<Vertex> degree(p_in->Vertices());
	vector<Vertex> degree_to_U(p_in->Vertices(), 0);

	V.clear();
	U.clear();

	array = rg->generate_array(p_in->Vertices());
	for (Vertex i = 0; i < p_in->Vertices(); i++) {
		v = array[i];
		V.push_back(v);
		ass.col[v] = -1;
		degree[v] = p_in->degree[v];
	}

	do {

		while (!V.empty()) {
			pt = vpt = V.begin();
			v = *vpt;
			pt++;

			if (U.empty()) //pick max degree vertex
			{
				while (pt != V.end()) {
					if (degree[*pt] > degree[v]) {
						v = *pt;
						vpt = pt;
					}
					pt++;
				}
			} else {
				//pick max degree to U; break ties in favour of min overall degree
				while (pt != V.end()) {
					if (degree_to_U[*pt] > degree_to_U[v]) {
						v = *pt;
						vpt = pt;
					} else if ((degree_to_U[v] == degree_to_U[*pt])
							&& (degree[v] > degree[*pt])) {
						v = *pt;
						vpt = pt;
					}
					pt++;
				}
			}

			ass.col[v] = num_col;
			V.erase(vpt);

			//move all neighbors of v from V to U
			//and update degrees data
			//for (p = p_in->adjacency_set[v].begin(); p != p_in->adjacency_set[v].end(); p++)

			pt = V.begin();
			while (pt != V.end()) {

				if (p_in->Edge(v, *pt)) {
					if (rg->next() > 0.5)
						U.push_back(*pt);
					else
						U.push_front(*pt);

					degree[*pt]--;
					ass.col[*pt] = -2;
					for (p = p_in->adjacency_set[*pt].begin(); p
							!= p_in->adjacency_set[*pt].end(); p++) {
						if (ass.col[*p] == -1)
							degree_to_U[*p]++;
					}
					pt = V.erase(pt);
				} else {
					pt++;
				}
			}
		}
		num_col++;

		for (pt = U.begin(); pt != U.end(); pt++) {
			ass.col[*pt] = -1;
			V.push_back(*pt);
			degree_to_U[*pt] = 0;
		}
		U.clear();

	} while (!V.empty() && num_col < max_col);

	for (pt = V.begin(); pt != V.end(); pt++) {
		if (representation==COMPLETE_REPR)
			ass.col[*pt] = rg->ran0N(num_col - 1);
		else if (representation==PARTIAL_REPR)
			ass.col[*pt] = NOT_A_COLOR;
	}

	delete[] array;

	ass.SetColors(num_col);
	updateColoringData(ass,representation);
	p_ctrl->getOutputStream() << "done\n";
}

void ColoringManager::constructiveRecursiveLargestFirstReverse(Coloring& ass, Color max_col, int representation) {
	Color num_col = 0;
	Vertex v;

	//set<Vertex,compare> V;
	//set<Vertex,compare> U;

	list<Vertex> U, V;
	list<Vertex>::iterator pt, vpt;

	VertexSet::iterator p;
	//set<Vertex,compare>::iterator pt;
	//  set<Vertex,compare>::iterator z;

	vector<Vertex> degree(p_in->Vertices());
	vector<Vertex> degree_to_U(p_in->Vertices(), 0);

	V.clear();
	U.clear();

	array = rg->generate_array(p_in->Vertices());
	for (Vertex i = 0; i < p_in->Vertices(); i++) {
		v = array[i];
		V.push_back(v);
		ass.col[v] = -1;
		degree[v] = p_in->degree[v];
	}
	//copy(V.begin(), V.end(), ostream_iterator<Vertex>(cout, " "));
	//
	do {

		//      V = U;
		//U.clear();

		while (!V.empty()) {
			pt = vpt = V.begin();
			v = *vpt;
			pt++;

			if (U.empty()) //pick max degree vertex
			{
				while (pt != V.end()) {
					if (degree[*pt] < degree[v]) {
						v = *pt;
						vpt = pt;
					}
					pt++;
				}
			} else {//pick max degree to U; break ties in favour of min overall degree
				while (pt != V.end()) {
					if (degree_to_U[*pt] < degree_to_U[v]) {
						v = *pt;
						vpt = pt;
					} else if ((degree_to_U[v] == degree_to_U[*pt])
							&& (degree[v] < degree[*pt])) {
						v = *pt;
						vpt = pt;
					}
					pt++;
				}
			}

			ass.col[v] = num_col;
			V.erase(vpt);

			//move all neighbors of v from V to U
			//and update degrees data
			//for (p = p_in->adjacency_set[v].begin(); p != p_in->adjacency_set[v].end(); p++)
			pt = V.begin();
			while (pt != V.end()) {

				if (p_in->Edge(v, *pt)) {
					if (rg->next() > 0.5)
						U.push_back(*pt);
					else
						U.push_front(*pt);

					degree[*pt]--;
					ass.col[*pt] = -2;
					for (p = p_in->adjacency_set[*pt].begin(); p
							!= p_in->adjacency_set[*pt].end(); p++) {
						if (ass.col[*p] == -1)
							degree_to_U[*p]++;
					}
					pt = V.erase(pt);
				} else {
					pt++;
				}
			}
		}
		num_col++;

		for (pt = U.begin(); pt != U.end(); pt++) {
			ass.col[*pt] = -1;
			V.push_back(*pt);
			degree_to_U[*pt] = 0;
		}
		U.clear();

	} while (!V.empty() && (num_col < max_col));

	for (pt = V.begin(); pt != V.end(); pt++) {
		if (representation==COMPLETE_REPR)
			ass.col[*pt] = rg->ran0N(num_col - 1);
		else if (representation==PARTIAL_REPR)
			ass.col[*pt] = NOT_A_COLOR;
	}

	delete[] array;

	ass.SetColors(num_col);
	updateColoringData(ass,representation);
}


//as decribed in the paper by Walter Klotz. it works in O(V3) but gets
//worse results
void ColoringManager::constructiveRecursiveLargestFirstStatic(Coloring& ass, Color max_col, int representation) {
	Color num_col = 0;
	Vertex v, last_v = 0;

	VertexSet::iterator p;

	list<Vertex> U, V;
	list<Vertex>::iterator pt, vpt;

	V.clear();
	U.clear();

	array = rg->generate_array(p_in->Vertices());

	Vertex **com_adj_in_V;
	com_adj_in_V = new Vertex *[p_in->Vertices()];
	for (Vertex i = 0; i < p_in->Vertices(); i++) {
		com_adj_in_V[i] = new Vertex[p_in->Vertices()];
		v = array[i];
		V.push_back(v);
		ass.col[v] = -1;
	}

	for (Vertex i = 0; i < p_in->Vertices(); i++) {
		for (Vertex j = i + 1; j < p_in->Vertices(); j++) {
			com_adj_in_V[i][j] = 0;
			for (p = p_in->adjacency_set[i].begin(); p
					!= p_in->adjacency_set[i].end(); p++) {
				if (p_in->Edge(*p, j)) {
					com_adj_in_V[i][j]++;

				}
			}
			com_adj_in_V[j][i] = com_adj_in_V[i][j];
		}
	}

	do {

		while (!V.empty()) {
			//determine a vertex v of maximal degree in V
			pt = vpt = V.begin();
			v = *vpt;
			pt++;

			if (U.empty()) {
				while (pt != V.end()) {
					if (p_in->degree[*pt] > p_in->degree[v]) {
						v = *pt;
						vpt = pt;
					}
					pt++;
				}
			} else {
				while (pt != V.end()) {
					if (com_adj_in_V[last_v][*pt] > com_adj_in_V[last_v][v]) {
						v = *pt;
						vpt = pt;
					} else if (com_adj_in_V[last_v][*pt]
							== com_adj_in_V[last_v][v] && p_in->degree[*pt]
							< p_in->degree[v]) {
						v = *pt;
						vpt = pt;
					}
					pt++;
				}

				if (com_adj_in_V[last_v][v] == 0) //the case in which V is unconnected
				{
					pt = vpt = V.begin();
					v = *vpt;
					pt++;
					while (pt != V.end()) {
						if (p_in->degree[*pt] > p_in->degree[v]) {
							v = *pt;
							vpt = pt;
						}
						pt++;
					}
				}
			}
			//put v into colour class
			ass.col[v] = num_col;
			V.erase(vpt);
			//V set of non neighbohrs of v
			last_v = v;
			pt = V.begin();
			while (pt != V.end()) {
				if (p_in->Edge(v, *pt)) {
					if (rg->next() > 0.5)
						U.push_back(*pt);
					else
						U.push_front(*pt);
					pt = V.erase(pt);
				} else {
					pt++;
				}
			}

		}

		V = U;
		U.clear();
		num_col++;
	} while (!V.empty() && (num_col < max_col));

	for (pt = V.begin(); pt != V.end(); pt++) {
		if (representation==COMPLETE_REPR)
			ass.col[*pt] = rg->ran0N(num_col - 1);
		else if (representation==PARTIAL_REPR)
			ass.col[*pt] = NOT_A_COLOR;
	}

	delete[] array;
	for (Vertex i = 0; i < p_in->Vertices(); i++)
		delete[] com_adj_in_V[i];
	delete[] com_adj_in_V;

	ass.SetColors(num_col);
	updateColoringData(ass,representation);
}

