/* GCP version 6.0 Copyright (C) 2003 Marco Chiarandini  
 Technische Universitaet Darmstadt
 email: machud@intellektik.informatik.tu-darmstadt.de

 This file is part of GCP.

 GCP is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.
 
 Foobar is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with GCP; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "Runners.h"

#include <algorithm>
#include <cmath>
#include <iostream>
using namespace std;

//------------------------------------------------------------
//--------------------COLORING RUNNERS------------------------
//------------------------------------------------------------
ColoringRunner::ColoringRunner(ColoringManager* s, Random *rnd, Control *ctrl,
		Graph* in = NULL) {
	p_sm = s;
	p_in = in;
	rg = rnd;
	p_ctrl = ctrl;
	number_of_iterations = 0;

}

ColoringRunner::ColoringRunner(ColoringManager* s, ColNeighborhoodExplorer* e,
		Random *rnd, Control *ctrl, Graph* in = NULL) {
	p_sm = s;
	p_ne = e;
	p_in = in;
	rg = rnd;
	p_ctrl = ctrl;
	number_of_iterations = 0;

}

void ColoringRunner::InitializeRun() {
	idle_iterations = 0;
	number_of_iterations = 0;
}

void ColoringRunner::initialization_procedure(Coloring &ass, int representation) {

	switch (p_ctrl->init) {
	case RANDOM:
		p_sm->randomAssignment(ass, p_ctrl->k);
		break;
	case ROS:
		p_sm->constructiveGreedy(ass, p_in->Vertices(), representation);
		break;
	case ROS_FIXED:
		p_sm->constructiveGreedy(ass, p_ctrl->k, representation);
		break;
	case RLF:
		p_sm->constructiveRecursiveLargestFirst(ass, p_in->Vertices(),
				representation);
		break;
	case RLF_FIXED:
		//cout<<p_ctrl->k<<" "<<p_ctrl->init <<endl;
		p_sm->constructiveRecursiveLargestFirst(ass, p_ctrl->k,
				representation);
		break;
	case DSATUR:
		p_sm->constructiveDSATUR(ass, p_in->Vertices(), representation);
		break;
	case DSATUR_FIXED:
		p_sm->constructiveDSATUR(ass, p_ctrl->k, representation);
		break;
	case FROM_INPUT_FILE:
		p_sm->readStartingSolution(ass, p_ctrl->k);
		break;
	case RLF_STATIC:
		// cout<<p_ctrl->k<<" "<<p_ctrl->init <<endl;
		p_sm->constructiveRecursiveLargestFirstStatic(ass,p_in->Vertices(), COMPLETE_REPR);
		break;
	case RLF_REVERSE:
		p_sm->constructiveRecursiveLargestFirstReverse(ass,p_in->Vertices(), COMPLETE_REPR);
		break;
	default:
		cerr << "Errors in initialization\n";
		exit(1);
	}
}

void ColoringRunner::decreasing_procedure(Coloring &ass,
		int representation) {
	switch (p_ctrl->decr) {
	case RANDOM:
		p_sm->decrease_colors(ass, representation);
		break;
	case ROS_FIXED:
		ass.SetColors(ass.CurrColors() - 1);
		p_sm->constructiveGreedy(ass, ass.CurrColors(), representation);
		break;
	case RLF_FIXED:
		ass.SetColors(ass.CurrColors() - 1);
		p_sm->constructiveRecursiveLargestFirst(ass, ass.CurrColors(),
				representation);
		break;
	case DSATUR_FIXED:
		ass.SetColors(ass.CurrColors() - 1);
		p_sm->constructiveDSATUR(ass, ass.CurrColors(), representation);
		break;
	case 10:
		ass.SetColors(ass.CurrColors() - 1);
		p_sm->randomAssignment(ass, ass.CurrColors());
		break;
	default:
		cerr << "something wrong!!\n";
		exit(1);
	}
}

void ColoringRunner::Run(Coloring &initial_ass) {
	p_ctrl->getOutputStream() << "Enter: Runner::Run\n";
	InitializeRun();
	Coloring current_state(initial_ass);
	ColMove mv;

	while ((current_state.cost != 0) && !(MaxIterationExpired())
			&& (idle_iterations < max_idle_iteration) && p_ctrl->isTimeLeft()) {
		p_ne->BestMove(current_state, mv);

		if (mv.type > 0 && mv.delta < 0) {
#ifdef TEST_RUNNER
			Coloring test_state(current_state);

			if (!(test_state == current_state)) {
				cerr << "before" << endl;
				exit(1);
			}

			cout << mv << endl;

#endif
			p_ne->MakeMove(current_state, mv);

#ifdef TEST_RUNNER
			p_ne->MakeMove(test_state, mv);
			p_sm->updateColoringData(test_state);
			//cout << test_state<<endl;
			if (!(test_state == current_state)) {
				exit(1);
			}

			//cout << current_state;
#endif
			if (current_state.cost != 0)
				current_state.feasible = false;
			else
				current_state.feasible = true;

			if (current_state.cost < initial_ass.cost) {
				initial_ass = current_state;
				p_ctrl->setCurrentCost(current_state.cost,
						current_state.CurrColors(), number_of_iterations);
				idle_iterations = 0;
			}

		} else
			break;
		idle_iterations++;
		UpdateIterationCounter();
	}
	TerminateRun();
}

void ColoringRunnerUpHill::Run(Coloring &initial_ass) {
	p_ctrl->getOutputStream() << "Enter: UpHill)\n";
	InitializeRun();
	Coloring current_state(initial_ass);
	ColMove mv;

	TabuList tl(p_in->Vertices(), current_state.CurrColors(), rg, p_ctrl);

	while ((current_state.cost != 0) && !(MaxIterationExpired())
			&& (idle_iterations < max_idle_iteration) && p_ctrl->isTimeLeft()) {
		UpdateIterationCounter();
		p_ne->BestMove(current_state, mv);
#ifdef TEST_RUNNER
		Coloring test_state(current_state);

		if (!(test_state == current_state)) {
			cerr << "before" << endl;
			exit(1);
		}
		cout << current_state.cost << " " << current_state.CurrColors() << endl;
		cout << mv << endl;
#endif
		if (mv.type > 0) {
			p_ne->MakeMove(current_state, mv);
#ifdef TEST_RUNNER
			p_ne->MakeMove(test_state, mv);
#endif
		} else
			break;
#ifdef TEST_RUNNER
		p_sm->updateColoringData(test_state);
		if (!(test_state == current_state)) {
			exit(1);
		}
#endif
		if (current_state.cost != 0)
			current_state.feasible = false;
		else
			current_state.feasible = true;

		if (current_state.cost < initial_ass.cost) {

			initial_ass = current_state;
			p_ctrl->setCurrentCost(current_state.cost,
					current_state.CurrColors(), number_of_iterations);
			idle_iterations = 0;

		}
		idle_iterations++;
	}
	TerminateRun();
}

void ColoringTabuSearchWithoutAsp::Run(CompleteColoring &initial_ass) {
	p_ctrl->getOutputStream() << "TabuSearchWithoutAsp\n";
	InitializeRun();
	CompleteColoring current_state(initial_ass);
	ColMove mv;

	TabuList tl(p_in->Vertices(), current_state.CurrColors(), rg, p_ctrl);

	while ((current_state.cost != 0) && !(MaxIterationExpired())
			&& (idle_iterations < max_idle_iteration) && p_ctrl->isTimeLeft()) {
		UpdateIterationCounter();
		p_ne->BestMove(current_state, mv, tl, number_of_iterations);
#ifdef TEST_RUNNER
		CompleteColoring test_state(current_state);

		if (!(test_state == current_state)) {
			cerr << "before" << endl;
			exit(1);
		}
		cout << current_state.cost << " " << current_state.CurrColors() << endl;
		cout << mv << endl;
#endif
		tl.UpdateTabuList(current_state, mv, number_of_iterations);
		p_ne->MakeMove(current_state, mv);
#ifdef TEST_RUNNER
		p_ne->MakeMove(test_state, mv);
		//test_state.col[mv.seq[0]] = mv.seq[1];

		//	  current_state.cost += mv.delta;

		p_sm->updateColoringData(test_state);
		//cout << test_state<<endl;
		if (!(test_state == current_state)) {
			exit(1);
		}

		//cout << current_state;
#endif
		if (current_state.cost != 0)
			current_state.feasible = false;
		else
			current_state.feasible = true;

		if (current_state.cost < initial_ass.cost) {

			initial_ass = current_state;
			p_ctrl->setCurrentCost(current_state.cost,
					current_state.CurrColors(), number_of_iterations);
			idle_iterations = 0;

		}
		idle_iterations++;
	}
	TerminateRun();
	p_ctrl->getOutputStream() << "Done (TabuSearchWithoutAsp)\n";
}

