
/* GCP version 3.5 Copyright (C) 2003 Marco Chiarandini  
   Technische Universitaet Darmstadt
   email: machud@intellektik.informatik.tu-darmstadt.de

 This file is part of GCP.

 GCP is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 Foobar is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with GCP; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#ifndef RUNNERS_H
#define RUNNERS_H


#include <stdio.h>
#include <stdlib.h>
#include <cfloat>
#include <vector>
#include <list>
#include "Problem.h"
#include "Control.h"
#include "Random.h"
#include "Solution.h"
#include "Explorer.h"
#include "Manager.h"
#include "Move.h"



class ColoringRunner 
{
public:
	ColoringRunner(ColoringManager* s, ColNeighborhoodExplorer* e,
			Random *rnd, Control *ctrl, Graph* in);

	ColoringRunner(ColoringManager* s,
			Random *rnd, Control *ctrl, Graph* in);

	virtual ~ColoringRunner() {};

	void SetCurrentState(const Coloring& s);

	Coloring GetCurrentState();
	fvalue CurrentStateCost();
	Coloring GetBestState();
	fvalue BestStateCost();

	void ComputeCost();

	//bool LowerBoundReached();
	Longint NumberOfIterations() const {return number_of_iterations;};
	//unsigned long MaxIteration() const;
	void SetMaxIteration(Longint max) {max_iteration = max;}
	void SetMaxIdleIteration(Longint max) {max_idle_iteration = max;}

	void initialization_procedure(Coloring& ass, int representation);
	void decreasing_procedure(Coloring &ass, int representation);

	virtual void Run(Coloring &initial_ass);
	virtual void Run(CompleteColoring &initial_ass) {};
	virtual void Run(PartialColoring &initial_ass) {};

	void setColExplorer(ColNeighborhoodExplorer* e) { p_ne = e;}
	void update(Coloring &a) { p_sm->updateColoringData(a);}
	ColoringManager *getColoringManager() {return p_sm;};
	ColNeighborhoodExplorer *p_ne;

protected:
	void InitializeRun() ;
	void TerminateRun() {};
	void ComputeMoveCost();
	void UpdateIterationCounter() {number_of_iterations++;}
	bool MaxIterationExpired() {return number_of_iterations>max_iteration;}

	// input
	Graph* p_in;
	// helpers
	ColoringManager* p_sm;


	// state data
	Random *rg;
	Control *p_ctrl;

	Longint idle_iterations;
	Longint number_of_iterations;

	Longint max_iteration;
	Longint max_idle_iteration;

	// std::ostream* pos;
};



class ColoringRunnerUpHill:public ColoringRunner
{
public:
	ColoringRunnerUpHill(ColoringManager* s, ColNeighborhoodExplorer* e,
			Random *rnd, Control *ctrl, Graph* in = NULL):
				ColoringRunner(s,e,rnd,ctrl,in) {};

	virtual ~ColoringRunnerUpHill() {};
	void Run(Coloring &initial_ass);
protected:
};


class ColoringTabuSearch:public ColoringRunner
{
public: // FIXME: da riportare anche nel .cpp
	ColoringTabuSearch(ColoringManager* s, ColNeighborhoodExplorer* e,
			Random *rnd, Control *ctrl, TabuList *tl, Graph* in = NULL):
				ColoringRunner(s,e,rnd,ctrl,in), p_tl(tl)
	{};

	virtual ~ColoringTabuSearch() {};
	void Run(CompleteColoring &initial_ass);
protected:
	TabuList *p_tl;
};

class ColoringTabuHashSearch:public ColoringRunner
{
public:
	ColoringTabuHashSearch(ColoringManager* s, ColNeighborhoodExplorer* e,
			Random *rnd, Control *ctrl, Graph* in = NULL):
				ColoringRunner(s,e,rnd,ctrl,in) {};

	virtual ~ColoringTabuHashSearch() {};
	void Run(CompleteColoring &initial_ass);
protected:

};

class ColoringReactiveTabuSearch:public ColoringRunner
{
public:
	ColoringReactiveTabuSearch(ColoringManager* s, ColNeighborhoodExplorer* e,
			Random *rnd, Control *ctrl, TabuList *tl, Graph* in = NULL):
				ColoringRunner(s,e,rnd,ctrl,in), p_tl(tl)
	{};

	virtual ~ColoringReactiveTabuSearch() {};
	void Run(CompleteColoring &initial_ass);
protected:
	TabuList *p_tl;
};


//tabu search without aspiration criteria
class ColoringTabuSearchWithoutAsp:public ColoringRunner
{
public:
	ColoringTabuSearchWithoutAsp(ColoringManager* s, ColNeighborhoodExplorer* e,
			Random *rnd, Control *ctrl, Graph* in = NULL):
				ColoringRunner(s,e,rnd,ctrl,in) {};

	virtual ~ColoringTabuSearchWithoutAsp() {};
	void Run(CompleteColoring &initial_ass);
protected:
};


class ColoringPartialCol:public ColoringRunner
{
public:
	ColoringPartialCol(ColoringManager* s, ColNeighborhoodExplorer* e,
			Random *rnd, Control *ctrl, TabuList *tl, Graph* in = NULL):
				ColoringRunner(s,e,rnd,ctrl,in), p_tl(tl)
	{};

	virtual ~ColoringPartialCol() {};
	void Run(PartialColoring &initial_ass);
protected:
	TabuList *p_tl;
};



class SimulatedAnnealing:public ColoringRunner
{
public:
	SimulatedAnnealing(ColoringManager* s, ColNeighborhoodExplorer* e,
			Random *rnd, Control *ctrl, Graph* in = NULL):
				ColoringRunner(s,e,rnd,ctrl,in) {};

	virtual ~SimulatedAnnealing() {};
	void Run(CompleteColoring &initial_ass);
protected:
	bool SACriterion(fvalue delta, double Temp);
};




class GuidedLocalSearch:public ColoringRunner
{
public:
	GuidedLocalSearch(ColoringManager* s, ColNeighborhoodExplorer* e,
			Random *rnd, Control *ctrl, Graph* in = NULL):
				ColoringRunner(s,e,rnd,ctrl,in) {};

	virtual ~GuidedLocalSearch() {};

	void Run(CompleteColoring &initial_ass);

protected:
};

class GuidedLocalSearchScaled:public ColoringRunner
{
public:
	GuidedLocalSearchScaled(ColoringManager* s, ColNeighborhoodExplorer* e,
			Random *rnd, Control *ctrl, Graph* in = NULL):
				ColoringRunner(s,e,rnd,ctrl,in) {};

	virtual ~GuidedLocalSearchScaled() {};

	void Run(CompleteColoring &initial_ass);

protected:
};




class IteratedLocalSearch:public ColoringRunner
{
public:
	IteratedLocalSearch(ColoringManager* s, ColNeighborhoodExplorer* e,
			Random *rnd, Control *ctrl, Graph* in = NULL):
				ColoringRunner(s,e,rnd,ctrl,in) {};

	virtual ~IteratedLocalSearch() {};
	void Run(CompleteColoring &initial_ass);
protected:
	void LocalSearch(CompleteColoring &initial_state, TabuList &tl,
			Longint max_idle_iterations_ls);
};

class ILS_adj:public IteratedLocalSearch
{
public:
	ILS_adj(ColoringManager* s, ColNeighborhoodExplorer* e,
			Random *rnd, Control *ctrl, Graph* in = NULL):
				IteratedLocalSearch(s,e,rnd,ctrl,in) {};

	virtual ~ILS_adj() {};
	void Run(CompleteColoring &initial_ass);

};




class MinConflicts:public ColoringRunner
{
public:
	MinConflicts(ColoringManager* s, ColNeighborhoodExplorer* e,
			Random *rnd, Control *ctrl, TabuList *tl, Graph* in = NULL):
				ColoringRunner(s,e,rnd,ctrl,in), p_tl(tl) {};

	virtual ~MinConflicts() {};
	void Run(CompleteColoring &initial_ass);
protected:
	TabuList* p_tl;
};

class NoveltyRunner:public ColoringRunner
{
public:
	NoveltyRunner(ColoringManager* s, ColNeighborhoodExplorer* e,
			Random *rnd, Control *ctrl, Graph* in = NULL):
				ColoringRunner(s,e,rnd,ctrl,in) {};

	virtual ~NoveltyRunner() {};
	void Run(CompleteColoring &initial_ass);

};



class XRLF:public ColoringRunner {
public:
	XRLF(ColoringManager* s,
			Random *rnd, Control *ctrl, Graph* in);

	virtual ~XRLF();
	void Run(Coloring &initial_ass);
protected:
	//int edge(int x,int y);
	inline int edge(Vertex x,Vertex y) {
		return (p_in->Edge(Gr[x],Gr[y]));
	}
	void findMaxIndepSet(Vertex Grsize);

	void exhaustive_search(register vector<Vertex> &Wi,
			register Vertex size,
			register list<Vertex> &Wprime,
			register list<Vertex> &Wstar,
			register Vertex current_cost);
	bool check(list<Vertex> &a);
	bool check(vector<Vertex> &a);

	//unsigned char **Bitmap;
	vector<Vertex> Gr; //check correspondence between graph vertices
	vector<Vertex> degree;
	Vertex N;			/* number of vertices in graph */

	unsigned int setlim;	/* size at which exhaustive search begins */
	unsigned int trialnum;	/* number of random starts per color class */
	unsigned int candnum;	/* number of candidates for "random" vertices */
	unsigned int iterations;	/* number of iterations */

	Vertex best_edges;
	vector< Vertex > Cstar;
	Vertex Wsize;

	vector< Vertex > worth_edges;
};



class Ex_DSATUR:public ColoringRunner {
public:
	Ex_DSATUR(ColoringManager* s,
			Random *rnd, Control *ctrl, Graph* in);

	~Ex_DSATUR();
	void Run(Coloring &initial_ass);

private:
	int greedy_clique(int *valid,int *clique);

	int max_w_clique(int *valid,int *clique,int lower,int target);
	void AssignColor(int node,int color);
	void RemoveColor(int node,int color);
	int color(int i,int current_color);
	void print_colors();

	int *ColorClass;
	int prob_count;
	int *Order;
	int *Handled;
	int **ColorAdj;
	int *ColorCount;
	int lb;
	int num_prob,max_prob;
	int num_node;
	int best_clique;
	int BestColoring;
	Coloring *p_col;
};




#endif


