/* GCP version 3.5 Copyright (C) 2003 Marco Chiarandini  
 Technische Universitaet Darmstadt
 email: machud@intellektik.informatik.tu-darmstadt.de

 This file is part of GCP.

 GCP is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.
 
 Foobar is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with GCP; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <algorithm>
#include <cmath>

#include <iostream>
#include <fstream>
#include "Solution.h"
//#include <string>
//#include <sstream>

using namespace std;
//using std::ostringstream; using std::string;


std::ostream& operator<<(std::ostream& os, const Coloring& coloring) {
	const Graph& g = *coloring.p_in; // g is an alias for the input Graph

#ifdef DEBUGGING
	int viol = 0;

	// write the coloring
	for (Vertex v = 0; v < g.Vertices(); v++) {
		os << v << "[" << coloring.col[v] << "], ";
		for (Vertex w = v + 1; w < g.Vertices(); w++)
		if (g.Edge(v, w) && coloring.col[v] == coloring.col[w])
		viol++;
	}
	for (Color c=0; c<coloring.k; c++)
	os<<coloring.partitions[c].size()<<" ";
	os<<endl;
	os << endl << " Colors: " << coloring.k << " Conflicting edges: "
	<< viol << ", Conflicting nodes: "
	<< coloring.conflicts_set.size() << " Fitness: "
	<< coloring.cost << endl;
	//copy(coloring.conflicts.begin(), coloring.conflicts.end(),
	//ostream_iterator<Vertex>(os, " "));
	//copy(coloring.adj_x_color.begin(), coloring.adj_x_color.end(), ostream_iterator<Vertex>(os, " "));
#else

	for (Vertex v = 0; v < g.Vertices(); v++) {
		os << coloring.col[v] + 1 << endl;
	}
#endif
	return os;
}

Coloring::Coloring(Random* rnd, Graph* pin) {
	rg = rnd;
	// initialize the delta map
	p_in = pin;
	col.resize(pin->Vertices());

	k = COLOR_MAX;
	feasible = false;
}

Coloring::Coloring(const Coloring &a) {
	p_in = a.p_in;
	rg = a.rg;

	k = a.k;
	cost = a.cost;

	col = a.col;

	partitions = a.partitions;
	loop_on_partition = a.loop_on_partition;
	feasible = a.feasible;

	pen_delta = a.pen_delta;
	pen_adj_x_color = a.pen_adj_x_color;

}


Coloring& Coloring::operator=(const Coloring &a) {
	if (this == &a)
		return *this;

	p_in = a.p_in;
	rg = a.rg;

	k = a.k;
	cost = a.cost;

	col = a.col;

	adj_x_color = a.adj_x_color;

	partitions = a.partitions;
	feasible = a.feasible;
	loop_on_partition = a.loop_on_partition;

	pen_delta = a.pen_delta;
	pen_adj_x_color = a.pen_adj_x_color;
	return *this;
}


bool Coloring::operator==(const Coloring &a) {

	if (p_in != a.p_in) {
		cerr << "input data!" << endl;
		return false;
	}
	if (rg != a.rg) {
		cerr << "random data!" << endl;
		return false;
	}

	if (k != a.k) {
		cerr << "k!" << endl;
		return false;
	}

	if (cost != a.cost) {
		cerr << "fitness data!" << endl;
		cerr << cost << " " << a.cost << endl;
		return false;
	}

	if (feasible != a.feasible) {
		cerr << "feasible data!";

		return false;
	}

	if (loop_on_partition != a.loop_on_partition) {
		cerr << "loop_on_partition data!" << endl;
		return false;
	}

	if (col != a.col) {
		cerr << "col data!" << endl;
		return false;
	}

	if (adj_x_color != a.adj_x_color) {
		cerr << "adj_color data!" << endl;
		return false;
	}

	for (Color i = 0; i < a.k; i++)
		if (partitions[i] != a.partitions[i]) {
			cerr << "partitions data!" << endl;
			return false;
		}

	return true;
}


void Coloring::SetInput(Graph* pin) {
	p_in = pin;
	col.resize(pin->Vertices());
}

void Coloring::SetColors(Color num_col) {
	this->k = num_col;
	adj_x_color.clear();
	adj_x_color.resize(k * p_in->Vertices());
}


void Coloring::print_pen_delta() {
	ofstream pen_os("tmp/gls-pen_profile.dat");
	for (Vertex i = 0; i < p_in->Vertices(); i++)
		for (Color j = 0; j < k; j++)
			pen_os << PenDelta(i, j) << " ";
	pen_os << endl;
	pen_os.close();
}

fvalue& Coloring::AdjColor(const Vertex& v, const Color& c) {
	return adj_x_color[v * k + c];
}

const fvalue& Coloring::AdjColor(const Vertex& v, const Color& c) const {
	return adj_x_color[v * k + c];
}

float Coloring::PenDelta(const Vertex& v, const Color& c) const {
	if (c == col[v])
		return FLT_MAX;
	else {
		return pen_adj_x_color[v * k + c] - pen_adj_x_color[v * k + col[v]];
	}
}

float& Coloring::PenAdjColor(const Vertex& v, const Color& c) {
	return pen_adj_x_color[v * k + c];
}

const float& Coloring::PenAdjColor(const Vertex& v, const Color& c) const {
	return pen_adj_x_color[v * k + c];
}

//-----------------------------------------------------------------------------

CompleteColoring::CompleteColoring(const CompleteColoring &a) {
	p_in = a.p_in;
	rg = a.rg;

	k = a.k;
	cost = a.cost;

	col = a.col;
	adj_x_color = a.adj_x_color;

	conflicts = a.conflicts;
#ifdef CONFLICT_EDGES
	conflict_edges = a.conflict_edges;
#endif
	conflicts_set = a.conflicts_set;
	partitions = a.partitions;
	loop_on_partition = a.loop_on_partition;
	feasible = a.feasible;

	pen_delta = a.pen_delta;
	pen_adj_x_color = a.pen_adj_x_color;

}

CompleteColoring::CompleteColoring(const Coloring &a) {
	p_in = a.p_in;
	rg = a.rg;

	k = a.k;
	cost = a.cost;

	col = a.col;
	adj_x_color = a.adj_x_color;

	partitions = a.partitions;
	loop_on_partition = a.loop_on_partition;
	feasible = a.feasible;

	pen_delta = a.pen_delta;
	pen_adj_x_color = a.pen_adj_x_color;
}

CompleteColoring& CompleteColoring::operator=(const CompleteColoring &a) {
	if (this == &a)
		return *this;

	p_in = a.p_in;
	rg = a.rg;

	k = a.k;
	cost = a.cost;
	col = a.col;

	adj_x_color = a.adj_x_color;
	conflicts = a.conflicts;
	conflicts_set = a.conflicts_set;
#ifdef CONFLICT_EDGES
	conflict_edges = a.conflict_edges;
#endif
	partitions = a.partitions;
	feasible = a.feasible;
	loop_on_partition = a.loop_on_partition;

	pen_delta = a.pen_delta;
	pen_adj_x_color = a.pen_adj_x_color;
	return *this;
}


bool CompleteColoring::operator==(const CompleteColoring &a) {

	if (p_in != a.p_in) {
		cerr << "input data!" << endl;
		return false;
	}
	if (rg != a.rg) {
		cerr << "random data!" << endl;
		return false;
	}

	if (k != a.k) {
		cerr << "k!" << endl;
		return false;
	}
	if (conflicts != a.conflicts) {
		cerr << "conflicts data! " << endl;
		return false;
	}
#ifdef CONFLICT_EDGES
	if (conflict_edges != a.conflict_edges) {
		cerr << "conflicting edges data! " << endl;
		set<pair<Vertex, Vertex> >::iterator pt;
		for (pt = conflict_edges.begin(); pt != conflict_edges.end(); pt++) {
			cout << "[" << (*pt).first << "," << (*pt).second << "] ";
		}
		cout << endl;
		for (pt = a.conflict_edges.begin(); pt != a.conflict_edges.end(); pt++) {
			cout << "[" << (*pt).first << "," << (*pt).second << "] ";
		}
		cout << endl;
		return false;
	}
#endif
	if (cost != a.cost) {
		cerr << "fitness data!" << endl;
		cerr << cost << " " << a.cost << endl;
		return false;
	}

	if (feasible != a.feasible) {
		cerr << "feasible data!";

		return false;
	}

	if (loop_on_partition != a.loop_on_partition) {
		cerr << "loop_on_partition data!" << endl;
		return false;
	}

	if (col != a.col) {
		cerr << "col data!" << endl;
		return false;
	}

	if (adj_x_color != a.adj_x_color) {
		cerr << "adj_color data!" << endl;
		return false;
	}

	for (Color i = 0; i < a.k; i++)
		if (partitions[i] != a.partitions[i]) {
			cerr << "partitions data!" << endl;
			return false;
		}

	if (conflicts_set != a.conflicts_set) {
		cerr << "conflicts set data!" << endl;
		return false;
	}

	return true;
}



void CompleteColoring::SetInput(Graph* pin) {
	p_in = pin;
	col.resize(pin->Vertices());

	// finally clear the conflict set
	conflicts_set.clear();
#ifdef CONFLICT_EDGES
	conflict_edges.clear();
#endif
	conflicts.resize(pin->Vertices(), 0);
}



Vertex CompleteColoring::choose_conflict() const {
	VertexSet::const_iterator ci = conflicts_set.begin();
	unsigned int n = rg->ran0N(conflicts_set.size() - 1);
	for (unsigned int i = 0; i < n; i++, ci++)
		;
	return *ci;
}



inline fvalue CompleteColoring::Delta(const ColMove& m) const {
	switch (m.type) {
	case ONE_EX:
		return Delta(m.seq[0], m.seq[1]);
	case PEN_ONE_EX:
		return (fvalue) PenDelta(m.seq[0], m.seq[1]);
	default:
		cerr << "Move type not installed in Coloring::Delta(const ColMove &)";
		exit(EXIT_FAILURE);
	}
}



fvalue CompleteColoring::ImpGraph(const Vertex& v, const Vertex& w) const {
	if (v == w)
		return FVALUE_MAX;
	else if (w >= p_in->Vertices() && v >= p_in->Vertices()) {
		return 0;
	} else if (w >= p_in->Vertices())//v<p_in->Vertices()
	{
		return adj_x_color[v * k + w - p_in->Vertices()];
	} else if (v >= p_in->Vertices())//v>p_in->Vertices()
	{
		return 0 - adj_x_color[w * k + col[w]];
	} else if (col[v] == col[w])
		return FVALUE_MAX;
	else {
		return adj_x_color[v * k + col[w]] - adj_x_color[w * k + col[w]]
				- p_in->Edge(v, w);
	}
}

Color CompleteColoring::Class(const Vertex v) const {
	Color l = (Color) v - p_in->Vertices();
	if (l >= 0)
		return l;
	else
		return col[v];
}

//-----------------------------------------------------------------------------


PartialColoring::PartialColoring(const PartialColoring &a) {
	p_in = a.p_in;
	rg = a.rg;

	k = a.k;
	cost = a.cost;
	//col.resize(p_in->Vertices());
	col = a.col;
	//delta = a.delta;
	//adj_x_color.resize(k*p_in->Vertices());
	adj_x_color = a.adj_x_color;

	impasse_class = a.impasse_class;
	partitions = a.partitions;
	loop_on_partition = a.loop_on_partition;
	feasible = a.feasible;

	pen_delta = a.pen_delta;
	pen_adj_x_color = a.pen_adj_x_color;

}

PartialColoring::PartialColoring(const Coloring &a) {
	p_in = a.p_in;
	rg = a.rg;

	k = a.k;
	cost = a.cost;
	//col.resize(p_in->Vertices());
	col = a.col;
	//delta = a.delta;
	//adj_x_color.resize(k*p_in->Vertices());
	adj_x_color = a.adj_x_color;

	partitions = a.partitions;
	loop_on_partition = a.loop_on_partition;
	feasible = a.feasible;

	pen_delta = a.pen_delta;
	pen_adj_x_color = a.pen_adj_x_color;

}

PartialColoring& PartialColoring::operator=(const PartialColoring &a) {
	if (this == &a)
		return *this;

	p_in = a.p_in;
	rg = a.rg;

	k = a.k;
	cost = a.cost;

	col = a.col;

	adj_x_color = a.adj_x_color;
	impasse_class = a.impasse_class;
	partitions = a.partitions;
	feasible = a.feasible;
	loop_on_partition = a.loop_on_partition;

	pen_delta = a.pen_delta;
	pen_adj_x_color = a.pen_adj_x_color;
	return *this;
}

bool PartialColoring::operator==(const PartialColoring &a) {

	if (p_in != a.p_in) {
		cerr << "input data!" << endl;
		return false;
	}
	if (rg != a.rg) {
		cerr << "random data!" << endl;
		return false;
	}

	if (k != a.k) {
		cerr << "k!" << endl;
		return false;
	}

	if (loop_on_partition != a.loop_on_partition) {
		cerr << "loop_on_partition data!" << endl;
		return false;
	}

	if (col != a.col) {
		cerr << "col data!" << endl;
		return false;
	}

	if (adj_x_color != a.adj_x_color) {
		cerr << "adj_color data!" << endl;
		for (Vertex i =0; i<p_in->Vertices(); i++)
			for (int c =0; c<k; c++)
				if (AdjColor(i,c)!=a.AdjColor(i,c))
					cout<<AdjColor(i,c)<<" "<<a.AdjColor(i,c)<<endl;

		return false;
	}

	for (Color i = 0; i < a.k; i++)
		if (partitions[i] != a.partitions[i]) {
			cerr << "partitions data!" << endl;
			return false;
		}

	if (impasse_class != a.impasse_class) {
		cerr << "impasse class differ!" << endl;
		return false;
	}

	if (cost != a.cost) {
			cerr << "fitness data!" << endl;
			cerr << cost << " " << a.cost << endl;
			return false;
		}

		if (feasible != a.feasible) {
			cerr << "feasible data!";

			return false;
		}
	return true;
}





fvalue PartialColoring::Delta(const ColMove& m) const {
	switch (m.type) {
	case I_SWAP:
		return Delta(m.seq[0], m.seq[1]);
	default:
		cerr
				<< "Move type not installed in PartialColoring::Delta(const ColMove &)";
		exit(EXIT_FAILURE);
	}
}




/*
 *
Coloring& operator=(const PartialColoring&) {
	if (this == &a)
			return *this;

		p_in = a.p_in;
		rg = a.rg;

		k = a.k;
		cost = a.cost;

		col = a.col;

		adj_x_color = a.adj_x_color;

		partitions = a.partitions;
		feasible = a.feasible;
		loop_on_partition = a.loop_on_partition;

		pen_delta = a.pen_delta;
		pen_adj_x_color = a.pen_adj_x_color;
		return *this;
}


Coloring& operator=(const CompleteColoring&) {
	if (this == &a)
			return *this;

		p_in = a.p_in;
		rg = a.rg;

		k = a.k;
		cost = a.cost;

		col = a.col;

		adj_x_color = a.adj_x_color;

		partitions = a.partitions;
		feasible = a.feasible;
		loop_on_partition = a.loop_on_partition;

		pen_delta = a.pen_delta;
		pen_adj_x_color = a.pen_adj_x_color;
		return *this;
}
*/

