/* GCP version 3.5 Copyright (C) 2003 Marco Chiarandini  
 Technische Universitaet Darmstadt
 email: machud@intellektik.informatik.tu-darmstadt.de

 This file is part of GCP.

 GCP is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 GCP is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with GCP; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef SOLUTION_H
#define SOLUTION_H
#include <vector>
#include "Problem.h"
#include "Move.h"
#include "Control.h"
#include "Random.h"

using namespace std;

class ColMove;

class Coloring {
	friend std::ostream& operator<<(std::ostream&, const Coloring&);

public:
	Coloring(Random* rnd, Graph* pin);
	Coloring() :
		p_in(0), rg(0) {
	}
	;
	~Coloring() {};
	Coloring(const Coloring &);
	bool operator==(const Coloring &a);
	Coloring& operator=(const Coloring &a);


	virtual void SetInput(Graph* pin);
	void SetColors(Color num_col);
	const Color CurrColors() const {
		return k;
	}
	;

	virtual fvalue Delta(const ColMove& m) const {return 0;};
	inline virtual fvalue Delta(const Vertex& v, const Color& c) const {return 0;};

	fvalue& AdjColor(const Vertex& v, const Color& c);
	const fvalue &AdjColor(const Vertex& v, const Color& c) const;

	float PenDelta(const Vertex& v, const Color& c) const;
	float& PenAdjColor(const Vertex& v, const Color& c);
	const float &PenAdjColor(const Vertex& v, const Color& c) const;

	ColorMap col; // the map between nodes of the graphs and colors
	vector<VertexSet> partitions;
	fvalue cost;

	bool feasible;
	bool loop_on_partition;

	vector<fvalue> adj_x_color;
	void print_pen_delta();
	vector<vector<float> > pen_delta;


	Graph *p_in;
	Random* rg;
	vector<float> pen_adj_x_color;
	Color k; // the actual number of colors used
};

class CompleteColoring : public Coloring {
	friend std::ostream& operator<<(std::ostream&, const Coloring&);
public:
	CompleteColoring(Random* rnd, Graph* pin) :
		Coloring(rnd, pin) {
		conflicts_set.clear();
		conflicts.resize(pin->Vertices(), 0);
#ifdef CONFLICT_EDGES
		conflict_edges.clear();
#endif
	}
	;
	CompleteColoring() :
		Coloring(0, 0) {
		conflicts_set.clear();
#ifdef CONFLICT_EDGES
		conflict_edges.clear();
#endif
	}
	;
	~CompleteColoring() {};
	CompleteColoring(const Coloring & s);
	CompleteColoring(const CompleteColoring & s);
	bool operator==(const CompleteColoring &a);
	CompleteColoring& operator=(const CompleteColoring &a);

	virtual void SetInput(Graph* pin);

	fvalue Delta(const ColMove& m) const;
	inline fvalue Delta(const Vertex& v, const Color& c) const {
		if (c == col[v])
		  return FVALUE_MAX;
		else {
		  return adj_x_color[v * k + c] - adj_x_color[v * k + col[v]];
		}

	}

	Color Class(const Vertex v) const;
	fvalue ImpGraph(const Vertex& v, const Vertex& w) const;

	Vertex choose_conflict() const;

	bool InConflicts(const Vertex& v) const {
		return conflicts_set.find(v) != conflicts_set.end();
	}

	//protected:
	vector<Vertex> conflicts; // the set of conflicting nodes
	VertexSet conflicts_set;
#ifdef CONFLICT_EDGES
	set<pair<Vertex, Vertex> > conflict_edges;
#endif
};

class PartialColoring: public Coloring {

public:
	PartialColoring(Random* rnd=NULL, Graph* pin=NULL) :
		Coloring(rnd, pin) {
		impasse_class.clear();
	}

	~PartialColoring() {};
	PartialColoring(const Coloring &s);
	PartialColoring(const PartialColoring &s);
	bool operator==(const PartialColoring &a);
	PartialColoring& operator=(const PartialColoring &a);

	fvalue Delta(const ColMove& m) const;
	inline fvalue Delta(const Vertex& v, const Color& c) const {
		if (c == col[v])
			return FVALUE_MAX;
		else {
		  return adj_x_color[v * k + c] - 1;
		}
	}

	VertexSet impasse_class;
};

#endif
