/*
 * TabuCol.cpp
 *
 *  Created on: Sep 27, 2010
 *      Author: marco
 */


#include "Runners.h"

#include <algorithm>
#include <cmath>
#include <iostream>
using namespace std;
/*
 * Scheme from
 * A. Hertz and D. de Werra, Using Tabu search techniques for graph coloring, Computing 39 (1987), pp. 345351.
 * with tabu tenure 0.6 * |V^c| + RANDOM(0, 9) from
 * P. Galinier and J.K. Hao, Hybrid evolutionary algorithms for graph coloring, Journal of Combinatorial Optimization 3 (4) (1999), pp. 379397.
 */
void ColoringTabuSearch::Run(CompleteColoring &initial_ass) {
	p_ctrl->getOutputStream() << "TabuSearch (TabuCol)\n";
	InitializeRun();
	CompleteColoring current_state(initial_ass);
	ColMove mv;
	bool tl_created = false;


	if (p_tl == NULL) {
		p_tl = new TabuList();
		tl_created = true;
	}
	p_tl->Init(p_in->Vertices(), current_state.CurrColors(), rg, p_ctrl);

	while ((current_state.cost != 0) && !(MaxIterationExpired())
			&& (idle_iterations < max_idle_iteration) && p_ctrl->isTimeLeft()) {

		UpdateIterationCounter();
		p_ne->BestMove(current_state, mv, *p_tl, number_of_iterations,
				initial_ass.cost);
#ifdef TEST_RUNNER
		CompleteColoring test_state(current_state);
		//cout << test_state<<endl;
		assert(test_state == current_state);
		cout << mv << endl;
#endif
		p_tl->UpdateTabuList(current_state, mv, number_of_iterations);
		p_ne->MakeMove(current_state, mv);
#ifdef TEST_RUNNER
		p_ne->MakeMove(test_state, mv);
		//test_state.col[mv.seq[0]] = mv.seq[1];

		//	  current_state.cost += mv.delta;

		p_sm->updateColoringData(test_state);
		//cout << test_state<<endl;
		if (!(test_state == current_state)) {
			exit(1);
		}
		//cerr<<mv.delta<<" "<<current_state.cost<<" "<<number_of_iterations<<endl;
		//cout << current_state;
#endif
		if (current_state.cost != 0)
			current_state.feasible = false;
		else
			current_state.feasible = true;

		if (current_state.cost < initial_ass.cost) {
			initial_ass = current_state;
			//	  ofstream ooo("tmp-sol.col");
			//	  ooo<<current_state;
			//	  ooo.close();
			p_ctrl->setCurrentCost(current_state.cost,
					current_state.CurrColors(), number_of_iterations);
			idle_iterations = 0;

		}
		idle_iterations++;

	}
	TerminateRun();
	if (tl_created) {
		delete p_tl;
		p_tl = NULL;
	}
	p_ctrl->getOutputStream() << "Done (TabuCol)\n";
}

