/*
 * TabuHashSearch.cpp
 *
 *  Created on: Jul 28, 2010
 *      Author: marco
 * implements an exact tabu search that
 * uses hashing to understand whether a past solution has been seen
 * does not need parameter for tabu length
 *
 */

#include <algorithm>
#include <cmath>
#include <iostream>

#include "Runners.h"
#include "Hash.h"

using namespace std;

void ColoringTabuHashSearch::Run(CompleteColoring &initial_ass) {
	p_ctrl->getOutputStream() << "TabuHashSearch\n";
	InitializeRun();
	CompleteColoring current_state(initial_ass);
	ColMove mv;

	HashForTabu *hh = new HashForTabu(p_in->Vertices(),
			current_state.CurrColors());

	hh->compute_hash_value(current_state);

	while ((current_state.cost != 0) && !(MaxIterationExpired())
			&& (idle_iterations < max_idle_iteration) && p_ctrl->isTimeLeft()) {
		UpdateIterationCounter();
		p_ne->BestMove(current_state, mv, hh, initial_ass.cost);

#ifdef TEST_RUNNER
		CompleteColoring test_state(current_state);
		//cout << test_state<<endl;
		if (!(test_state == current_state)) {
			cerr << "before" << endl;
			exit(1);
		}
		cout << mv << endl;
		;
#endif
		hh->set_current_hash_value(hh->compute_hash_value_incrementally(
				current_state, mv));

		hh->update_data(current_state, mv);
		hh->insert_in_hash_table(number_of_iterations);
		p_ne->MakeMove(current_state, mv);

#ifdef TEST_RUNNER
		p_ne->MakeMove(test_state, mv);
		//test_state.col[mv.seq[0]] = mv.seq[1];
		//	  current_state.cost += mv.delta;
		p_sm->updateColoringData(test_state);
		//cout << test_state<<endl;
		if (!(test_state == current_state)) {
			exit(1);
		}
		//cout << current_state;
#endif
#ifdef TEST_HASH
		cout << mv << endl;
		Longint tmp = hh->get_current_hash_value();
		assert(tmp == hh->compute_hash_value(current_state));
#endif

		if (current_state.cost != 0)
			current_state.feasible = false;
		else
			current_state.feasible = true;
		if (current_state.cost < initial_ass.cost) {
			initial_ass = current_state;
			p_ctrl->setCurrentCost(current_state.cost,
					current_state.CurrColors(), number_of_iterations);
			idle_iterations = 0;

		}
		idle_iterations++;

	}
	TerminateRun();
	delete hh;
	p_ctrl->getOutputStream() << "Done (TabuHashSearch)\n";
}

