/*
 * TabuList.h
 *
 *  Created on: Jul 19, 2010
 *      Author: marco
 */

#ifndef TABULIST_H_
#define TABULIST_H_

#include "Settings.h"
#include "Solution.h"
#include "Move.h"

class TabuList {
public:
	TabuList(Vertex v, Color c, Random *rnd, Control *ctrl);
	TabuList(Control *ctrl);
	TabuList() {
	}
	;
	virtual ~TabuList() {
	}
	;
	virtual void Init(Vertex v, Color c, Random *rnd, Control *ctrl);
	virtual bool isTabu(const Vertex& v, const Color& c, unsigned long iter) {
		return tabu_list[v * colors + c] >= iter;
	}
	virtual unsigned long int &IterTabu(const Vertex& v, const Color& c) {
		return tabu_list[v * colors + c];
	}
	virtual const unsigned long int IterTabu(const Vertex& v, const Color& c) const {
		return tabu_list[v * colors + c];
	}
	virtual bool isTabu(const Vertex& v, unsigned long iter) {
		cerr << "wrong access 1" << endl;
		return 0;
	}
	virtual unsigned long int &IterTabu(const Vertex& v) {
		cerr << "wrong access 2" << endl;
		return tabu_list[0];
	}
	virtual const unsigned long int IterTabu(const Vertex& v) const {
		cerr << "wrong access 3" << endl;
		return 0;
	}

	virtual void UpdateTabuList(const Coloring& st, const ColMove& mv,
			unsigned long iter);
	double prob;

	virtual bool areEqual(ColMove mv) {
		return false;
	}
	;
	virtual const unsigned long int IterTabuMatrix(const Vertex& v,
			const Vertex& c) const {
		cerr << "wrong access 4" << endl;
		return 0;
	}
	;
	virtual void cleanTable() {
	}
	;

	virtual void UpdateTabuList(ColMove&, ColMove&) {
	}
	;
	virtual set<ColMove, lessMove> &List(const ColMove&) {
		cerr << "&List" << endl;
		return tabu_matrix_list[0][0];
	}
	;
	vector<vector<set<ColMove, lessMove> > > tabu_matrix_list;

	void setTabuLengthType(unsigned int new_type) {
		tl_type = new_type;
	}
	;
	void setTabuLengthParameter(unsigned int new_tl) {
		tl = new_tl;
	}
	;
	unsigned int getTabuLengthParameter() {
		return tl;
	}
	;

protected:
	unsigned long ComputeLength(const fvalue& conflicts);

	Color colors;
	Vertex vertices;
	vector<unsigned long int> tabu_list;
	Random *rg;
	Control *p_ctrl;

	float alpha;
	unsigned int tl_type;
	unsigned int tl;
	unsigned int tl1;
	unsigned int tl2;
	unsigned int invtype;
};

class TabuListVLSN: public TabuList {
public:
	TabuListVLSN(Control *ctrl) :
		TabuList(ctrl) {
	}
	;
	TabuListVLSN(Vertex v, Color col, Random *rnd, Control *ctrl);
	TabuListVLSN() {
	}
	;
	virtual ~TabuListVLSN() {
	}
	;
	void Init(Vertex v, Color c, Random *rnd, Control *ctrl);
	bool isTabu(const Vertex& v1, const Color& c, unsigned long iter);
	void cleanTable();
	void UpdateTabuList(const Coloring& st, const ColMove& mv,
			unsigned long iter);
private:
	Vertex vertices;
};

class TabuListVertices: public TabuList {
public:
	TabuListVertices(Vertex v, Random *rnd, Control *ctrl);
	TabuListVertices() {
	}
	;
	virtual ~TabuListVertices() {
	}
	;
	bool isTabu(const Vertex& v, unsigned long iter) {
		return tabu_list[v] >= iter;
	}
	unsigned long int &IterTabu(const Vertex& v) {
		return tabu_list[v];
	}
	const unsigned long int IterTabu(const Vertex& v) const {
		return tabu_list[v];
	}
	void cleanTable();
	void UpdateTabuList(const Coloring& st, const ColMove& mv,
			unsigned long iter);
};

class TabuListPart: public TabuList {
public:
			TabuListPart(Vertex v, Color c, Random *rnd, Control *ctrl,
					Coloring *state);
	virtual ~TabuListPart() {
	}
	;
	bool isTabu(const Vertex& v, const Color& c, unsigned long iter);

	// bool isTabu(const Vertex v, const Vertex w, unsigned long iter)
	//  {return tabu_list[v*vertices+w]>=iter;}

	unsigned long int &IterTabu(const Vertex& v, const Vertex& w) {
		return tabu_list[v * vertices + w];
	}
	const unsigned long int IterTabu(const Vertex& v, const Vertex& w) const {
		return tabu_list[v * vertices + w];
	}
	void UpdateTabuList(const Coloring& st, const ColMove& mv,
			unsigned long iter);
protected:
	Vertex vertices;
	Coloring *p_st;
};

class TabuListMinConf: public TabuList {
public:
	TabuListMinConf(Vertex v, Color c, Random *rnd, Control *ctrl) :
		TabuList(v, c, rnd, ctrl) {
	}
	;
	virtual ~TabuListMinConf() {
	}
	;
	void UpdateTabuList(const Coloring& st, const ColMove& mv,
			unsigned long iter);
};

class HistoryNovelty: public TabuList {
public:
	HistoryNovelty(Vertex v, Color c, Random *rnd, Control *ctrl) :
		TabuList(v, c, rnd, ctrl) {
	}
	;
	virtual ~HistoryNovelty() {
	}
	;
	void UpdateTabuList(const Coloring& st, const ColMove& mv,
			unsigned long iter);

};

class TabuListDoubleRunner: public TabuList {
public:
	TabuListDoubleRunner(Vertex v, Random *rnd, Control *ctrl);
	virtual ~TabuListDoubleRunner() {
	}
	;
	virtual bool isTabu(const Vertex& u, const Vertex& v, unsigned long iter) {
		return tabu_matrix[u][v] >= iter;
	}
	const unsigned long int IterTabuMatrix(const Vertex& u, const Vertex& v) const {
		return tabu_matrix[u][v];
	}
	virtual void UpdateTabuList(const Coloring& st, const ColMove& mv,
			unsigned long iter);
	bool areEqual(ColMove mv) {
		return (tabu_matrix[mv.seq[0]][mv.seq[2]]
				== tabu_matrix[mv.seq[1]][mv.seq[3]]
				&& tabu_matrix[mv.seq[0]][mv.seq[2]] != 0);
	}
	virtual void cleanTable();
protected:
	vector<vector<unsigned long int> > tabu_matrix;
};

class TabuListPerMoves: public TabuList {
public:
	TabuListPerMoves(Vertex v, Random *rnd, Control *ctrl);
	virtual ~TabuListPerMoves() {
	}
	;
	void UpdateTabuList(ColMove& mv, ColMove& prec_mv);
	set<ColMove, lessMove> &List(const ColMove&);
	void cleanTable();
	//vector< vector < set<ColMove, lessMove> > > tabu_matrix_list;
};

class TabuListNew: public TabuList {
public:
	TabuListNew(Control* ctrl) :
		TabuList(ctrl) {
	}
	;
	virtual ~TabuListNew() {
	}
	;
	virtual void UpdateTabuList(const Coloring& st, const ColMove& mv,
			unsigned long iter);
};

class TabuListNewOld: public TabuList {
public:
	TabuListNewOld(Control* ctrl) :
		TabuList(ctrl) {
	}
	;
	virtual ~TabuListNewOld() {
	}
	;
	virtual void UpdateTabuList(const Coloring& st, const ColMove& mv,
			unsigned long iter);
};

class TabuListVertex: public TabuList {
public:
	TabuListVertex(Control* ctrl) :
		TabuList(ctrl) {
	}
	;
	virtual ~TabuListVertex() {
	}
	;
	virtual void UpdateTabuList(const Coloring& st, const ColMove& mv,
			unsigned long iter);
};

#endif /* TABULIST_H_ */
