/*
 * ReactiveTabuSearch.cpp
 *
 *  Created on: Jul 19, 2010
 *      Author: marco
 */

#include <algorithm>
#include <cmath>
#include <iostream>

#include "Runners.h"
#include "Hash.h"

using namespace std;

void ColoringReactiveTabuSearch::Run(CompleteColoring &initial_ass) {
	p_ctrl->getOutputStream() << "TabuReactiveSearch\n";
	InitializeRun();
	CompleteColoring current_state(initial_ass);
	ColMove mv;
	int new_tl;
	bool tl_created = false;
	unsigned long int iter_without_detection = 0;

	HashForTabu hh(p_in->Vertices(), current_state.CurrColors());

	if (p_tl == NULL) {
		p_tl = new TabuList();
		tl_created = true;
	}
	p_tl->Init(p_in->Vertices(), current_state.CurrColors(), rg, p_ctrl);

	hh.compute_hash_value(current_state);

	while ((current_state.cost != 0) && !(MaxIterationExpired())
			&& (idle_iterations < max_idle_iteration) && p_ctrl->isTimeLeft()) {
		UpdateIterationCounter();
		p_ne->BestMove(current_state, mv, *p_tl, number_of_iterations,
				initial_ass.cost);
		//--------------------------------------
#ifdef TEST_RUNNER
		CompleteColoring test_state(current_state);
		//cout << test_state<<endl;
		if (!(test_state == current_state)) {
			cerr << "before" << endl;
			exit(1);
		}
		cout << mv << endl;
		;
#endif
		//--------------------------------------

		hh.set_current_hash_value(hh.compute_hash_value_incrementally(
				current_state, mv));
		hh.update_data(current_state, mv);

		// change values of tl in a reactive way
		if (!hh.insert_in_hash_table(number_of_iterations)) {
			iter_without_detection = 0;
			p_tl->setTabuLengthType(TL_FIXED);
			new_tl = (int) ceil((Real) p_tl->getTabuLengthParameter() * 1.01);
			if (new_tl < (current_state.cost * current_state.CurrColors())) {
#ifdef MONITOR_TL_CHANGES
				cout << "Increase tl to " << new_tl << endl;
#endif
				p_tl->setTabuLengthParameter(new_tl);
			}

		} else {
			iter_without_detection++;
			if (iter_without_detection > p_in->Vertices()) {
				new_tl = (int) floor((Real) p_tl->getTabuLengthParameter()
						* 0.99);
				if (new_tl > 1) {
#ifdef MONITOR_TL_CHANGES
					cout << "Decrease tl to " << new_tl << endl;
#endif
					p_tl->setTabuLengthParameter(new_tl);

				}
			}
		}

		p_tl->UpdateTabuList(current_state, mv, number_of_iterations);

		p_ne->MakeMove(current_state, mv);

#ifdef TEST_RUNNER
		p_ne->MakeMove(test_state, mv);
		//test_state.col[mv.seq[0]] = mv.seq[1];
		//	  current_state.cost += mv.delta;
		p_sm->updateColoringData(test_state);
		//cout << test_state<<endl;
		if (!(test_state == current_state)) {
			exit(1);
		}
		//cout << current_state;
#endif
#ifdef TEST_HASH
		cout << mv << endl;
		Longint tmp = hh.get_current_hash_value();
		assert(tmp == hh.compute_hash_value(current_state));
#endif

		if (current_state.cost != 0)
			current_state.feasible = false;
		else
			current_state.feasible = true;
		if (current_state.cost < initial_ass.cost) {
			initial_ass = current_state;
			p_ctrl->setCurrentCost(current_state.cost,
					current_state.CurrColors(), number_of_iterations);
			idle_iterations = 0;

		}
		idle_iterations++;

	}
	TerminateRun();
	if (tl_created) {
		delete p_tl;
		p_tl = NULL;
	}
	p_ctrl->getOutputStream() << "Done (TabuReactiveSearch)\n";
}

