/* -*- mode: c++; mode: font-lock -*- */

#ifndef UTILITIES_H
#define UTILITIES_H

#include<cstdlib>
#include<cstdio>
#include<iostream>
#include<fstream>
#include<cassert>
#include <new>
#include<set>
#include<vector>
#include<cfloat>
#include<cstring>
#include<climits>
#include "Definitions.h"
#include "da_set.h"

using namespace std;

typedef unsigned long ELEMENT;
/*  // This can be changed. It represents the maximum number of */
/*  // colours allowed.  */
#define MAX_BITS   256
#define ELEM_BITS (sizeof(ELEMENT)*8)
#define ELEMS(cBits)  ( (cBits + ELEM_BITS - 1)/ELEM_BITS )

typedef ELEMENT BITSET[ELEMS(MAX_BITS)];

#ifndef max
#define max(a, b) ((a) >= (b) ? (a) : (b))
#endif

#ifndef min
#define min(a, b) ((a) <= (b) ? (a) : (b))
#endif

#define NOT_A_COLOR -1
inline bool isNotAColor(Color x) {return x==NOT_A_COLOR;}


/* // Initializes Bitset with 0. */
inline void InitBitset(BITSET bitset, int cBits) {
	assert(cBits <= MAX_BITS);

	memset(bitset, 0, ELEMS(MAX_BITS) * sizeof(ELEMENT));
}

/* // Copies one bitset over another. */

inline void CopyBitset(BITSET to, const BITSET from) {
	memcpy(to, from, ELEMS(MAX_BITS) * sizeof(ELEMENT));
}

/* // Puts 1 on the position corresponding to "index". */

inline void Set(BITSET bitset, int index) {
	bitset[index / ELEM_BITS] |= 1L << (index % ELEM_BITS);
}

/* // If there is 0 on the position "index", then the function */
/* // returns 0, otherwise it returns 1.  */
inline int Test(const BITSET bitset, int index) {
	return bitset[index / ELEM_BITS] & (1L << (index % ELEM_BITS));
}

/* // Returns 1 if the candidate bitset represents a subset of */
/* // the reference one.  */
inline int Subset(BITSET reference, const BITSET candidate) {
	unsigned int iElem;

	for (iElem = 0; iElem < ELEMS(MAX_BITS); iElem++) {

		if ((reference[iElem] | candidate[iElem]) != reference[iElem]) {

			return 0;
		}
	}

	return 1;
}

/* // Used just to control that everything is fine.  */
inline void PrintSet(BITSET set) {
	int i, first = 1;
	;

	printf("{");

	for (i = 0; i < MAX_BITS; i++) {

		if (Test(set, i)) {

			if (first) {
				printf("%d", i);
				first = 0;
			} else {
				printf(",%d", i);
			}
		}
	}

	printf("}");
}

//  void InitBitset (BITSET bitset,int cBits);
//  void CopyBitset (BITSET to, const BITSET from);
//  void Set(BITSET bitset,int index);
//  int Test(const BITSET bitset,int index);
//  int Subset(BITSET reference, const BITSET candidate);
//  void PrintSet(BITSET set);

//------------------------------------------------------------
// Needed by DSJ XRLF


//#define INT 32		/* computer word size */

extern long int *array;

struct compare {
	bool operator()(const Vertex &s1, const Vertex &s2) const {
		return array[s1] > array[s2];
	}
};

template<typename T> T* docast(T*, void *v) { return reinterpret_cast<T*>(v); }

extern void duplicateMemoryArray(char **to, const char *from, size_t size);


#endif

